/* cie.c */
/*
  The BREP Library.
  Copyright (C) 1996 Philippe Bekaert

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Float.h"
#include "cie.h"

/* this code was taken from the MGF documentation */

#ifdef  NTSC
#define  CIE_x_r                0.670           /* standard NTSC primaries */
#define  CIE_y_r                0.330
#define  CIE_x_g                0.210
#define  CIE_y_g                0.710
#define  CIE_x_b                0.140
#define  CIE_y_b                0.080
#define  CIE_x_w                0.3333333333          /* monitor white point */
#define  CIE_y_w                0.3333333333
#else
#define  CIE_x_r                0.640           /* nominal CRT primaries */
#define  CIE_y_r                0.330
#define  CIE_x_g                0.290
#define  CIE_y_g                0.600
#define  CIE_x_b                0.150
#define  CIE_y_b                0.060
#define  CIE_x_w                0.3333333333          /* monitor white point */
#define  CIE_y_w                0.3333333333
#endif

#define CIE_D           (       CIE_x_r*(CIE_y_g - CIE_y_b) + \
                                CIE_x_g*(CIE_y_b - CIE_y_r) + \
                                CIE_x_b*(CIE_y_r - CIE_y_g)     )
#define CIE_C_rD        ( (1./CIE_y_w) * \
                                ( CIE_x_w*(CIE_y_g - CIE_y_b) - \
                                  CIE_y_w*(CIE_x_g - CIE_x_b) + \
                                  CIE_x_g*CIE_y_b - CIE_x_b*CIE_y_g     ) )
#define CIE_C_gD        ( (1./CIE_y_w) * \
                                ( CIE_x_w*(CIE_y_b - CIE_y_r) - \
                                  CIE_y_w*(CIE_x_b - CIE_x_r) - \
                                  CIE_x_r*CIE_y_b + CIE_x_b*CIE_y_r     ) )
#define CIE_C_bD        ( (1./CIE_y_w) * \
                                ( CIE_x_w*(CIE_y_r - CIE_y_g) - \
                                  CIE_y_w*(CIE_x_r - CIE_x_g) + \
                                  CIE_x_r*CIE_y_g - CIE_x_g*CIE_y_r     ) )

#define CIE_rf          (CIE_y_r*CIE_C_rD/CIE_D)
#define CIE_gf          (CIE_y_g*CIE_C_gD/CIE_D)
#define CIE_bf          (CIE_y_b*CIE_C_bD/CIE_D)

static float  xyz2rgbmat[3][3] = {     /* XYZ to RGB */
        {(CIE_y_g - CIE_y_b - CIE_x_b*CIE_y_g + CIE_y_b*CIE_x_g)/CIE_C_rD,
         (CIE_x_b - CIE_x_g - CIE_x_b*CIE_y_g + CIE_x_g*CIE_y_b)/CIE_C_rD,
         (CIE_x_g*CIE_y_b - CIE_x_b*CIE_y_g)/CIE_C_rD},
        {(CIE_y_b - CIE_y_r - CIE_y_b*CIE_x_r + CIE_y_r*CIE_x_b)/CIE_C_gD,
         (CIE_x_r - CIE_x_b - CIE_x_r*CIE_y_b + CIE_x_b*CIE_y_r)/CIE_C_gD,
         (CIE_x_b*CIE_y_r - CIE_x_r*CIE_y_b)/CIE_C_gD},
        {(CIE_y_r - CIE_y_g - CIE_y_r*CIE_x_g + CIE_y_g*CIE_x_r)/CIE_C_bD,
         (CIE_x_g - CIE_x_r - CIE_x_g*CIE_y_r + CIE_x_r*CIE_y_g)/CIE_C_bD,
         (CIE_x_r*CIE_y_g - CIE_x_g*CIE_y_r)/CIE_C_bD}
};

static float  rgb2xyzmat[3][3] = {     /* RGB to XYZ */
        {CIE_x_r*CIE_C_rD/CIE_D,CIE_x_g*CIE_C_gD/CIE_D,CIE_x_b*CIE_C_bD/CIE_D},
        {CIE_y_r*CIE_C_rD/CIE_D,CIE_y_g*CIE_C_gD/CIE_D,CIE_y_b*CIE_C_bD/CIE_D},
        {(1.-CIE_x_r-CIE_y_r)*CIE_C_rD/CIE_D,
         (1.-CIE_x_g-CIE_y_g)*CIE_C_gD/CIE_D,
         (1.-CIE_x_b-CIE_y_b)*CIE_C_bD/CIE_D}
};

void cie_rgb(float *ciecolor, float *rgbcolor)             /* convert CIE to RGB */
{
  int  i;

  for (i = 0; i < 3; i++) {
    rgbcolor[i] =   xyz2rgbmat[i][0]*ciecolor[0] +
                    xyz2rgbmat[i][1]*ciecolor[1] +
                    xyz2rgbmat[i][2]*ciecolor[2] ;
    if (rgbcolor[i] < 0.0)          /* watch for negative values */
      rgbcolor[i] = 0.0;
  }
}

void rgb_cie(float *rgbcolor, float *ciecolor)             /* convert RGB to CIE */
{
  int  i;

  for (i = 0; i < 3; i++) {
    ciecolor[i] =   rgb2xyzmat[i][0]*rgbcolor[0] +
                    rgb2xyzmat[i][1]*rgbcolor[1] +
                    rgb2xyzmat[i][2]*rgbcolor[2] ;
  }
}

void cie2rgb(float *cie, float *rgb)
{
  float sum;
  
  cie_rgb(cie, rgb);
  sum = rgb[0] + rgb[1] + rgb[2];
  if (sum > 0.) {
    rgb[0] /= sum; rgb[1] /= sum; rgb[2] /= sum; 
  }
}

void CIExyToRGB(Float x, Float y, RGB *color)
{
  Float cie[3], rgb[3];
  Float mexx;

  cie[0] = x;
  cie[1] = y;
  cie[2] = 1.-x-y;
  
  cie2rgb(cie, rgb);

  mexx = rgb[0];
  if (rgb[1] > mexx) mexx = rgb[1];
  if (rgb[2] > mexx) mexx = rgb[2];
  
  if (mexx > EPSILON) {
    rgb[0] /= mexx; 
    rgb[1] /= mexx; 
    rgb[2] /= mexx; 
  }

  RGBSET(*color, rgb[0], rgb[1], rgb[2]);
  RGBCLIP(*color);
}

/* convert MGF color to RGB */
void MGFGetColor(C_COLOR *cin, double intensity, RGB *cout)
{
  c_ccvt(cin, C_CSXY);
  CIExyToRGB(cin->cx, cin->cy, cout);
  RGBSCALE(intensity, *cout, *cout);
}

