/*
  The BREP Library.
  Copyright (C) 1996 Philippe Bekaert

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* geometry.c: geometric data for topological entities such as vertices,
 * edges, ... */

#include <stdio.h>
#include <math.h>

#include "Float.h"
#include "geometry.h"
#include "brep.h"
#include "pools.h"
#include "error.h"
#include "vector.h"
#include "globals.h"

/* #define DDEBUG */
#undef DEBUG

/* ************************* tolerances *********************************** */
/* minimum (relative) vertex, edge and face tolerance */
double vertex_tolerance = 1e-6,
       edge_tolerance = 1e-12,
       face_tolerance = 1e-12;

void SetMinimumVertexTolerance(double tol)
{
  vertex_tolerance = tol;
}

void SetMinimumEdgeTolerance(double tol)
{
  edge_tolerance = tol;
}

void SetMinimumFaceTolerance(double tol)
{
  face_tolerance = tol;
}


/* ********************** Vertex functions ******************************* */
#ifndef NOPOOLS
static POOL *vertexPool = (POOL *)NULL;
#define NEWVERTEX()  	(VERTEX *)New(sizeof(VERTEX), &vertexPool)
#define DISPOSEVERTEX(ptr) Dispose((unsigned char *)(ptr), &vertexPool)
#else /*NOPOOLS*/
#define NEWVERTEX()	(VERTEX *)Alloc(sizeof(VERTEX))
#define DISPOSEVERTEX(ptr) Free((char *)ptr, sizeof(VERTEX))
#endif /*NOPOOLS*/

/* Vertex compare flags:
 *
 * The following data may be taken into consideration when comparing vertices:
 * - vertex location: normally always specified.
 * - vertex normal: (0,0,0) if not specified, normalized vector if specified.
 * - vertex name: NULL if not specified.
 * What data should be taken into account when comparing vertices is 
 * controlled with the vertex compare flags VCMP_LOCATION, 
 * VCMP_NORMAL and VCMP_NAME. 
 *
 * The order in which the data is taken into account by VertexCompare is
 * as follows: first location is tested, than normals, than name (if relevant).
 *
 * When the normal or name is relevant, it also depends on the situation
 * whether or not two vertices without normal (or name) and equal other data
 * are to be considered equal or not. This is controlled with the flags 
 * VCMP_NO_NORMAL_IS_EQUAL_NORMAL and VCMP_NO_NAME_IS_EQUAL_NAME.
 * When the normal (or name) is relevant, and only one of the two
 * vertices has a normal (or name), the vertices are considered different.
 *
 * The flags can be |-ed together.
 */
static unsigned int vertex_compare_flags;

void SetVertexCompareFlags(unsigned int flags)
{
  vertex_compare_flags = flags;
}

/* Compares two vertices and returns a code useful for ordering vertices
 * into an octree. Returns 8 if the two vertices are to be considered
 * the same vertices and 0-7, the index of an octree branch to be expored
 * further, if not. This function is needed in order to sort created
 * vertices into an octree for efficient later lookup. This is taken care of
 * with the functions BrepInstallVertex() and BrepLookupVertex(). For each 
 * solid, a vertex octree is created at the time the solid is created. */
int VertexCompare(VERTEX *v1, VERTEX *v2)
{
  int code;
  double tolerance;

  code = 8;

  /* first compare the coordinates */
  if (vertex_compare_flags & VCMP_LOCATION) {
    /* two entities intersect if their tolerance region intersects, i.o.w. if
     * the distance between them is smaller than the sum of the two entity's 
     * tolerances */
    tolerance = v1->tolerance + v2->tolerance;

    code = 0;

    if (v1->point.x > v2->point.x + tolerance) code += 1;
    if (v1->point.y > v2->point.y + tolerance) code += 2;
    if (v1->point.z > v2->point.z + tolerance) code += 4;
    if (code!=0) 	/* x1 > x2 || y1 > y2 || z1 > z2 */
      return code;

    if (v1->point.x < v2->point.x - tolerance ||
	v1->point.y < v2->point.y - tolerance ||
	v1->point.z < v2->point.z - tolerance)
      return code;	/* not the same coordinates */

    code = 8;
  }

  /* same coordinates, test the normals */
  if (vertex_compare_flags & VCMP_NORMAL) {
    code = 0;
    
    if (v1->normal.x > v2->normal.x + EPSILON) code += 1;
    if (v1->normal.y > v2->normal.y + EPSILON) code += 2;
    if (v1->normal.z > v2->normal.z + EPSILON) code += 4;
    if (code!=0) 	/* x1 > x2 || y1 > y2 || z1 > z2 */
      return code;

    if (v1->normal.x < v2->normal.x - EPSILON ||
	v1->normal.y < v2->normal.y - EPSILON ||
	v1->normal.z < v2->normal.z - EPSILON)
      return code;	/* not the same normal */

    if (!(vertex_compare_flags & VCMP_NO_NORMAL_IS_EQUAL_NORMAL))
      if (v1->normal.x == 0. && v1->normal.y == 0. && v1->normal.z == 0.)
	return 0;				/* no valid normal => not the same vertex */

    code = 8;
  }

  if (vertex_compare_flags & VCMP_NAME) {
    if (v1->name && v2->name) {		/* both have a name, compare */
      int cmp = strcmp(v1->name, v2->name);
      if (cmp < 0)
	return 1;
      else if (cmp > 0)
	return 2;
      else
	return 8;			/* same name => same vertex */
    } else if (v1->name) {		/* v1 has a name, v2 not */
      return 3;
    } else if (v2->name) {		/* v1 has no name, v2 has */
      return 4;
    } else if (vertex_compare_flags & VCMP_NO_NAME_IS_EQUAL_NAME) {
      return 8;				/* none has a name => same vertex */
    } else
      return 0;				/* none has a name => not same vertex */
  }

  return code;
}

/* This routine only compares the location of the two vertices. */
int VertexCompareLocation(VERTEX *v1, VERTEX *v2)
{
  int code;
  double tolerance;

  /* two entities intersect if their tolerance region intersects, i.o.w. if
  * the distance between them is smaller than the sum of the two entity's 
  * tolerances */
  tolerance = v1->tolerance + v2->tolerance;

  code = 0;
  if (v1->point.x > v2->point.x + tolerance) code += 1;
  if (v1->point.y > v2->point.y + tolerance) code += 2;
  if (v1->point.z > v2->point.z + tolerance) code += 4;
  if (code!=0) 	/* x1 > x2 || y1 > y2 || z1 > z2 */
    return code;

  if (v1->point.x < v2->point.x - tolerance ||
      v1->point.y < v2->point.y - tolerance ||
      v1->point.z < v2->point.z - tolerance)
    return code;	/* not the same coordinates */

  return 8;
}

void PrintVertexData(FILE *out, VERTEX *v)
{
  if (v) {
    fprintf(out, "vertex id %d, name '%s', point = (%g,%g,%g), normal = (%g,%g,%g), tolerance = %g, ",
	  v->id & ~IDMASK,
	  v->name ? v->name : "(null)",
	  v->point.x, v->point.y, v->point.z, 
	  v->normal.x, v->normal.y, v->normal.z,
	  v->tolerance);
    fprintf(out, "%shidden, %sprocessed.\n", 
	    v->id&HIDDEN ? "" : "not ",
	    v->id&PROCESSED ? "" : "not ");
  } else
    fprintf(out, "(vertex without data)");
}

static void PrintWingStderr(BREP_WING *wing)
{
  PrintWingShort(stderr, wing); fprintf(stderr, "\n");
}

void PrintVertex(FILE *out, BREP_VERTEX *vertex)
{
  PrintVertexData(out, (VERTEX *)(vertex->client_data));
  fprintf(stderr, "\n\tWings leaving at the vertex:\n");
  BrepVertexIterateWings(vertex, PrintWingStderr);
}

static int TotalVertices = 0, vertexcount = 0;

int GetNumberOfVertices(void)
{
  return TotalVertices;
}

VERTEX *VertexCreateCallback(BREP_VERTEX *v)
{
  VERTEX *vertex, *vertex_data = v->client_data;

  /* duplicate the vertex data */
  vertex = NEWVERTEX();

  if (vertex_data) {
    *vertex = *vertex_data;
    if (vertex_data->name && vertex_data->name[0]) {
      vertex->name = Alloc(strlen(vertex_data->name)+1);
      strcpy(vertex->name, vertex_data->name);
    } else
      vertex->name = NULL;
  }
  vertex->id = ((vertexcount++)&~IDMASK) | VERTEXMASK;	/* so it will be recognized as a vertex */

#ifdef DEBUG
  fprintf(stderr, "VertexCreateCallback:\n");
  PrintVertexData(stderr, vertex_data);
  fprintf(stderr, "\n");
#endif

  TotalVertices ++;
    
  return vertex;
}

/* Computes a normals for the vertex if it doesn't have a normal yet
 * unless normals in the input should be ignored. */
void ComputeVertexNormal(BREP_VERTEX *vertex)
{
  VECTOR *normal = &((VERTEX *)(vertex->client_data))->normal;

  /* don't recompute vertex normals that are given in the input, unless
   * they should be ignored. */
  if (ignore_normals_on_input ||
      (FLOATEQUAL(normal->x, 0., EPSILON) &&
       FLOATEQUAL(normal->y, 0., EPSILON) && 
       FLOATEQUAL(normal->z, 0., EPSILON)))

  /* Compute a normal for the vertex: average normal of the faces sharing 
   * the vertex. */
  VECTORSET(*normal, 0., 0., 0.);
  ForAllWingsLeavingVertex(wing, vertex) {
    VECTORADD(*normal, ((FACE *)(wing->contour->face->client_data))->normal, *normal);    
  } EndForAll;
  VECTORNORMALIZE(*normal);
}

static void ComputeVertexTolerance(VERTEX *vertex_data)
{
  double max, m;
  max = fabs(vertex_data->point.x);
  m = fabs(vertex_data->point.y);
  if (m > max) max = m;
  m = fabs(vertex_data->point.z);
  if (m > max) max = m;
  vertex_data->tolerance = vertex_tolerance * max;
}

VERTEX *VertexCloseCallback(BREP_VERTEX *vertex)
{
  ComputeVertexTolerance(vertex->client_data);
#ifdef DEBUG
  fprintf(stderr, "VertexCloseCallback:\n");
  PrintVertexData(stderr, (VERTEX *)(vertex->client_data));
  fprintf(stderr, "\n");
#endif

  return vertex->client_data;
}

void VertexDestroyCallback(BREP_VERTEX *vertex)
{
  VERTEX *vertex_data = vertex->client_data;

#ifdef DEBUG
  fprintf(stderr, "VertexDestroyCallback:\n");
#endif
  if (vertex_data) {
#ifdef DEBUG
    fprintf(stderr, "\t");
    PrintVertexData(stderr, (VERTEX *)(vertex_data));
    fprintf(stderr, "\n");
#endif
    if (vertex_data->name)
      Free(vertex_data->name, strlen(vertex_data->name)+1);
    DISPOSEVERTEX(vertex_data);
  }

  TotalVertices --;
}


/* ************************ Edge functions **************************** */
#ifndef NOPOOLS
static POOL *edgePool = (POOL *)NULL;
#define NEWEDGE()  	(EDGE *)New(sizeof(EDGE), &edgePool)
#define DISPOSEEDGE(ptr) Dispose((unsigned char *)(ptr), &edgePool)
#else /*NOPOOLS*/
#define NEWEDGE()	(EDGE *)Alloc(sizeof(EDGE))
#define DISPOSEEDGE(ptr) Free((char *)ptr, sizeof(EDGE))
#endif /*NOPOOLS*/

void PrintEdgeData(FILE *out, EDGE *edge_data)
{
  if (edge_data) {
    fprintf(out, "edge id %d, tolerance = %g, extent = %g, near = %g, ", 
	    edge_data->id & ~IDMASK,
	    edge_data->tolerance, edge_data->extent, edge_data->near);
    fprintf(out, "%shidden, %sprocessed.\n", 
	    edge_data->id&HIDDEN ? "" : "not ",
	    edge_data->id&PROCESSED ? "" : "not ");
  } else {
    fprintf(out, "(edge without data)");
  }
}

void PrintEdge(FILE *out, BREP_EDGE *edge)
{
  PrintEdgeData(out, (EDGE *)(edge->client_data));
  fprintf(out, "\n");
  fprintf(out, "\tvertex1 = "); PrintVertexData(out, (VERTEX *)(edge->wing[0].vertex->client_data));
  fprintf(out, "\n");
  fprintf(out, "\tvertex2 = "); PrintVertexData(out, (VERTEX *)(edge->wing[1].vertex->client_data));
  fprintf(out, "\n");
}

void PrintWing(FILE *out, BREP_WING *wing)
{
  BREP_EDGE *edge = wing->edge;

  PrintEdgeData(out, (EDGE *)(edge->client_data));

  if (wing == &(edge->wing[0])) {
    fprintf(out, ", wing 0\n"); 
    fprintf(stderr, "\tvertex1 = "); PrintVertexData(stderr, (VERTEX *)(edge->wing[0].vertex->client_data));
    fprintf(stderr, "\n");
    fprintf(stderr, "\tvertex2 = "); PrintVertexData(stderr, (VERTEX *)(edge->wing[1].vertex->client_data));
    fprintf(stderr, "\n");
  } else if (wing == &(edge->wing[1])) {
    fprintf(out, ", wing 1\n"); 
    fprintf(stderr, "\tvertex1 = "); PrintVertexData(stderr, (VERTEX *)(edge->wing[1].vertex->client_data));
    fprintf(stderr, "\n");
    fprintf(stderr, "\tvertex2 = "); PrintVertexData(stderr, (VERTEX *)(edge->wing[0].vertex->client_data));
    fprintf(stderr, "\n");
  } else
    fprintf(out, ", wing ???\n");

  if (wing->contour) {
    fprintf(out, "\tbelongs to ");
    PrintContourData(out, (CONTOUR *)(wing->contour->client_data));
    fprintf(out, "\n");
  } else
    fprintf(out, "\tdoesn't belong to a contour\n");
}

void PrintWingShort(FILE *out, BREP_WING *wing)
{
  if (!wing)
    fprintf(out, "(null)");
  else
    fprintf(out, "edge %d wing %s contour %d, vertex %d -> %d, ", 
	  ((EDGE *)(wing->edge->client_data))->id & ~IDMASK,
	  ((wing == &(wing->edge->wing[0])) ? "0" :
	   (wing == &(wing->edge->wing[1])) ? "1" : "ERROR"),
	  (wing->contour ? 
	  ((CONTOUR *)(wing->contour->client_data))->id & ~IDMASK :
	  -1),
	  ((VERTEX *)(wing->vertex->client_data))->id & ~IDMASK,
	  (wing->next ?
	   ((VERTEX *)(wing->next->vertex->client_data))->id & ~IDMASK :
	   -1));
}

static int TotalEdges = 0, edgecount = 0;

int GetNumberOfEdges(void)
{
  return TotalEdges;
}

EDGE *EdgeCreateCallback(BREP_EDGE *edge)
{
  EDGE *e, *edge_data = edge->client_data;

  e = NEWEDGE();
  if (edge_data)
    *e = *edge_data;
  e->id = ((edgecount++) & ~IDMASK) | EDGEMASK;

#ifdef DEBUG
  fprintf(stderr, "EdgeCreateCallback:\n");
#endif
  TotalEdges++;
  return e;
}

EDGE *EdgeCloseCallback(BREP_EDGE *edge)
{
  EDGE *edge_data = (EDGE *)(edge->client_data);
  float e;

#ifdef DEBUG 
  fprintf(stderr, "EdgeCloseCallback:\n");
#endif 

  /* compute tolerance, extent and near as in M. Segal, SIGGRAPH '90 p 105 */
  edge_data->tolerance = edge_tolerance;
  edge_data->extent = ((VERTEX *)(edge->wing[0].vertex->client_data))->tolerance;
  e = ((VERTEX *)(edge->wing[1].vertex->client_data))->tolerance;
  if (e > edge_data->extent)
    edge_data->extent = e;
  edge_data->near = HUGE;

#ifdef DEBUG 
  PrintEdge(stderr, edge);
#endif 

  return edge_data;
}

void EdgeDestroyCallback(BREP_EDGE *edge)
{
  EDGE *edge_data = edge->client_data;

#ifdef DEBUG
  fprintf(stderr, "EdgeDestroyCallback:\n");
#endif
  if (edge_data) {
#ifdef DEBUG
    fprintf(stderr, "\t");
    PrintEdgeData(stderr, (EDGE *)(edge_data));
    fprintf(stderr, "\n");
#endif
    DISPOSEEDGE(edge_data);
  }

  TotalEdges--;
}


/* ********************** Contour functions ****************************** */
#ifndef NOPOOLS
static POOL *contourPool = (POOL *)NULL;
#define NEWCONTOUR()  	(CONTOUR *)New(sizeof(CONTOUR), &contourPool)
#define DISPOSECONTOUR(ptr) Dispose((unsigned char *)(ptr), &contourPool)
#else /*NOPOOLS*/
#define NEWCONTOUR()	(CONTOUR *)Alloc(sizeof(CONTOUR))
#define DISPOSECONTOUR(ptr) Free((char *)ptr, sizeof(CONTOUR))
#endif /*NOPOOLS*/
static int TotalContours = 0, contourcount = 0;

int IsHole(BREP_CONTOUR *contour)
{
  CONTOUR *c = (CONTOUR *)(contour->client_data);
  FACE *f = (FACE *)(contour->face->client_data);
  return (VECTORDOTPRODUCT(f->normal, c->normal) <= 0.);
}

void PrintContourData(FILE *out, CONTOUR *contour_data)
{
  if (contour_data) {
    fprintf(out, "contour id %d normal = (%f,%f,%f), ", 
	    contour_data->id & ~IDMASK,
	    contour_data->normal.x, 
	    contour_data->normal.y,
	    contour_data->normal.z); 
    fprintf(out, "%shidden, %sprocessed.\n", 
	    contour_data->id&HIDDEN ? "" : "not ",
	    contour_data->id&PROCESSED ? "" : "not ");
  } else {
    fprintf(out, "(no contour data)");
  }
}

int GetNumberOfContours(void)
{
  return TotalContours;
}

CONTOUR *ContourCreateCallback(BREP_CONTOUR *contour)
{
  CONTOUR *c, *contour_data = contour->client_data;

  c = NEWCONTOUR();
  if (contour_data)
    *c = *contour_data;
  c->id = ((contourcount++)&~IDMASK) | CONTOURMASK;

#ifdef DEBUG
  fprintf(stderr, "ContourCreateCallback:\n");
  PrintContourData(stderr, contour_data);
  fprintf(stderr, "\n");
#endif
  TotalContours++;
  return c;
}

/* Newells method for computing the normal on the plane defined by the contour.
 * Ref: F. Tampieri, Graphics Gems III p 231 */
void ContourNormal(BREP_CONTOUR *contour, VECTOR *norm)
{
  BREP_WING *wing;
  DVECTOR *cur, *next;

  VECTORSET(*norm, 0. ,0., 0.);

  next = &(((VERTEX *)(contour->wings->vertex->client_data))->point);
  for (wing = contour->wings; wing; wing = (wing->next == contour->wings ? NULL : wing->next)) {
    cur = next;
    next = &(((VERTEX *)(wing->next->vertex->client_data))->point);

    norm->x += (cur->y - next->y) * (cur->z + next->z);
    norm->y += (cur->z - next->z) * (cur->x + next->x);
    norm->z += (cur->x - next->x) * (cur->y + next->y);
  }    

  VECTORNORMALIZE(*norm);
}

CONTOUR *ContourCloseCallback(BREP_CONTOUR *contour)
{
#ifdef DEBUG
  fprintf(stderr, "ContourCloseCallback:\n");
  PrintContourData(stderr, (CONTOUR *)(contour->client_data));
  fprintf(stderr, "\n");
  CheckContour(contour);
#endif
  ContourNormal(contour, &(((CONTOUR *)(contour->client_data))->normal));
  return contour->client_data;
}

void ContourDestroyCallback(BREP_CONTOUR *contour)
{
  CONTOUR *contour_data = contour->client_data;
  if (contour_data)
    DISPOSECONTOUR(contour_data);
  TotalContours--;
}


/* ************************** Face functions ***************************** */
#ifndef NOPOOLS
static POOL *facePool = (POOL *)NULL;
#define NEWFACE()  	(FACE *)New(sizeof(FACE), &facePool)
#define DISPOSEFACE(ptr) Dispose((unsigned char *)(ptr), &facePool)
#else /*NOPOOLS*/
#define NEWFACE()	(FACE *)Alloc(sizeof(FACE))
#define DISPOSEFACE(ptr) Free((char *)ptr, sizeof(FACE))
#endif /*NOPOOLS*/

void PrintFaceData(FILE *out, FACE *face_data)
{
  if (face_data) {
    fprintf(out, "face id %d, ", face_data->id & ~IDMASK);
    fprintf(out, "normal = (%g, %g, %g), plane constant = %g, tolerance = %g, extent = %g, near = %g, ", 
	    face_data->normal.x, face_data->normal.y, face_data->normal.z,
	    face_data->d,
	    face_data->tolerance,
	    face_data->extent,
	    face_data->near);
    fprintf(out, "%shidden, %sprocessed.\n", 
	    face_data->id&HIDDEN ? "" : "not ",
	    face_data->id&PROCESSED ? "" : "not ");
  } else {
    fprintf(out, "(no face data)");
  }
}

static int TotalFaces = 0, facecount = 0;

int GetNumberOfFaces(void)
{
  return TotalFaces;
}

FACE *FaceCreateCallback(BREP_FACE *face)
{
  FACE *f, *face_data = face->client_data;

  f = NEWFACE();
  if (face_data)
    *f = *face_data;
  f->id = ((facecount++)&~IDMASK) | FACEMASK;

  TotalFaces++;
  return f;
}

/* for computing the plane equation of a face with Newells method, including
 * face tolerance, see
 * F. Tampieri, Graphics Gems III p 231 
 * M. Segal, IEEE CG&A Jan. 1988 p 53 */
static FACE *face_data;
static double dmin, dmax, emax;

static void do_close_face_vertex(BREP_VERTEX *vertex)
{
  VERTEX *vertex_data = (VERTEX *)(vertex->client_data);
  double d;
  POINT p;
  
  d = -VECTORDOTPRODUCT(face_data->normal, vertex_data->point);
  if (d < dmin) dmin = d;
  if (d > dmax) dmax = d;

  if (vertex_data->tolerance > emax)
    emax = vertex_data->tolerance;

  VECTORCOPY(vertex_data->point, p);
  BoundsEnlargePoint(face_data->bounds, &p);
}

/* computes the plane equation of the face using Newells method */
void ComputePlaneEquation(BREP_FACE *face)
{
  face_data = (FACE *)(face->client_data);
  
  face_data->normal = ((CONTOUR *)(face->outer_contour->client_data))->normal;
  BoundsInit(face_data->bounds);

  emax = 0.;		/* maximum vertex tolerance */
  dmin = HUGE;		/* minimum plane constant */
  dmax = -HUGE;		/* maximum plane constant */
  BrepFaceIterateContours1A(face, (void (*)(BREP_CONTOUR *, void *))BrepContourIterateVertices, (void *)do_close_face_vertex);
  face_data->d = (dmin + dmax)/2.;
  face_data->tolerance = MAX((dmax - dmin)/2., face_tolerance);
  face_data->extent = face_data->tolerance + emax;
  face_data->near = HUGE;

  face_data->bounds[MIN_X] -= face_data->extent;
  face_data->bounds[MIN_Y] -= face_data->extent;
  face_data->bounds[MIN_Z] -= face_data->extent;
  face_data->bounds[MAX_X] += face_data->extent;
  face_data->bounds[MAX_Y] += face_data->extent;
  face_data->bounds[MAX_Z] += face_data->extent;
}

FACE *FaceCloseCallback(BREP_FACE *face)
{
#ifdef DEBUG 
  fprintf(stderr, "FaceCloseCallback:\n");
#endif 
  /* compute the plane equation of the face with Newell's method */
  ComputePlaneEquation(face);

#ifdef DEBUG 
  PrintFaceData(stderr, (FACE *)(face->client_data));
  fprintf(stderr, "\n");
#endif 

  return face->client_data;
}

void FaceDestroyCallback(BREP_FACE *face)
{
  FACE *face_data = face->client_data;
  if (face_data)
    DISPOSEFACE(face_data);
  TotalFaces--;
}

static float minlen;

static void TestDecreaseEdgeLength(BREP_WING *wing)
{
  BREP_VERTEX *v1 = wing->vertex, *v2 = wing->next->vertex;
  DVECTOR p1 = ((VERTEX *)(v1->client_data))->point,
          p2 = ((VERTEX *)(v2->client_data))->point;
  double len;
  VECTORDIST(p1, p2, len);
  if (len < minlen) minlen = len;
}

float FaceShortestEdgeLength(BREP_FACE *face)
{
  minlen = HUGE;
  BrepContourIterateWings(face->outer_contour, TestDecreaseEdgeLength);
  return minlen;
}

float minedgelength;
void SetMinimumEdgeLength(float length)
{
  minedgelength = length;
}

/* ********************* Shell functions ********************************* */
#ifndef NOPOOLS
static POOL *shellPool = (POOL *)NULL;
#define NEWSHELL()  	(SHELL *)New(sizeof(SHELL), &shellPool)
#define DISPOSESHELL(ptr) Dispose((unsigned char *)(ptr), &shellPool)
#else /*NOPOOLS*/
#define NEWSHELL()	(SHELL *)Alloc(sizeof(SHELL))
#define DISPOSESHELL(ptr) Free((char *)ptr, sizeof(SHELL))
#endif /*NOPOOLS*/

void PrintShellData(FILE *out, SHELL *shell_data)
{
  if (shell_data) {
    fprintf(out, "shell id %d, ", shell_data->id & ~IDMASK);
    fprintf(out, "name = '%s', ", shell_data->name ? shell_data->name : "(null)");
    fprintf(out, "material = '%s', ", shell_data->matname ? shell_data->matname : "(null)");
    fprintf(out, "%shidden, %sprocessed.\n", 
	    shell_data->id&HIDDEN ? "" : "not ",
	    shell_data->id&PROCESSED ? "" : "not ");
  } else {
    fprintf(out, "(no shell data)");
  }
}

static int TotalShells = 0;

int GetNumberOfShells(void)
{
  return TotalShells;
}

SHELL *ShellCreateCallback(BREP_SHELL *shell)
{
  SHELL *shell_data= shell->client_data;
  SHELL *s;

  s = NEWSHELL();
  if (shell_data) {
    *s = *shell_data;
    if (shell_data->name) {
      s->name = Alloc(strlen(shell_data->name)+1);
      strcpy(s->name, shell_data->name);
    }
    if (shell_data->matname) {
      s->matname = Alloc(strlen(shell_data->matname)+1);
      strcpy(s->matname, shell_data->matname);
    }
  }
  s->id |= SHELLMASK;

  TotalShells++;
  return s;
}

static SHELL *shell_data;

static void do_close_shell_face(BREP_FACE *face)
{
  FACE *face_data = (FACE *)(face->client_data);
  BoundsEnlarge(shell_data->bounds, face_data->bounds);
}

SHELL *ShellCloseCallback(BREP_SHELL *shell)
{
#ifdef DEBUG
  fprintf(stderr, "ShellCloseCallback:\n");
  PrintShellData(stderr, (SHELL *)(shell->client_data));
  fprintf(stderr, "\n");
#endif
  shell_data = (SHELL *)shell->client_data;
  BoundsInit(shell_data->bounds);
  BrepShellIterateFaces(shell, do_close_shell_face);

  return shell->client_data;
}

void ShellDestroyCallback(BREP_SHELL *shell)
{
  SHELL *shell_data = shell->client_data;
  if (shell_data) {
    if (shell_data->name)
      Free(shell_data->name, strlen(shell_data->name)+1);
    if (shell_data->matname)
      Free(shell_data->matname, strlen(shell_data->matname)+1);
    DISPOSESHELL(shell_data);
  }
  TotalShells--;
}


/* *********************** Solid functions *************************** */
#ifndef NOPOOLS
static POOL *solidPool = (POOL *)NULL;
#define NEWSOLID()  	(SOLID *)New(sizeof(SOLID), &solidPool)
#define DISPOSESOLID(ptr) Dispose((unsigned char *)(ptr), &solidPool)
#else /*NOPOOLS*/
#define NEWSOLID()	(SOLID *)Alloc(sizeof(SOLID))
#define DISPOSESOLID(ptr) Free((char *)ptr, sizeof(SOLID))
#endif /*NOPOOLS*/

void PrintSolidData(FILE *out, SOLID *solid_data)
{
  if (solid_data) {
    fprintf(out, "solid id %d, ", solid_data->id & ~IDMASK);
    fprintf(out, "filename = '%s'", solid_data->filename ? solid_data->filename : "(null)");
    fprintf(out, "bounds: "); BoundsPrint(out, solid_data->bounds);
    fprintf(out, "%shidden, %sprocessed.\n", 
	    solid_data->id&HIDDEN ? "" : "not ",
	    solid_data->id&PROCESSED ? "" : "not ");
  } else {
    fprintf(out, "(no solid data)");
  }
}

static int TotalSolids = 0;

int GetNumberOfSolids(void)
{
  return TotalSolids;
}

SOLID *SolidCreateCallback(BREP_SOLID *solid)
{
  SOLID *solid_data = solid->client_data;
  SOLID *s;

  s = NEWSOLID();
  if (solid_data) {
    *s = *solid_data;
    if (solid_data->filename) {
      s->filename = Alloc(strlen(solid_data->filename)+1);
      strcpy(s->filename, solid_data->filename);
    }
  }
  s->id |= SOLIDMASK;

  TotalSolids++;
  return s;
}

static SOLID *solid_data;

static void do_close_solid_shell(BREP_SHELL *shell)
{
  SHELL *shell_data = (SHELL *)(shell->client_data);
  BoundsEnlarge(solid_data->bounds, shell_data->bounds);
}

SOLID *SolidCloseCallback(BREP_SOLID *solid)
{
#ifdef DEBUG
  fprintf(stderr, "SolidCloseCallback:\n");
  PrintSolidData(stderr, (SOLID *)(solid->client_data));
  fprintf(stderr, "\n");
#endif
  solid_data = (SOLID *)solid->client_data;
  BoundsInit(solid_data->bounds);
  BrepSolidIterateShells(solid, do_close_solid_shell);

  return solid->client_data;
}

void SolidDestroyCallback(BREP_SOLID *solid)
{
  SOLID *solid_data = solid->client_data;
  if (solid_data) {
    if (solid_data->filename)
      Free(solid_data->filename, strlen(solid_data->filename)+1);
    DISPOSESOLID(solid_data);
  }
  TotalSolids--;
}


/* ****************** entity type identification etc... ****************** */
unsigned int EntityType(ENTITY *data)
{
  unsigned int typemask;

  for (typemask=0x80000000; typemask&IDMASK; typemask >>= 1) {
    if (data->id & typemask)
      return typemask;
  }
  return 0;
}

char *EntityTypeName(ENTITY *data)
{
  switch (EntityType(data)) {
  case VERTEXMASK:	return "vertex";
  case EDGEMASK:	return "edge";
  case CONTOURMASK:	return "contour";
  case FACEMASK:	return "face";
  case SHELLMASK:	return "shell";
  case SOLIDMASK:	return "solid";
  default:		return "unknown entity type";
  }
}

void PrintEntityData(FILE *out, ENTITY *data)
{
  if (mg_file) {
    fprintf(out, "MGF file '%s' line '%d':\n", 
	    mg_file->fname ? mg_file->fname : "(null)",
	    mg_file->lineno);
  }

  if (!data) {
    fprintf(out, "(no entity data)");
    return;
  }

  switch (EntityType(data)) {
  case VERTEXMASK:      PrintVertexData(out, (VERTEX *)data); break;
  case EDGEMASK:	PrintEdgeData(out, (EDGE *)data); break;
  case CONTOURMASK:	PrintContourData(out, (CONTOUR *)data); break;
  case FACEMASK:	PrintFaceData(out, (FACE *)data); break;
  case SHELLMASK:	PrintShellData(out, (SHELL *)data); break;
  case SOLIDMASK:	PrintSolidData(out, (SOLID *)data); break;
  default:		    
    fprintf(out, "(unknown entity type, id = 0x%08x nr = %d)", 
	    data->id, data->id & ~IDMASK);
  }

  return;
}


/* ****************** brep library error reporting *********************** */
/* prints an info message */
static void do_info(ENTITY *client_data, char *message)
{
#ifdef DEBUG
  PrintEntityData(stderr, client_data);
  fprintf(stderr, "\n%s.\n", message);
#endif
}

/* prints a warning message */
static void do_warning(ENTITY *client_data, char *message)
{
  Warning(NULL, "%s id %d: %s", 
	  EntityTypeName(client_data), 
	  EntityId(client_data), 
	  message);
}

/* pritns an error message */
static void do_error(ENTITY *client_data, char *message)
{
  Error(NULL, "%s id %d: %s", 
	EntityTypeName(client_data), 
	EntityId(client_data), 
	message);
}

/* prints a fatal error message and exits the program */
static void do_fatal(ENTITY *client_data, char *message)
{
  Fatal(-1, NULL, "%s id %d: %s", 
	EntityTypeName(client_data), 
	EntityId(client_data), 
	message);
}


/* ******************* various initialisations ********************** */
void InitGeometry(void)
{
  TotalVertices = TotalEdges = TotalContours = TotalFaces = TotalShells = TotalSolids = 0;

  /* Default vertex compare flags: vertex location, normal and name are
   * all taken into accoutn when comparing vertices. If two vertices to be
   * compared have the same location, and the same normal,but both of them 
   * have no name (e.g. automatically created vertices by the MGF parser
   * library, when discretising spheres ...) the vertices are to be
   * considered equal. */
  SetVertexCompareFlags(VCMP_LOCATION | VCMP_NORMAL | VCMP_NAME |
			VCMP_NO_NAME_IS_EQUAL_NAME);

  /* --- mandatory --- */
  /* specify the routine two compare the geometry data of two vertices */
  BrepSetVertexCompareRoutine((BREP_COMPARE_FUNC)VertexCompare);
  BrepSetVertexCompareLocationRoutine((BREP_COMPARE_FUNC)VertexCompareLocation);

  /* --- optional --- */
  /* specify routines for printing warning, error and fatal error messages */ 
  BrepSetInfoCallback((BREP_MSG_CALLBACK_FUNC)do_info);
  BrepSetWarningCallback((BREP_MSG_CALLBACK_FUNC)do_warning);
  BrepSetErrorCallback((BREP_MSG_CALLBACK_FUNC)do_error);
  BrepSetFatalCallback((BREP_MSG_CALLBACK_FUNC)do_fatal);

  /* vertex callback functions */
  BrepSetCreateVertexCallback((BREP_CALLBACK_FUNC)VertexCreateCallback);
  BrepSetCloseVertexCallback((BREP_CALLBACK_FUNC)VertexCloseCallback);
  BrepSetDestroyVertexCallback((BREP_CALLBACK_FUNC)VertexDestroyCallback);

  /* edge callback functions */
  BrepSetCreateEdgeCallback((BREP_CALLBACK_FUNC)EdgeCreateCallback);
  BrepSetCloseEdgeCallback((BREP_CALLBACK_FUNC)EdgeCloseCallback);
  BrepSetDestroyEdgeCallback((BREP_CALLBACK_FUNC)EdgeDestroyCallback);

  /* contour callback functions */
  BrepSetCreateContourCallback((BREP_CALLBACK_FUNC)ContourCreateCallback);
  BrepSetCloseContourCallback((BREP_CALLBACK_FUNC)ContourCloseCallback);
  BrepSetDestroyContourCallback((BREP_CALLBACK_FUNC)ContourDestroyCallback);

  /* face callback functions */
  BrepSetCreateFaceCallback((BREP_CALLBACK_FUNC)FaceCreateCallback);
  BrepSetCloseFaceCallback((BREP_CALLBACK_FUNC)FaceCloseCallback);
  BrepSetDestroyFaceCallback((BREP_CALLBACK_FUNC)FaceDestroyCallback);

  /* shell callback functions */
  BrepSetCreateShellCallback((BREP_CALLBACK_FUNC)ShellCreateCallback);
  BrepSetCloseShellCallback((BREP_CALLBACK_FUNC)ShellCloseCallback);
  BrepSetDestroyShellCallback((BREP_CALLBACK_FUNC)ShellDestroyCallback);

  /* solid callback functions */
  BrepSetCreateSolidCallback((BREP_CALLBACK_FUNC)SolidCreateCallback);
  BrepSetCloseSolidCallback((BREP_CALLBACK_FUNC)SolidCloseCallback);
  BrepSetDestroySolidCallback((BREP_CALLBACK_FUNC)SolidDestroyCallback);
}


