/* geometry.h: geometric data for vertices, edges, ... */
/*
  The BREP Library.
  Copyright (C) 1996 Philippe Bekaert

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef _GEOMETRY_H_
#define _GEOMETRY_H_

#include <stdio.h>
#include "vectortype.h"
#include "brep.h"
#include "mgf.h"	/* MGF parser header file */
#include "bounds.h"

/* see Mark Segal, "Using Tolerances to Guarantee Valid Polyhedral Modeling 
 * Results", SIGGRAPH '90 p 105 for everything about tolerances, extents and
 * near values and what they are used for. */

/* The first field of the following structs is always an unsigned int used for 
 * identification and other purposes. An unsigned int should contain at least
 * 4 bytes (32 bits). */

/* Geometric data for a vertex */
typedef struct VERTEX {
  unsigned int id;	/* identification for debugging etc... */
  DVECTOR point;	/* coordinates in 3D space: double floating point! */
  VECTOR normal;	/* normal vector to the surface at the vertex */
  float tolerance;	/* coordinate tolerance. Two vertices nearer than
			* this should be considered one vertex. */
  char *name;		/* name for the vertex */
} VERTEX;

/* compares two vertices and returns a code usable for ordering vertices
 * into an octree. */
extern int VertexCompare(VERTEX *v1, VERTEX *v2);
extern int VertexCompareLocation(VERTEX *v1, VERTEX *v2);

extern void PrintVertexData(FILE *out, VERTEX *vertex_data);
extern void PrintVertex(FILE *out, BREP_VERTEX *vertex);
extern int GetNumberOfVertices(void);
extern VERTEX *VertexCreateCallback(BREP_VERTEX *vertex);
extern VERTEX *VertexCloseCallback(BREP_VERTEX *vertex);
extern void VertexDestroyCallback(BREP_VERTEX *vertex);

/* Computes a normals for the vertex if it doesn't have a normal yet
 * unless normals in the input should be ignored. */
extern void ComputeVertexNormal(BREP_VERTEX *vertex);

/* Geometric data for an edge */
typedef struct EDGE {
  unsigned int id;
  float tolerance, extent, near;
} EDGE;

extern void PrintEdgeData(FILE *out, EDGE *edge_data);
extern void PrintEdge(FILE *out, BREP_EDGE *edge);
extern void PrintWing(FILE *out, BREP_WING *wing);
extern void PrintWingShort(FILE *out, BREP_WING *wing);
extern int GetNumberOfEdges(void);
extern EDGE *EdgeCreateCallback(BREP_EDGE *edge);
extern EDGE *EdgeCloseCallback(BREP_EDGE *edge);
extern void EdgeDestroyCallback(BREP_EDGE *edge);


/* Geometric data for a contour */
typedef struct CONTOUR {
  unsigned int id;
  VECTOR normal;
} CONTOUR;

extern void ContourNormal(BREP_CONTOUR *contour, VECTOR *norm);
extern int IsHole(BREP_CONTOUR *contour);

extern void PrintContourData(FILE *out, CONTOUR *contour_data);
extern int GetNumberOfContours(void);
extern CONTOUR *ContourCreateCallback(BREP_CONTOUR *contour);
extern CONTOUR *ContourCloseCallback(BREP_CONTOUR *contour);
extern void ContourDestroyCallback(BREP_CONTOUR *contour);


/* Geometric data for a face */
typedef struct FACE {
  unsigned int id;
  float tolerance, extent, near;
  BOUNDINGBOX bounds;
  VECTOR normal;
  float d;	/* plane constant */
} FACE;

#define DisjunctFaceBounds(faceA, faceB) \
     DisjunctBounds(((FACE *)(faceA->client_data))->bounds, \
		    ((FACE *)(faceB->client_data))->bounds)

extern void PrintFaceData(FILE *out, FACE *face_data);
extern int GetNumberOfFaces(void);
extern FACE *FaceCreateCallback(BREP_FACE *face);
extern FACE *FaceCloseCallback(BREP_FACE *face);
extern void FaceDestroyCallback(BREP_FACE *face);


/* Geometric data for a shell */
#include "color.h"

typedef struct RGB_MATERIAL {
  RGB diffuse, emission, specular;
  float shininess, alpha;
} RGB_MATERIAL;

extern int SameRGBMaterial(RGB_MATERIAL *mat1, RGB_MATERIAL *mat2);

typedef struct SHELL {
  unsigned int id;
  char *name;
  char *matname; 
  C_MATERIAL material;
  RGB_MATERIAL rgbmat;
  BOUNDINGBOX bounds;
} SHELL;

#define DisjunctShellBounds(shellA, shellB) \
     DisjunctBounds(((SHELL *)(shellA->client_data))->bounds, \
		    ((SHELL *)(shellB->client_data))->bounds)

extern void PrintShellData(FILE *out, SHELL *shell_data);
extern int GetNumberOfShells(void);
extern SHELL *ShellCreateCallback(BREP_SHELL *shell);
extern SHELL *ShellCloseCallback(BREP_SHELL *shell);
extern void ShellDestroyCallback(BREP_SHELL *shell);


/* Geometric data for a solid */
typedef struct SOLID {
  unsigned int id;
  char *filename;
  BOUNDINGBOX bounds;
} SOLID;

#define DisjunctSolidBounds(solidA, solidB) \
     DisjunctBounds(((SOLID *)(solidA->client_data))->bounds, \
		    ((SOLID *)(solidB->client_data))->bounds)

extern void PrintSolidData(FILE *out, SOLID *solid_data);
extern int GetNumberOfSolids(void);
extern SOLID *SolidCreateCallback(BREP_SOLID *solid);
extern SOLID *SolidCloseCallback(BREP_SOLID *solid);
extern void SolidDestroyCallback(BREP_SOLID *solid);


/* ************************** type- and other bits ****************************** */
/* All entities have as a first field of data an unsigned integer used
 * for identification and to hold flags for various other purposes, such as to
 * indicate whether a contour lies on the inside or outside of another object
 * after partitioning for e.g. CSG operations (see "various other masks" below). */

/* The most significant set bit indicates the type of an entity. The type bits
 * are defined below. */
#define CONTOURMASK 	0x80000000
#define VERTEXMASK 	0x40000000	
#define EDGEMASK 	0x20000000
#define FACEMASK 	0x10000000
#define SHELLMASK 	0x08000000
#define SOLIDMASK 	0x04000000

/* Various masks for other purposes than identification of an entity: These masks
 * set only bits that are less significant than the type bit of an entity. */

/* masks to indicate whether a contour lies inside or outside another object after
 * partitioning with PartitionFaces() in isect.c + for indicating whether the vertex
 * data for a vertex have been written out already before in writemgf.c */
#define OUTMASK 	0x04000000
#define INMASK  	0x08000000

/* masks indicating whether or not a contour needs to be written with the
 * routines in writemgf.c */
#define DONTWRITEMASK	0x10000000
#define WRITEMASK	0x20000000
#define WRITEREVERSEMASK 0x40000000

/* mask indicating that an entity is hidden */
#define HIDDEN 		0x02000000

/* mask indicating that an entity has been processed */
#define PROCESSED	0x01000000

/* The IDMASK below is used to separate the type bit and other flags from
 * the id-number of an entity. In order to provide as much as possible
 * space for id-numbers of entities, entities for which most other flags
 * are needed (contours) have the highest type bit. Doing so, about 16 Mio 
 * (2^24) topological entities of the same type are possible. */
#define IDMASK 		0xff000000

/* struct with common first field of VERTEX, EDGE, CONTOUR, ... */
typedef struct ENTITY {
  unsigned int id;
} ENTITY;

#define MARK_UNHIDDEN(entity) ((ENTITY *)(entity->client_data))->id &= ~HIDDEN
#define MARK_HIDDEN(entity) ((ENTITY *)(entity->client_data))->id |= HIDDEN
#define IS_HIDDEN(entity) (((ENTITY *)(entity->client_data))->id & HIDDEN)

#define MARK_UNPROCESSED(entity) ((ENTITY *)(entity->client_data))->id &= ~PROCESSED
#define MARK_PROCESSED(entity) ((ENTITY *)(entity->client_data))->id |= PROCESSED
#define IS_PROCESSED(entity) (((ENTITY *)(entity->client_data))->id & PROCESSED)

#define TOGGLE_WRITEREVERSE(contour) ((CONTOUR *)(contour->client_data))->id ^= WRITEREVERSEMASK
#define IS_REVERSED(contour) ((CONTOUR *)(contour->client_data))->id & WRITEREVERSEMASK

/* Determines the entity type of an entity. The type mask is returned. */
extern unsigned int EntityType(ENTITY *data);

/* returns the entity type name of an entity */
extern char *EntityTypeName(ENTITY *data);

/* "returns" the id number of an entity. */
#define EntityId(entity) (((ENTITY *)(entity))->id & ~IDMASK)

/* prints all geometric and type information about the entity to the file
 * pointed to by out */
extern void PrintEntityData(FILE *out, ENTITY *data);


/* ******************* Vertex comparison flags ***************************** */
/*
 * The following data may be taken into consideration when comparing vertices:
 * - vertex location: normally always specified.
 * - vertex normal: (0,0,0) if not specified, normalized vector if specified.
 * - vertex name: NULL if not specified.
 * What data should be taken into account when comparing vertices is 
 * controlled with the vertex compare flags VCMP_LOCATION, 
 * VCMP_NORMAL and VCMP_NAME. 
 *
 * The order in which the data is taken into account by VertexCompare is
 * as follows: first location is tested, than normals, than name (if relevant).
 *
 * When the normal or name is relevant, it also depends on the situation
 * whether or not two vertices without normal (or name) and equal other data
 * are to be considered equal or not. This is controlled with the flags 
 * VCMP_NO_NORMAL_IS_EQUAL_NORMAL and VCMP_NO_NAME_IS_EQUAL_NAME.
 * When the normal (or name) is relevant, and only one of the two
 * vertices has a normal (or name), the vertices are considered different.
 *
 * The flags can be |-ed together.
 *
 * Default vertex compare flags: vertex location, normal and name are
 * all taken into account when comparing vertices. If two vertices to be
 * compared have the same location, and the same normal,but both of them 
 * have no name (e.g. automatically created vertices by the MGF parser
 * library, when discretising spheres ...) the vertices are to be
 * considered equal. */

#define VCMP_LOCATION			1
#define VCMP_NORMAL			2
#define VCMP_NAME			4
#define VCMP_NO_NORMAL_IS_EQUAL_NORMAL	8
#define VCMP_NO_NAME_IS_EQUAL_NAME     16

extern void SetVertexCompareFlags(unsigned int flags);


/* ******************* various initialisations ********************** */
extern void InitGeometry(void);

extern void SetMinimumVertexTolerance(double tol);
extern void SetMinimumEdgeTolerance(double tol);
extern void SetMinimumFaceTolerance(double tol);

extern void SetMinimumEdgeLength(float length);

extern float FaceShortestEdgeLength(BREP_FACE *face);

#endif /*_GEOMETRY_H_*/

