/*
  The BREP Library.
  Copyright (C) 1996 Philippe Bekaert

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* #define DDEBUG  */
#undef DEBUG 
/* #define CHECK  */

/* isect.c: partitions pairs of faces into nonintersecting parts
 *
 * References: 
 *
 * - Marc Segal and Carlo H. Sequin, "Partitioning Polyhedral Objects into 
 *      Nonintersecting Parts", IEEE Computer graphics and Applications, 
 *	Januari 1988 p 53
 * - Marc Segal, "Using Tolerances to Guarantee Valid Polyhedral Modeling 
 *	Results", SIGGRAPH '90, p 105 
 *
 *
 * Known bugs:
 *
 * - endpoint determination may fail for small feautres, e.g. edges 
 * 	with length of about the same order of magnitude as the vertex 
 *	tolerances.
 *
 * In case of bugs: recompile this file with the symbol CHECK defined.
 * Face pairs causing warnings or errors are dumped to an MGF file, which
 * most of the time will be usable to investigate the problem with 
 * another instance of the program compiled after commenting out
 * the #undef DEBUG line above.
 *
 *
 * Still to do: 
 *
 * - feature merging (Segal, SIGGRAPH '90)
 *
 *
 * Author: 
 *
 *	Philippe Bekaert, Computer Graphics Research Group K.U.Leuven
 *	Philippe.Bekaert@cs.kuleuven.ac.be
 *	Aug, 29 1996
 *
 *
 * Changelog:
 *
 * Sun Sep 22 23:23:37 1996  Philippe Bekaert  <philippe@strojka.cs.kuleuven.ac.be>
 *
 *      * non-coplanar faces containing contours that lie entirely in the
 *	transversal face may cause problems. Solved??
 *
 * Wed Oct  9 18:10:11 1996  Philippe Bekaert  <philippe@flater.cs.kuleuven.ac.be>
 *
 *	* inside/outside marking of contours corrected.
 *
 */

#include <stdlib.h>

#include "isect.h"
#include "geometry.h"
#include "pools.h"
#include "error.h"
#include "vector.h"

#ifdef DEBUG
#include <stdio.h>
#include "check.h"
#endif

#define UNKNOWN_SIDE	10	/* arbitrary, but not -1, 0 or 1 */

/* information about intersection points of two faces: */

/* endpoint types: to be combined with | */
#define VERTEX_ENDPOINT		1	/* no faces need to be cut, only for merging
					 * with very nearby intersection points. */
#define PHANTOM_ENDPOINT	2	/* only the "other" face needs to be cut */
#define EXIT_ENDPOINT		4	/* both faces need to be partitioned */
#define ENTRY_ENDPOINT		8	/* "" */
#define DELETED_ENDPOINT       16	/* deleted endpoints after merging */

typedef struct ENDPOINT {
  double ti;		/* parameter of endpoint on intersection line */
  double te;		/* parmeter of endpoint on contributing edge */
  double tol;		/* intersection point coordinate tolerance */
  int	type;		/* ENTRY_ENDPOINT or EXIT_ENDPOINT or ... */
  BREP_WING *wing;	/* wing containing the intersection point */
  int source;		/* for debugging: source code line identifying the
			 * situation how the endpoints was contributed */
} ENDPOINT;

/* the endpoint list is automatically made large enough */
static ENDPOINT *endpoints = (ENDPOINT *)NULL;
static int nr_endpoints, max_nr_endpoints=0;

/* intersection line data */
typedef struct ISECT_LINE {
  DVECTOR U, V;		/* parametric line equation is t.U + V, U normalized. */
   double tol;		/* tolerance */
} ISECT_LINE;

static ISECT_LINE isect;

/* global parameters needed during partioning */
static BREP_FACE *faceA, *faceB;		/* pair of intersecting faces 
						 * being processed. */
static BREP_WING *lastwingA, *lastwingB;	/* wings leaving at the exit
						 * endpoint of the previous
						 * cutting interval. */
static double lastti;				/* exit endpoint parameter on the
						 * intersecting line of the previous
						 * cutting interval */
static double timinA, timaxA, timinB, timaxB;	/* minimum and maximum
						 * parameter along intersection line
						 * of endpoints contributed by
						 * faceA and faceB */

/* see Segal, SIGGRAPH '90 p 111. The coplanarity factor indicates to what 
 * extent two faces are coplanar. Coplanar faces have a large coplanarity
 * factor. For perpendicular faces, the coplanarity factor is sqrt(2). IN
 * order to decide whether two faces are to be considered coplanar, their
 * coplanarity factor is compared with a minimum coplanarity factor, which
 * should be a number between 100 and 1000. Smaller values force the algorithm
 * to deem pairs of faces that intersect in a relatively large dehedral 
 * angle to be coplanar. */
#ifndef MIN_COPLANARITY_FACTOR
#define MIN_COPLANARITY_FACTOR 500
#endif
double min_coplanarity_factor = MIN_COPLANARITY_FACTOR;

/* To set the minimum allowed coplanarity factor. */
void SetMinCoplanarityFactor(double factor)
{
  min_coplanarity_factor = factor;
}

/* When CHECK is defined, pairs of faces that cause an error or warning 
 * are dumped in an MGF file, so the situation can be examined later on. */
#ifdef CHECK
#include <stdio.h>
#include <stdarg.h>
#include "writemgf.h"

static int check = FALSE;

static int dumpid = 0;
static FILE *dumpf = (FILE *)NULL;

static void OpenDump(void)
{
  char fname[100];

  sprintf(fname, "dump%d.mgf", ++dumpid);
  dumpf = fopen(fname, "w");
  if (!dumpf) {
    fprintf(stderr, "Can't open dumpfile %s for writing.\n", fname);
  }
}

static void CloseDump(void)
{
  fclose(dumpf);
  dumpf = (FILE *)NULL;
}

static void DumpFace(BREP_FACE *face)
{
  if (!dumpf)
    return;

  fprintf(stderr, "Dumping face %s ...", face == faceA ? "A" : "B");

  InitMGFOutput(face->shell->solid, dumpf);
  OutputMaterial(((SHELL *)(face->shell->client_data))->matname,
		 &(((SHELL *)(face->shell->client_data))->material));
  OutputFace(face);
  CloseMGFOutput(face->shell->solid);

  fprintf(stderr, "done.\n");
}

static void IsectWarning(char *routine, char *text, ...)
{
  va_list pvar;

  fprintf(stderr, "Warning");
  if (routine) fprintf(stderr, " (in subroutine %s)", routine);
  fprintf(stderr, ": ");

  va_start(pvar, text);
  vfprintf(stderr, text, pvar);
  va_end(pvar);

  fprintf(stderr, ".\n");

  if (!faceA && !faceB)
    return;

  OpenDump();

  fprintf(dumpf, "# Warning");
  if (routine) fprintf(dumpf, " (in subroutine %s)", routine);
  fprintf(dumpf, ": ");

  va_start(pvar, text);
  vfprintf(dumpf, text, pvar);
  va_end(pvar);

  fprintf(dumpf, ".\n");

  fprintf(dumpf, "# face A:\n");
  if (faceA) DumpFace(faceA);
  fprintf(dumpf, "# face B:\n");
  if (faceB) DumpFace(faceB);
  CloseDump();
}

#define Warning IsectWarning
#define Error IsectWarning

#endif /*CHECK*/

/* computes the coplanarity factor of two planes */
double CoplanarityFactor(FACE *f1, FACE *f2)
{
  DVECTOR normal;
  double s, c, tol;
  
  VECTORCROSSPRODUCT(f1->normal, f2->normal, normal);
  s = VECTORNORM(normal);
  if (s < EPSILON)
    return HUGE;

  c = 1 - s*s;
  if (c > 0.)
    c = sqrt(c);
  else 
    c = 0.;

  tol = sqrt(f1->extent * f1->extent + f2->extent * f2->extent +
		   2. * f1->extent * f2->extent * c) / s;

  return tol / MAX(f1->extent, f2->extent);
}

/* Computes the parametric equation P = V + tU of the 
 * intersection line of two non-parallel (face-)planes.
 * Ref: Priamos Georgiades, "Plane-to-Plane Intersection", 
 * Graphics Gems III p 233 */
static ISECT_LINE ComputePlaneToPlaneIntersection(FACE *f1, FACE *f2)
{
  ISECT_LINE isect;
  int index;
  double abs, maxabs, s, c;
  double coplaf;

  VECTORCROSSPRODUCT(f1->normal, f2->normal, isect.U);

  s = VECTORNORM(isect.U);
  c = 1 - s*s; c = (c > 0. ? sqrt(c) : 0.);
  isect.tol = sqrt(f1->extent * f1->extent + f2->extent * f2->extent +
		   2. * f1->extent * f2->extent * c) / s;
  coplaf = isect.tol / MAX(f1->extent, f2->extent);
  if (coplaf > min_coplanarity_factor)
    Warning("ComputePlaneToPlaneIntersection", "attempt to compute intersection line of parallel planes (coplanarity factor = %g, min. allowed = %g)", coplaf, min_coplanarity_factor);

  maxabs = fabs(isect.U.x); index = XNORMAL;
  if ((abs = fabs(isect.U.y)) > maxabs) {maxabs = abs; index = YNORMAL;}
  if ((abs = fabs(isect.U.z)) > maxabs) {maxabs = abs; index = ZNORMAL;}

  switch (index) {
  case XNORMAL:
    isect.V.x = 0.;
    isect.V.y = (f1->normal.z * f2->d - f2->normal.z * f1->d) / isect.U.x;
    isect.V.z = (f2->normal.y * f1->d - f1->normal.y * f2->d) / isect.U.x;
    break;
  case YNORMAL:
    isect.V.y = 0.;
    isect.V.z = (f1->normal.x * f2->d - f2->normal.x * f1->d) / isect.U.y;
    isect.V.x = (f2->normal.z * f1->d - f1->normal.z * f2->d) / isect.U.y;
    break;
  case ZNORMAL:
    isect.V.z = 0.;
    isect.V.x = (f1->normal.y * f2->d - f2->normal.y * f1->d) / isect.U.z;
    isect.V.y = (f2->normal.x * f1->d - f1->normal.x * f2->d) / isect.U.z;
    break;
  default:
    Fatal(2, "PlaneToPlaneIntersection", "something impossible wrong");
  }

  VECTORSCALEINVERSE(s, isect.U, isect.U);
#ifdef DEBUG
fprintf(stderr, "PlaneToPlaneIntersection: U = (%f, %f, %f), V = (%f, %f, %f), tol = %g\n",
	isect.U.x, isect.U.y, isect.U.z, 
	isect.V.x, isect.V.y, isect.V.z, 
	isect.tol);
#endif
  return isect;
}

/* Returns the signed distance from the vertex with given
 * geometric data to the face with given geometric data. 
 * Fills in on what side (+1 or -1) of the plane of the face
 * the vertex lies. If the vertex lies within the extent region of the
 * plane (i.e. the distance to the plane is less than the extent
 * of the plane), 0 is filled in in side. */
static double SignedDistanceToPlane(VERTEX *v, FACE *f, int *side)
{
  double dist;

  dist = VECTORDOTPRODUCT(f->normal, v->point) + f->d;

  if (dist < -f->extent)
    *side = -1;
  else if (dist > f->extent)
    *side = +1;
  else
    *side = 0;
  
  return dist;
}

/* Returns TRUE if the vertex approximately is on the intersection line. See
 * Segal, SIGGRAPH '90 p 111 */
static int VertexOnIntersectionLine(VERTEX *v, double dist)
{
  if (fabs(dist) < v->tolerance + isect.tol)
    return TRUE;
  else
    return FALSE;

#ifdef NEVER
  double d, e;
  DVECTOR x;

  /* numerically unstable: sometimes, about equal large numbers are subtracted */
  VECTORDIFF(v->point, isect.V, x);
  d = VECTORDOTPRODUCT(isect.U, x);
  d *= -d;
  d += VECTORDOTPRODUCT(x, x);
  e = isect.tol + v->tolerance;
  e *= e;
  return (d <= e);
#endif
}

/* Computes the distance of the vertex to the plane of the face in *dist.
 * Returns 0 if the vertex approximately lies on the intersection line
 * of the two intersectin faces being run and -1 or +1, the side w.r.t.
 * the plane if not. */
static int VertexToFaceIntersection(BREP_VERTEX *v, BREP_FACE *f, double *dist)
{
  int side;

  *dist = SignedDistanceToPlane(v->client_data, f->client_data, &side);
  if (VertexOnIntersectionLine(v->client_data, *dist)) 
    side = 0;

  return side;
}

/* Computes the parameter ti of an intersection point P = v1 + te.(v2-v1),
 * on the intersection line ti.U + V of the planes of two intersecting faces. */
static double ProjectIntersection(VERTEX *v1, VERTEX *v2, double te)
{
  DVECTOR x;

  VECTORDIFF(v2->point, v1->point, x);
  VECTORSUMSCALED(v1->point, te, x, x);
  VECTORDIFF(x, isect.V, x);
  return VECTORDOTPRODUCT(x, isect.U);
}

/* Same as above, except that the intersection is at a vertex */
static double ProjectVertexIntersection(VERTEX *v)
{
  DVECTOR x;

  VECTORDIFF(v->point, isect.V, x);
  return VECTORDOTPRODUCT(x, isect.U);
}

/* The edge from v1 to v2 belonging to the face intersects another face. The
 * intersection line of the face planes has direction vector U. This function 
 * returns ENTRY_ENDPOINT if the intersection point will be an entry endpoint 
 * of a cutting interval or EXIT_ENDPOINT if it will be an exit endpoint. 
 * It relies on the convention that the interior of a face is on the
 * left side of each contour bounding the face, or: (v2-v1) x n, with v1 and v2
 * the endpoints of a wing and n the normal to the face plane, points to the
 * interior of the face. The intersection point will be an entry endpoint if
 * this vector and U point to the same direction. It will be an exit endpoint
 * if they point into opposite directions. */
static int TestEntryExit(VERTEX *v1, VERTEX *v2, FACE *f)
{
  DVECTOR x;
  
  VECTORDIFF(v2->point, v1->point, x);
  VECTORCROSSPRODUCT(f->normal, x, x);
#ifdef DDDEBUG
  fprintf(stderr, "TestEntryExit: v1 = (%f,%f,%f), v2 = (%f,%f,%f), f->normal = (%f,%f,%f), x = (%f,%f,%f), dotp isect.U = %f\n",
	  v1->point.x, v1->point.y, v1->point.z,
	  v2->point.x, v2->point.y, v2->point.z,
	  f->normal.x, f->normal.y, f->normal.z, 
	  x.x, x.y, x.z, VECTORDOTPRODUCT(x, isect.U));
#endif
  return (VECTORDOTPRODUCT(x, isect.U) > 0. ? ENTRY_ENDPOINT : EXIT_ENDPOINT);
}

/* v0->v1 and v1->v2 are two edges of the boundary of the face with given
 * data. This routine returns TRUE if the corner at vertex v1 is a convex corner,
 * and returns FALSE if it is a concave corner. Note that our conventions
 * about edge directions are different than in Segals papers. In our case, the
 * inside of a face is always on the left side of the boundary. Same situation
 * for outer and inner contours. */
static int ConvexCorner(VERTEX *v0, VERTEX *v1, VERTEX *v2, FACE *f)
{
  DVECTOR e1, e2, x;

  VECTORDIFF(v1->point, v0->point, e1);
  VECTORDIFF(v2->point, v1->point, e2);
  VECTORCROSSPRODUCT(e1, e2, x);
  return VECTORDOTPRODUCT(x, f->normal) > 0.;
}

/* v0->v1 and v1->v2 are two edges along the intersection line of the planes
 * of two faces. This function returns TRUE if the edges form a notch and
 * FALSE if not. The two edges form a notch if they point into opposite
 * direction. If the edges form a notch, v1 is an endpoint of a cutting interval
 * along which the face containing the edges will need to be partitioned.
 * In that caase, it is determined whether the endpoint is an entry or
 * exit endpoint in the contributing face: v1 is
 * an entry point if the edge v1->v2 points into opposite direction of the
 * unit direction vector U of the intersection line. If v1 is an entry endpoint,
 * ENTRY_ENDPOINT is filled in in type. EXIT_ENDPOINT if it is an exit endpoint.
 * The value of type is unaltered if the edges do not form a notch, in which
 * case v1 is not an endpoint. */
static int IsNotch(VERTEX *v0, VERTEX *v1, VERTEX *v2, int *type)
{
  DVECTOR e1, e2;

  VECTORDIFF(v1->point, v0->point, e1);
  VECTORDIFF(v2->point, v1->point, e2);
  if (VECTORDOTPRODUCT(e1, e2) > 0.)
    return FALSE;
  
  *type = (VECTORDOTPRODUCT(e2, isect.U) < 0. ? ENTRY_ENDPOINT : EXIT_ENDPOINT);
  return TRUE;
}

#ifdef DEBUG
static void PrintEndpointData(FILE *out, ENDPOINT *p)
{
  if (!p) {
    fprintf(out, "(null)\n");
    return;
  }

  fprintf(out, "intersection point %d: ti = %g, tol = %g, te = %g, source = %d,\n\t type = ",
	  p - endpoints, p->ti, p->tol, p->te, p->source);

  /*  fprintf(out, "0x%08x = ", p->type); */
  if (p->type & ENTRY_ENDPOINT) fprintf(out, "ENTRY, ");
  if (p->type & EXIT_ENDPOINT ) fprintf(out, "EXIT, ");
  if (p->type & PHANTOM_ENDPOINT ) fprintf(out, "PHANTOM, ");
  if (p->type & VERTEX_ENDPOINT ) fprintf(out, "VERTEX, ");
  if (p->type & DELETED_ENDPOINT ) fprintf(out, "DELETED, ");

  PrintWingShort(out, p->wing);
  if (!p->wing || !p->wing->contour || !p->wing->next || !p->wing->contour->face)
    fprintf(out, "ERROR: BAD WING: ");
  else if (faceA && p->wing->contour->face == faceA)
    fprintf(out, "faceA, ");
  else if (faceB && p->wing->contour->face == faceB)
    fprintf(out, "faceB, ");
  else if (faceA && faceB)
    fprintf(out, "NOT faceA or faceB!");

  fprintf(out, "\n");
}
#endif


/* adds endpoint information to the list of endpoints */
static void InsertEndpoint(double ti, double t, int type, BREP_WING *wing, int source)
{
  if (nr_endpoints >= max_nr_endpoints) {
    if (max_nr_endpoints <= 0) {
      endpoints = (ENDPOINT *)Alloc(32 * sizeof(ENDPOINT));
      max_nr_endpoints = 32;
    } else {
      ENDPOINT *newendpoints; int i;

      newendpoints = (ENDPOINT *)Alloc((max_nr_endpoints + 32) * sizeof(ENDPOINT));
      for (i=0; i<max_nr_endpoints; i++)
	newendpoints[i] = endpoints[i];
      Free((char *)endpoints, max_nr_endpoints * sizeof(ENDPOINT));
      endpoints = newendpoints;
      max_nr_endpoints += 32;
    }
  }

  endpoints[nr_endpoints].ti = ti;
  endpoints[nr_endpoints].te = t;
  endpoints[nr_endpoints].type = type;
  endpoints[nr_endpoints].source = source;
  endpoints[nr_endpoints].wing = wing;
  endpoints[nr_endpoints].tol = isect.tol;	/* see Segal SIGGRAPH '90 p 111 */

  if (wing->contour->face == faceA) {
    if (ti < timinA) timinA = ti;
    if (ti > timaxA) timaxA = ti;
  } else {
    if (ti < timinB) timinB = ti;
    if (ti > timaxB) timaxB = ti;
  }

#ifdef DEBUG
  fprintf(stderr, "%d: ", nr_endpoints);
  PrintEndpointData(stderr, &endpoints[nr_endpoints]);
#endif
  nr_endpoints++;
}

/* Wing is a "phantom" wing or a wing that results after splitting a contour.
 * This routine will mark the contour to which wing belongs and the contour
 * to which the wing in the reverse direction belongs whether they are
 * inside or outside the other solid. This information is needed to do
 * inside/outside classification for CSG operations. */
static void MarkContours(BREP_WING *wing)
{
  BREP_FACE *main_face, *transversal_face;
  DVECTOR s, u, p1, p2;
  VECTOR nm, nt;
  unsigned int mark, prevmark;
#ifdef DEBUG
  fprintf(stderr, "MarkContour:\n");
  fprintf(stderr, "\twing = "); PrintWingShort(stderr, wing); fprintf(stderr, "\n");
  fprintf(stderr, "\totherwing = "); PrintWingShort(stderr, BrepEdgeOtherWing(wing)); fprintf(stderr, "\n");
#endif /*DEBUG*/

  main_face = wing->contour->face;
  transversal_face = (main_face == faceA ? faceB : faceA);
#ifdef DDEBUG
  if (debug) {
    fprintf(stderr, "main_face = "); PrintFaceData(stderr, main_face->client_data); fprintf(stderr, "\n");
    fprintf(stderr, "transversal_face = "); PrintFaceData(stderr, transversal_face->client_data); fprintf(stderr, "\n");
  }
#endif
  
  p1 = ((VERTEX *)(wing->vertex->client_data))->point;
  p2 = ((VERTEX *)(wing->next->vertex->client_data))->point;
  VECTORDIFF(p2, p1, u);
  nm = ((FACE *)(main_face->client_data))->normal; 
  /* nm = ((CONTOUR *)(wing->contour->client_data))->normal; */
  VECTORCROSSPRODUCT(nm, u, s);
  nt = ((FACE *)(transversal_face->client_data))->normal;
  if (VECTORDOTPRODUCT(s, nt) > 0.) 
    mark = OUTMASK;
  else
    mark = INMASK;

  prevmark = ((CONTOUR *)(wing->contour->client_data))->id & (OUTMASK | INMASK);
#ifdef DDEBUG
    fprintf(stderr, "p1 = (%f,%f,%f)\n", p1.x, p1.y, p1.z);
    fprintf(stderr, "p2 = (%f,%f,%f)\n", p2.x, p2.y, p2.z);
    fprintf(stderr, "u  = (%f,%f,%f)\n", u.x, u.y, u.z);
    fprintf(stderr, "nm  = (%f,%f,%f)\n", nm.x, nm.y, nm.z);
    fprintf(stderr, "s  = (%f,%f,%f)\n", s.x, s.y, s.z);
    fprintf(stderr, "nt  = (%f,%f,%f)\n", nt.x, nt.y, nt.z);
    fprintf(stderr, "VECTORDOTPRODUCT(s, nt) = %f ==> ",
	    VECTORDOTPRODUCT(s, nt));
#endif
#ifdef DEBUG
    fprintf(stderr, "\tprevmark = %s%s -> mark = %s%s\n", 
	    (prevmark & INMASK ? "INSIDE " : ""),
	    (prevmark & OUTMASK ? "OUTSIDE " : ""),
	    (mark & INMASK ? "INSIDE " : ""),
	    (mark & OUTMASK ? "OUTSIDE " : ""));
#endif

  ((CONTOUR *)(wing->contour->client_data))->id &= ~(OUTMASK | INMASK);
  ((CONTOUR *)(wing->contour->client_data))->id |= mark;
    
  if (BrepEdgeOtherWing(wing)->contour) {  
    CONTOUR *ocontour_data = (CONTOUR *)(BrepEdgeOtherWing(wing)->contour->client_data);
    unsigned int omark = (mark == OUTMASK ? INMASK : OUTMASK);
#ifdef DEBUG
      fprintf(stderr, "\tocontour inout mask = %s%s -> omark = %s%s\n",
	      ocontour_data->id & OUTMASK ? "OUTSIDE" : "",
	      ocontour_data->id & INMASK ? "INSIDE" : "",
	      omark & OUTMASK ? "OUTSIDE" : "",
	      omark & INMASK ? "INSIDE" : "");
#endif
/*
    if (!(ocontour_data->id & (INMASK|OUTMASK))) {
      ocontour_data->id |= omark;
    } else if ((ocontour_data->id & (INMASK|OUTMASK)) != omark)
      Warning("MarkContour", "contour %d: inconsistent inside/outside determination", EntityId(wing->contour->client_data));
*/
    ocontour_data->id &= ~(INMASK|OUTMASK);
    ocontour_data->id |= omark;
  }
}

/* in/out-marks the contour to which wing belongs if it has not been
 * marked before */
static void ConditionallyMarkContours(BREP_WING *wing)
{
  if ((((CONTOUR *)(wing->contour->client_data))->id & (INMASK | OUTMASK)) == 0)
    MarkContours(wing);
}

/* Test each wing of the main face, starting with first_wing, for intersection with
 * the transversal face. Most of the intersection points will be an endpoint of a
 * cutting interval, along which the faces will need to be partitioned. Some
 * however will not. Record the endpoints with information which wing contributed
 * it and whether it is an entry or exit point on the contributing face. 
 * "first_wing" is known to be the first wing contributing an intersection point.
 * Previous wings and contours have been checked already in ClassifyIntersection() */
static void DetermineEndpoints(BREP_FACE *main_face, BREP_WING *first_wing, BREP_FACE *transversal_face)
{
  BREP_CONTOUR *contour;
  BREP_WING *wing;
  double d0, d1, d2, te, ti;
  int side0, side1, side2, type, inphantom, firstphantom;
#ifdef DEBUG
fprintf(stderr, "DetermineEndpoints:\n\tmain_face = %s: ",
	main_face == faceA ? "faceA" : "faceB");
PrintFaceData(stderr, main_face->client_data);
fprintf(stderr, "\n\ttransversal_face = %s: ",
	transversal_face == faceA ? "faceA" : "faceB");
PrintFaceData(stderr, transversal_face->client_data);
fprintf(stderr, "\n");
#endif

#define START_OVER	/* start over looking for intersection from the
			 * first wing of the first contour on */
#ifdef START_OVER
  contour = main_face->outer_contour;
#else /*START_OVER*/
  contour = first_wing->contour;
#endif /*START_OVER*/
  do {
#ifdef START_OVER
    wing = contour->wings;
#else /*START_OVER*/
    if (contour == first_wing->contour) {
/*
      if (first_wing != contour->wings)
	wing = first_wing->prev;
      else
*/
	wing = first_wing;
    } else
      wing = contour->wings;
#endif /*START_OVER*/
#ifdef DEBUG
    fprintf(stderr, "***** contour = %d:\n\twing = ", EntityId(contour->client_data));
    PrintWingShort(stderr, wing);
    fprintf(stderr, "\n\tfirst wing of contour = ");
    PrintWingShort(stderr, contour->wings);
    fprintf(stderr, "\n");
#endif

    side2 = VertexToFaceIntersection(wing->vertex, transversal_face, &d2);
    side1 = UNKNOWN_SIDE;	/* an impossible value indicating that we don't know 
				 * on what side of the transversal plane the starting 
				 * point of the previous wing lies --- this is only 
				 * needed in rare cases */
    firstphantom = inphantom = FALSE;		/* not in a phtanom interval */
    
    do {
      side0 = side1;
      d1 = d2; side1 = side2;
      side2 = VertexToFaceIntersection(wing->next->vertex, transversal_face, &d2);
#ifdef DEBUG
      fprintf(stderr, "v%d (%f,%f,%f) '%s' -> v%d: d1=%g, side1=%d, d2=%g, side2=%d:\n",
	      ((VERTEX *)(wing->vertex->client_data))->id & ~IDMASK,
	      ((VERTEX *)(wing->vertex->client_data))->point.x, 
	      ((VERTEX *)(wing->vertex->client_data))->point.y, 
	      ((VERTEX *)(wing->vertex->client_data))->point.z, 
	      (((VERTEX *)(wing->vertex->client_data))->name ? 
	       ((VERTEX *)(wing->vertex->client_data))->name : "(null)"),
	      ((VERTEX *)(wing->next->vertex->client_data))->id & ~IDMASK,
	      d1, side1, d2, side2);
#endif
      if (side1 != side2 && (side1+side2) == 0) { /* intersection halfway the 
						   * line segment */

	/* determine parameter of intersection point on the intersection line of 
	 * the main and transversal face */
	te = fabs(d1) / (fabs(d1) + fabs(d2));
	ti = ProjectIntersection(wing->vertex->client_data, wing->next->vertex->client_data, te);
	type = TestEntryExit(wing->vertex->client_data, wing->next->vertex->client_data, main_face->client_data);
	InsertEndpoint(ti, te, type, wing,  __LINE__);	/* always an endpoint */

      } else if (side1 == 0) {			/* starting point lies in the 
						 * transversal plane */
	te = 0.;
	ti = ProjectVertexIntersection(wing->vertex->client_data);

	/* we need to know whether the starting vertex of the previous wing also
	 * lies in the transversal plane or not */
	if (side0 == UNKNOWN_SIDE)  {
	  side0 = VertexToFaceIntersection(wing->prev->vertex, transversal_face, &d0);

	  /* if the last wing lies inside the transversal plane, determine whether 
	   * a previous vertex has contributed a phantom endpoint or not. If this
	   * "previous" vertex contributes a phantom endpoint, it will be generated 
	   * later, when the vertex will be processed. */
	  if (side0 == 0) {
	    int prevside0, prevside1, prevside2, prevtype;
	    double prevd0, prevd1, prevd2;
	    BREP_WING *prevwing;

	    /* wing->vertex and wing->prev->vertex lie in the transversal plane.
	     * Search backward to find the last vertex that is the top of a notch,
	     * or that doesn't lie in the transversal plane. Normally, there is such a
	     * vertex since the main face is known not to be coplanar with the
	     * transversal face, and contours are closed: so, either there is a
	     * vertex in the contour that does not lie in the transversal plane,
	     * or they all lie on the intersection line, but then there must be
	     * a notch (even two) somewhere. However, due to the finite
	     * precision and extents, it might happen that a contour of
	     * the main face is judged to be coplanar with the
	     * transversal face. */
	    prevside2 = side1; prevd2 = d1;
	    prevside1 = side0; prevd1 = d0; prevwing = wing->prev;
	    prevside0 = VertexToFaceIntersection(prevwing->prev->vertex, transversal_face, &prevd0);

	    while (prevside0 == 0 && !IsNotch(prevwing->prev->vertex->client_data,
					      prevwing->vertex->client_data,
					      prevwing->next->vertex->client_data,
					      &prevtype) &&
		   prevwing != wing) {
	      prevside2 = prevside1; prevd2 = prevd1;
	      prevside1 = prevside0; prevd1 = prevd0; prevwing = prevwing->prev;
	      prevside0 = VertexToFaceIntersection(prevwing->prev->vertex, transversal_face, &prevd0);	 
	    }

	    if (prevwing == wing) {	/* should not happen, but we prevent infinite
					 * loops if it would happen. */
	      Warning("DetermineEndpoints", "bad contour"); 
	      goto next_contour;
	    }

	    if (prevside0 != 0) {
	      inphantom = TRUE;
	    } else {
	      /* a previous vertex is the top of a notch, or the
	       * whole contour lies in the transversal plane */
#ifdef DEBUG
	      fprintf(stderr, "previous vertex %d is top of a notch or the contour lies entirely in the transversal plane.\n",
		      ((VERTEX *)(prevwing->vertex->client_data))->id & ~IDMASK);
#endif

	      /* find a vertex not lying in the transversal plane. */
	      while (prevside0 == 0 && prevwing != wing) {
		prevside2 = prevside1; prevd2 = prevd1;
		prevside1 = prevside0; prevd1 = prevd0; prevwing = prevwing->prev;
		prevside0 = VertexToFaceIntersection(prevwing->prev->vertex, transversal_face, &prevd0);	 
	      }

	      if (prevside0 != 0) {
		/* we found a vertex outside the transversal plane, so
		 * we have a notch */
		inphantom = FALSE;
	      } else {
#ifdef DEBUG
		fprintf(stderr, "the contour lies entirely in the transversal plane ... we skip it.\n");
		goto next_contour;
#endif
		inphantom = TRUE;	/* eens zien wat dit geeft */
	      }
	    }
	    firstphantom = inphantom;
	  }
	}

#ifdef DEBUG
      fprintf(stderr, "v%d <- : side0=%d, inphantom = %s\n",
	      ((VERTEX *)(wing->prev->vertex->client_data))->id & ~IDMASK,
	      side0, inphantom ? "TRUE" : "FALSE");
#endif	
	if (side0 != 0 && side2 != 0) {		/* both adjacent endpoints lie outside
						 * the transversal plane */

	  if (side0 != side2) {			/* the previous and next endpoint lie
						 * on opposite sides */

	    type = TestEntryExit(wing->vertex->client_data, wing->next->vertex->client_data, main_face->client_data);
	    InsertEndpoint(ti, te, type|VERTEX_ENDPOINT, wing, __LINE__);	/* always an endpoint */
	    
	  } else {				/* the previous and next endpoint lie 
						 * on the same side */

	    if (!ConvexCorner(wing->prev->vertex->client_data, wing->vertex->client_data, wing->next->vertex->client_data, main_face->client_data)) {
	      /* if the previous and current wing make a convex corner at the starting
	       * vertex of the current wing, no endpoints are induced. If it is a
	       * concave corner, two endpoints are induced: one entry and one exit,
	       * at the same physical location. */

	      InsertEndpoint(ti, te, EXIT_ENDPOINT|VERTEX_ENDPOINT, wing, __LINE__);
	      InsertEndpoint(ti, te, ENTRY_ENDPOINT|VERTEX_ENDPOINT, wing, __LINE__);

	   } else {

	      /* not an endpoint, no partitioning of the main or traversal face is
	       * necessary. The intersection point is stored anyways in order to
	       * merge with other intersection points that lie within tolerance 
	       * form it. */

	      InsertEndpoint(ti, te, VERTEX_ENDPOINT, wing, __LINE__);

	    }

	  }

	} else if (side0 == 0) {		/* previous endpoint lies in the
						 * transversal plane */

	  if (side2 != 0) {			/* next endpoint doesn't lie in the
						 * transversal plane */

	    /* The previous edge lies in the transversal plane, the current edge 
	     * not. No partitioning of the main face is necessary along the
	     * previous wing, but partitioning of the transversal plane might be, 
	     * so we add a phantom endpoint, induced by the previous edge, to the 
	     * endpoint list. */

	    if (inphantom) {
	      InsertEndpoint(ti, 1., PHANTOM_ENDPOINT|VERTEX_ENDPOINT, wing->prev, __LINE__);
	      inphantom = FALSE;
	    }

	    if (!ConvexCorner(wing->prev->vertex->client_data, wing->vertex->client_data, wing->next->vertex->client_data, main_face->client_data)) {

	      /* if the corner at the starting vertex of the current wing is not a 
	       * convex corner, the starting vertex is also a real endpoint: the 
	       * entry of exit of a cutting interval along which the main face
	       * *will* need to be partitioned. */
	      
	      type = TestEntryExit(wing->vertex->client_data, wing->next->vertex->client_data, main_face->client_data);
	      InsertEndpoint(ti, te, type|VERTEX_ENDPOINT, wing, __LINE__);	

	    } else {

	      /* no endpoint needs to be inserted. Not even a VERTEX
	       * endpoint */

	    }
	    
	  } else {				/* both previous and next lie in
						 * the transversal plane */

	    if (IsNotch(wing->prev->vertex->client_data, wing->vertex->client_data, wing->next->vertex->client_data, &type)) {
	      /* two wings along the intersection line, going into opposite
	       * direction: a notch. */

	      /* For a normal notch, inphantom should be TRUE. For a slit
	       * however, inphantom may be either TRUE or FALSE. */
	      InsertEndpoint(ti, 1., PHANTOM_ENDPOINT|VERTEX_ENDPOINT, wing->prev, __LINE__);
	      inphantom = 1 - inphantom;

	      InsertEndpoint(ti, te, type|VERTEX_ENDPOINT, wing, __LINE__);

	    } else {
	      /* Two wings along the intersection line, going in the same direction.
	       * No partitioning of the main face is necessary. Partitioning of
	       * the transversal face might be necessary, but the starting vertex of
	       * the current wing is not an endpoint. */

	      InsertEndpoint(ti, te, VERTEX_ENDPOINT, wing, __LINE__);

	    }

	  }

	} else if (side2 == 0) {		/* current and next endpoint, i.o.w.
						 * the current edge, lies in the
						 * traversal plane (but the previous 
						 * edge does not). */

	  if (!ConvexCorner(wing->prev->vertex->client_data, wing->vertex->client_data, wing->next->vertex->client_data, main_face->client_data)) {

	    /* if the corner at the starting vertex of the current wing is not a 
	     * convex corner, the starting vertex is a real endpoint: the 
	     * entry of exit of a cutting interval along which the main face
	     * *will* need to be partitioned. */
	      
	    type = TestEntryExit(wing->prev->vertex->client_data, wing->vertex->client_data, main_face->client_data);
	    InsertEndpoint(ti, 1.0, type|VERTEX_ENDPOINT, wing->prev, __LINE__);

	  } else {

	    /* no endpoint needs to be inserted. Not even a VERTEX
	     * endpoint */

	  }
	  
	  /* The main face doesn't need to be cut for the current wing, but the 
	   * transversal face might however. Insert the starting vector as a phantom 
	   * endpoint into the endpoint list. */

	  if (!inphantom) {
	    InsertEndpoint(ti, te, PHANTOM_ENDPOINT|VERTEX_ENDPOINT, wing, __LINE__);
	    inphantom = TRUE;
	  }

	  /* The main face doesn't need to be cut, but we can mark the contour
	   * as lying inside or outside the other solid.
	   * Out guess might not be correct, so, don't mark
	   * the contour if already marked before. If the contour was not
	   * yet marked, we might mark it wrong if the contour contains 
	   * wings on both sides of the other solid. In that case, however,
	   * the contour will be split later, and the fragments marked
	   * correctly. */
	  ConditionallyMarkContours(wing);
	}

      } else {

	/* either there is no intersection, or only the end vertex lies inside
	 * the transversal plane. In the latter case, the intersection will be handled
	 * as an intersection of the next wing with the transversal plane, or has been
	 * handled as an intersection of the first wing in the contour. */

      }

      wing = wing->next;
    } while (wing != contour->wings);

    if (inphantom != firstphantom) {
      Warning("DetermineEndpoints", "something wrong with phantom endpoints"); 
    }

  next_contour:
    contour = contour->next;
  } while (contour != main_face->outer_contour);
}

/* returns +1 if p1 comes after p2, -1 if p1 preceeds p2 and 0
 * if they are exactly at the same position */
static int CompareEndpointPosition(ENDPOINT *p1, ENDPOINT *p2)
{
#ifdef DDEBUG
  fprintf(stderr, "CompareEndpointPosition: endpoint %d vs. %d\n", p1-endpoints, p2-endpoints);
#endif
  return (p1->ti > p2->ti) ? +1 : -1;
}

/* returns TRUE if the endpoints approximately coincide */
static int CoincidingEndpoints(ENDPOINT *p1, ENDPOINT *p2)
{
  double d, tol;
  d = p2->ti - p1->ti; if (d < 0.) d = -d;
  tol = p1->tol + p2->tol;
  return (d <= tol);
}

/* Enlarges the tolerance region of a vertex. probably also a good place to merge
 * the vertex with other approximately coinciding vertices */
static void VertexEnlargeTolerance(VERTEX *v, double tol)
{
  if (tol > v->tolerance)
    v->tolerance = tol;
}

/* Align a series of endpoints that are within each others tolerance region */
static void AlignEndpoints(int start, int end)
{
  int i, nr_vertex_endpoints;
  double tmin, tmax, t_align, tol;
  ENDPOINT *p;
#ifdef DEBUG
  fprintf(stderr, "AlignEndpoints:\n");
#endif
  /* find a good position on the intersection line for alignment and corresponding
   * tolerance region. */
  nr_vertex_endpoints = 0; t_align = 0.;
  tmin = HUGE; tmax = -HUGE;
  for (i=start, p=&endpoints[start]; i<=end; i++, p++) {
    if (p->type != VERTEX_ENDPOINT) {
      if (p->ti - p->tol < tmin)
	tmin = p->ti - p->tol;
      if (p->ti + p->tol > tmax)
	tmax = p->ti + p->tol;
    }
    if (p->type & VERTEX_ENDPOINT) {
      t_align += p->ti;
      nr_vertex_endpoints++;
    }
  }

  if (nr_vertex_endpoints == 0) {
    t_align = (tmin + tmax) * 0.5;
    tol = (tmax - tmin) * 0.5;
  } else {
    double m;
    t_align /= (double)nr_vertex_endpoints;
    tol = tmax - t_align;
    m = t_align - tmin;
    if (m > tol) tol = m;
  }
#ifdef DEBUG
  fprintf(stderr, "file %s line %d: %d vertex endpoints found, tmin = %g, tmax = %g., t_align = %g, tol = %g\n",
	  __FILE__, __LINE__,
	  nr_vertex_endpoints, tmin, tmax, t_align, tol);
#endif

  /* Align the endpoints. Enlarge the tolerance region of vertex endpoints to
   * include the aligned intersection position. */
  for (i=start, p=&endpoints[start]; i<=end; i++, p++) {
    if (p->type & VERTEX_ENDPOINT) { 
      /* enlarge the tolerance region of the vertex to include the intersection point */
      VERTEX *v; double d;

      if (p->te == 0.)
	v = (VERTEX *)(p->wing->vertex->client_data);
      else {
	if (p->te != 1.)
	  Warning("MergeInterval", "vertex endppoint with te != 0. and != 1.");
	v = (VERTEX *)(p->wing->next->vertex->client_data);
      }
      
      d = (t_align > p->ti ? t_align - p->ti : p->ti - t_align);
      p->tol = MAX(d, p->tol);
      if (p->tol > v->tolerance)
	VertexEnlargeTolerance(v, p->tol);
      p->ti = t_align;
    } else {
      p->ti = t_align;
      p->tol = tol;
    }
  }
}

/* swap the endpoints */
static void SwapEndpoints(int i, int j)
{
  ENDPOINT r;

  r = endpoints[i];
  endpoints[i] = endpoints[j];
  endpoints[j] = r;
}

/* secondary endpoint sorting:
 * sorts the endpoints in the interval such that the endpoints induced by faceA
 * all preceed the endpoints induced by faceB. Returns in firstA and lastA
 * the index to the first and last endpoint induced by faceA and similar for faceB */
static void SortEndpointFace(int start, int end,
			     int *firstA, int *lastA, int *firstB, int *lastB)
{
#ifdef DEBUG
  fprintf(stderr, "SortEndpointFace:\n");
#endif
  *firstB = start; 
  while (*firstB <= end && endpoints[*firstB].wing->contour->face != faceB) (*firstB)++;
  /* *firstB is the index of the first endpoint induced by faceB */
  
  *lastA = end;   
  while (*lastA >= start && endpoints[*lastA].wing->contour->face != faceA) (*lastA)--;
  /* lastA is the index of the last endpoint induced by faceA */

  while (*lastA > *firstB) {
    SwapEndpoints(*firstB, *lastA);

    while (*firstB <= end && endpoints[*firstB].wing->contour->face != faceB) (*firstB)++;
    while (*lastA >= start && endpoints[*lastA].wing->contour->face != faceA) (*lastA)--;
  }

  *firstA = start;
  *lastB = end;
}

/* exit endpoints must preceed phantom endpoints, phantom endpoints
 * must preceed entry endpoints. This routine returns +1 if the
 * order needs to be changed, 0 is the endpoints have the same
 * type and -1 if the order is correct. */
static int CompareEndpointType(ENDPOINT *p1, ENDPOINT *p2)
{
  if (p1->type == p2->type) return 0;
  else if (p1->type & EXIT_ENDPOINT) return -1;
  else if (p2->type & EXIT_ENDPOINT) return +1;
  else if (p1->type & PHANTOM_ENDPOINT) return -1;
  else if (p2->type & PHANTOM_ENDPOINT) return +1;
  else if (p1->type & ENTRY_ENDPOINT) return -1;
  else if (p2->type & ENTRY_ENDPOINT) return +1;
  else	/* other irrelevant endpoint type */
    return 0;
}

/* tertiary sorting: first exit endpoints, then phantom endpoints,
 * then entry endpoints. Note that the series to be sorted will
 * be very small (at most a few: coinciding endpoints induced by
 * the same face) */
static void SortEndpointType(int first, int last)
{
  int i, j;
#ifdef DEBUG
  fprintf(stderr, "SortEndpointType(%d, %d):\n", first, last);
#endif
  for (i = first; i < last; i++) {
    for (j = i+1; j <= last; j++) {
      if (CompareEndpointType(&endpoints[i], &endpoints[j]) > 0) {
	SwapEndpoints(i, j);
      }
    }
  }
}

/* annihilates pairs of endpoints that would cause zero length cutting intervals */
static void AnnihilateEndpoints(int start, int end)
{
  int i, first_exit=end+1, last_exit=start-1, nr_exit = 0,
         first_phantom=end+1, last_phantom=start-1, nr_phantom = 0,
         first_entry=end+1, last_entry=start-1, nr_entry = 0;

#ifdef DEBUG
  fprintf(stderr, "AnnihilateEndpoints(%d, %d):\n", start, end);
#endif
  for (i=start; i<=end; i++) {
    if (endpoints[i].type & EXIT_ENDPOINT) {
      if (i < first_exit) first_exit = i;
      if (i > last_exit) last_exit = i;
      nr_exit ++;
    }
    else if (endpoints[i].type & PHANTOM_ENDPOINT) {
      if (i < first_phantom) first_phantom = i;
      if (i > last_phantom) last_phantom = i;
      nr_phantom ++;
    }
    else if (endpoints[i].type & ENTRY_ENDPOINT) {
      if (i < first_entry) first_entry = i;
      if (i > last_entry) last_entry = i;
      nr_entry ++;
    }
    else 
      endpoints[i].type |= DELETED_ENDPOINT;
  }

  while (nr_phantom >= 2) {
    endpoints[first_phantom++].type |= DELETED_ENDPOINT; 
    endpoints[first_phantom++].type |= DELETED_ENDPOINT; 
    nr_phantom -= 2;
#ifdef DEBUG
    fprintf(stderr, "file %s line %d: pair phantom endpoints annihilated\n", __FILE__, __LINE__);
#endif
  }

  while (nr_exit > 1 && nr_entry > 1) {
    endpoints[first_exit++].type |= DELETED_ENDPOINT; 
    endpoints[first_entry++].type |= DELETED_ENDPOINT; 
    nr_exit--; nr_entry--;
#ifdef DEBUG
    fprintf(stderr, "file %s line %d: entry/exit endpoint pair annihilated\n", __FILE__, __LINE__);
#endif
  }
}

/* Given start and end index of a series of approximately coinciding endpoints,
 * align the endpoints, (secondary) sort them according to their contributing
 * face, (tertiary) sort them according to their type and remove endpoints that 
 * would lead to zero length cutting intervals. */
static void MergeInterval(int start, int end)
{
  int firstA, firstB, lastA, lastB;

#ifdef DEBUG
  fprintf(stderr, "--- MergeInterval(%d,%d):\n", start, end);
#endif
#ifdef DDEBUG
  int i;
#endif
  AlignEndpoints(start, end);
  SortEndpointFace(start, end, &firstA, &lastA, &firstB, &lastB);
#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: after secondary sorting: firstA = %d, lastA = %d, firstB = %d, lastB = %d\n", 
	  __FILE__, __LINE__, firstA, lastA, firstB, lastB);
#endif /*DDEBUG*/
  SortEndpointType(firstA, lastA);
  SortEndpointType(firstB, lastB);
#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: after tertiary sorting:\n", __FILE__, __LINE__);
  for (i=start; i<=end; i++) 
    PrintEndpointData(stderr, &endpoints[i]);
#endif /*DDEBUG*/
  AnnihilateEndpoints(firstA, lastA);
  AnnihilateEndpoints(firstB, lastB);
#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: after annihilating:\n", __FILE__, __LINE__);
  for (i=start; i<=end; i++) 
    PrintEndpointData(stderr, &endpoints[i]);
#endif /*DDEBUG*/
}

/* Aligns coinciding endpoints, (secondary) sorts them according to their 
 * contributing face, (tertiary) sorts them according to their type and 
 * removes endpoints that would lead to zero length cutting intervals. */
static void MergeEndpoints(void)
{
  int i, start;
  ENDPOINT *p, *prev;
#ifdef DDEBUG
fprintf(stderr, "MergeSortAnnihilateEndpoints:\n");
#endif
  start = 0;	/* index of start of series of approximately coinciding endpoints */
  prev = &endpoints[0]; 
  for (i=1, p=&endpoints[1]; i<nr_endpoints; prev=p, i++, p++) {
    if (!CoincidingEndpoints(prev, p)) {
      if (i-1 > start)
	MergeInterval(start, i-1);
      start = i;
    }
  }
  if (nr_endpoints-1 > start)
    MergeInterval(start, nr_endpoints-1);
}

/* Sorts the endpoints contributed by the intersecting faces. */
static void SortEndpoints(void)
{
#ifdef DEBUG
fprintf(stderr, "SortEndpoints: %d endpoints\n", nr_endpoints);
#endif
  qsort(endpoints, nr_endpoints, sizeof(ENDPOINT), (int (*)(const void *, const void *))CompareEndpointPosition);
#ifdef DEBUG
  fprintf(stderr, "file %s line %d: after primary sorting:\n", __FILE__, __LINE__);
  {int i; for (i=0; i<nr_endpoints; i++) PrintEndpointData(stderr, &(endpoints[i]));}
#endif
  MergeEndpoints();
#ifdef DEBUG
  fprintf(stderr, "file %s line %d: after merging etc...:\n", __FILE__, __LINE__);
  {int i; for (i=0; i<nr_endpoints; i++) PrintEndpointData(stderr, &(endpoints[i]));}
#endif
}

/* Given the endpoint data p for an endpoint, this routine computes what
 * vertex data correspond to it, looks up whether a vertex with this data
 * already exists in the solid to which the face belongs, and if not found, 
 * creates such a vertex. A pointer to the vertex is returned. The face does
 * not have to be the same face as the face contributing the endpoint. */
static BREP_VERTEX *MakeVertex(BREP_FACE *face, ENDPOINT *p)
{
  BREP_VERTEX *thevertex;
#ifdef DEBUG
  fprintf(stderr, "MakeVertex:\n");
  fprintf(stderr, "\tface = %s\n", face == faceA ? "faceA" : (face == faceB ? "faceB" : "NOT faceA or faceB"));
  fprintf(stderr, "\t"); PrintEndpointData(stderr, p); fprintf(stderr, "\n");
#endif

  if ((p->type & VERTEX_ENDPOINT) && p->wing->contour->face == face) {
    if (p->te == 0.) 
      thevertex = p->wing->vertex;
    else
      thevertex = p->wing->next->vertex;
  } else {		/* endpoint somewhere halfway the edge or on 
			 * the other face */
    VERTEX v, *v1, *v2;
    v1 = (VERTEX *)(p->wing->vertex->client_data);
    v2 = (VERTEX *)(p->wing->next->vertex->client_data);

    /* Sometimes, MakeVertex() is called to create a new vertex at the end of 
     * the previous cutting interval for a face. Return the previous end
     * vertex in this case instead of creating a new vertex at the same
     * position. */
#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: face = %s, p->ti = %f, lastti = %f\n", __FILE__, __LINE__,
	  (face == faceA ? "faceA" : (face == faceB ? "faceB" : "face ???")), 
	  p->ti, lastti);
  fprintf(stderr, "\tlastwingA = "); PrintWingShort(stderr, lastwingA); fprintf(stderr, "\n");
  fprintf(stderr, "\tlastwingB = "); PrintWingShort(stderr, lastwingB); fprintf(stderr, "\n");
#endif
    if (p->ti == lastti) {
      if (face == faceA && lastwingA) {
#ifdef DEBUG
	fprintf(stderr, "file %s line %d: new vertex = lastwingA->vertex\n", __FILE__, __LINE__);
#endif
	return lastwingA->vertex;
      } else if (face == faceB && lastwingB) {
#ifdef DEBUG
	fprintf(stderr, "file %s line %d: new vertex = lastwingB->vertex\n", __FILE__, __LINE__);
#endif
	return lastwingB->vertex;
      }
    }

    v.id = p - endpoints;
    VECTORSUMSCALED(isect.V, p->ti, isect.U, v.point); 

    if (p->wing->contour->face == face) {
      /* compute a nice interpolated normal */
      VECTORSCALE((1. - p->te), v1->normal, v.normal);
      VECTORSUMSCALED(v.normal, p->te, v2->normal, v.normal);
      VECTORNORMALIZE(v.normal);
    } else {
      if (p->type & VERTEX_ENDPOINT) {
	if (p->te == 0.)
	  v.point = v1->point;
	else
	  v.point = v2->point;
      }
      VECTORSET(v.normal, 0., 0., 0.);
    }

    v.tolerance = p->tol;
    v.name = NULL;
#ifdef DDEBUG
    fprintf(stderr, "file %s line %d: vertex data = ", __FILE__, __LINE__);
    PrintVertexData(stderr, &v);
    fprintf(stderr, "\n");
#endif
    thevertex = BrepFindVertex(&v, face->shell->solid->vertices);
#ifdef DDEBUG
    fprintf(stderr, "file %s line %d: vertex %s.\n", __FILE__, __LINE__,
	    thevertex ? "found" : "not found --- installing");
#endif
    if (!thevertex) {
      thevertex = BrepInstallVertex(&v, &(face->shell->solid->vertices));
#ifdef DEBUG
      fprintf(stderr, "file %s line %d: new ", __FILE__ ,__LINE__);
      PrintVertexData(stderr, thevertex->client_data);
      fprintf(stderr, "\n");
#endif
    }
  }

  return thevertex;
}

/* Creates a BREP_VERTEX corresponding to the intersection point and
 * belonging to the solid that contains the face that contributed the
 * intersection point. Unless this vertex is the endpoint of the
 * contributing wing, the contributing edge is split (on both sides
 * if used in two contours). The, possibly new, wing leaving at the
 * intersection point is returned. This routine also replaces all
 * other references to the split wing in the (global) endpoint list. 
 * Since the intersection point contributed by a wing is unique (within 
 * the tolerances of course), the edge line parameter (te) of the
 * updates endpoints is set to zero: the intersection point is the starting
 * point of the newly created wing that is referenced. */
static BREP_WING *MakeVertexSplitEdge(ENDPOINT *p)
{
  BREP_VERTEX *vertex;
  BREP_WING *wing;
#ifdef DEBUG
  fprintf(stderr, "MakeVertexSplitEdge: ");
  PrintEndpointData(stderr, p);
#endif  
  vertex = MakeVertex(p->wing->contour->face, p);
#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: the vertex:", __FILE__, __LINE__);
  PrintVertex(stderr, vertex);
  fprintf(stderr, "file %s line %d: the wing:", __FILE__, __LINE__);
  PrintWing(stderr, p->wing);
#endif
  if (vertex == p->wing->vertex)
    wing = p->wing;
  else if (vertex == p->wing->next->vertex)
    wing = p->wing->next;
  else {
    BREP_WING *otherwing = BrepEdgeOtherWing(p->wing), 
              *newotherwing = (BREP_WING *)NULL;
    int i;

#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: before BrepSplitWing(otherwing)\n", __FILE__, __LINE__);
  PrintWing(stderr, otherwing);
#endif
  if (otherwing->contour) {
      newotherwing = BrepSplitWing(otherwing, vertex);
#ifdef DEBUG
    fprintf(stderr, "file %s line %d: new wings ('other' side):\n", __FILE__, __LINE__);
    PrintWingShort(stderr, newotherwing->prev);
    PrintWingShort(stderr, newotherwing);
    fprintf(stderr, "\n");
#endif
  }

#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: before BrepSplitWing(p->wing)\n", __FILE__, __LINE__);
#endif
    wing = BrepSplitWing(p->wing, vertex);
#ifdef DEBUG
    fprintf(stderr, "file %s line %d: new wings ('this' side):\n", __FILE__, __LINE__);
    PrintWingShort(stderr, wing->prev);
    PrintWingShort(stderr, wing);
    fprintf(stderr, "\n");
#endif

#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: updating endpoint list.\n", __FILE__, __LINE__);
#endif
    /* Replace all references to wings of the split edge in the endpoint list. */
    for (i=0; i<nr_endpoints; i++) {
      if (endpoints[i].wing == p->wing) {
	endpoints[i].wing = wing;
	endpoints[i].te = 0.;	/* should be the same intersection point */
#ifdef DDEBUG
	fprintf(stderr, "file %s line %d: endpoint %d updated ('this' side).\n", __FILE__, __LINE__, i);
#endif
      } else if (endpoints[i].wing == otherwing) {
	endpoints[i].wing = newotherwing;
	endpoints[i].te = 0.;	/* should be the same intersection point */
#ifdef DDEBUG
	fprintf(stderr, "file %s line %d: endpoint %d updated ('other' side).\n", __FILE__, __LINE__, i);
#endif
      }
    }
  }

#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: wing = ", __FILE__, __LINE__);
  PrintWing(stderr, wing);
#endif
  return wing;
}

/* Introduces a cut from the starting vertex of wing1 to the starting vertex of
 * ing2. Assumes that the starting vertex of wing1 comes before the starting
 * vertex of wing2 on the intersection line. Returns a wing leaving at the
 * exit end of the cut. Also replaces contributing wings in the global
 * endpoint list to belong to the right contour for further processing. */
static BREP_WING *MakeCut(BREP_WING *wing1, BREP_WING *wing2)
{
  BREP_WING *wing;
  EDGE edge_data;
  CONTOUR contour_data;

  edge_data.id = 0;
  contour_data.id = 0;
#ifdef DEBUG
  fprintf(stderr, "MakeCut:\n");
  fprintf(stderr, "\twing1       = "); PrintWingShort(stderr, wing1); fprintf(stderr, "\n");
  fprintf(stderr, "\twing2       = "); PrintWingShort(stderr, wing2); fprintf(stderr, "\n");
#endif
  if (wing1->contour == wing2->contour) {
#ifdef DEBUG
    fprintf(stderr, "file %s line %d: before BrepMakeEdgeSplitContour():\n", __FILE__, __LINE__);
#endif
#ifdef DDEBUG
    fprintf(stderr, "\twing1->prev = "); PrintWingShort(stderr, wing1->prev); fprintf(stderr, "\n");
    fprintf(stderr, "\twing1       = "); PrintWingShort(stderr, wing1); fprintf(stderr, "\n");
    fprintf(stderr, "\twing2->prev = "); PrintWingShort(stderr, wing2->prev); fprintf(stderr, "\n");
    fprintf(stderr, "\twing2       = "); PrintWingShort(stderr, wing2); fprintf(stderr, "\n");
#endif
    wing = BrepMakeEdgeSplitContour(wing1, wing2, &edge_data, &contour_data);
#ifdef DEBUG
    fprintf(stderr, "file %s line %d: new wings:\n", __FILE__, __LINE__);
    PrintWingShort(stderr, wing);
    PrintWingShort(stderr, BrepEdgeOtherWing(wing));
    fprintf(stderr, "\n");
#endif
#ifdef DDEBUG
    fprintf(stderr, "file %s line %d: after BrepMakeEdgeSplitContour():\n", __FILE__, __LINE__);
    fprintf(stderr, "\t  wing->prev = "); PrintWingShort(stderr, wing->prev); fprintf(stderr, "\n");
    fprintf(stderr, "\t  wing       = "); PrintWingShort(stderr, wing); fprintf(stderr, "\n");
    fprintf(stderr, "\t  wing->next = "); PrintWingShort(stderr, wing->next); fprintf(stderr, "\n");
    fprintf(stderr, "\to wing->prev = "); PrintWingShort(stderr, BrepEdgeOtherWing(wing)->prev); fprintf(stderr, "\n");
    fprintf(stderr, "\to wing       = "); PrintWingShort(stderr, BrepEdgeOtherWing(wing)); fprintf(stderr, "\n");
    fprintf(stderr, "\to wing->next = "); PrintWingShort(stderr, BrepEdgeOtherWing(wing)->next); fprintf(stderr, "\n");
#endif
#ifdef DDEBUG
    fprintf(stderr, "file %s line %d: checking wing1->contour:\n", __FILE__, __LINE__);
    CheckContour(wing1->contour);
    fprintf(stderr, "file %s line %d: checking wing2->contour:\n", __FILE__, __LINE__);
    CheckContour(wing2->contour);
    fprintf(stderr, "file %s line %d: checking vertices:\n", __FILE__, __LINE__);
    CheckVertex(wing1->vertex);
    CheckVertex(wing2->vertex);
#endif
    /* Update the endpoints list: the exitant edge at p2 must be an edge on the contour
     * that is still to be processed. Failing to do so may cause the split of contour
     * to be merged again with the original contour when processing the next
     * cutting interval if it starts at the exit endpoint of the one being
     * processed. This situation occurs may occur when an isolated vertex of one
     * face lies in the other face. Cfr. Segal IEEE CG&A jan. 88 p 60. */
    if (ConvexCorner(wing1->vertex->client_data, wing2->vertex->client_data, wing2->next->vertex->client_data, wing2->contour->face->client_data)) {
      BREP_WING *owing = BrepEdgeOtherWing(wing);
      int i;
      for (i=0; i<nr_endpoints; i++) {
	if (endpoints[i].wing == wing2) {
	  endpoints[i].wing = owing;
#ifdef DEBUG
	fprintf(stderr, "file %s line %d: endpoint %d updated: wing ", __FILE__, __LINE__, i);
	PrintWingShort(stderr, wing2);
	fprintf(stderr, " -> ");
	PrintWingShort(stderr, owing);
	fprintf(stderr, "\n");
#endif
	}
      }
      wing2 = owing;
    }
  } else {
#ifdef DEBUG
    fprintf(stderr, "file %s line %d: before BrepMakeEdgeJoinContours():\n", __FILE__, __LINE__);
#endif
#ifdef DDEBUG
    fprintf(stderr, "\twing1->prev = "); PrintWingShort(stderr, wing1->prev); fprintf(stderr, "\n");
    fprintf(stderr, "\twing1       = "); PrintWingShort(stderr, wing1); fprintf(stderr, "\n");
    fprintf(stderr, "\twing2->prev = "); PrintWingShort(stderr, wing2->prev); fprintf(stderr, "\n");
    fprintf(stderr, "\twing2       = "); PrintWingShort(stderr, wing2); fprintf(stderr, "\n");
#endif
    wing = BrepMakeEdgeJoinContours(wing1, wing2, &edge_data);
#ifdef DEBUG
    fprintf(stderr, "file %s line %d: new wings:\n", __FILE__, __LINE__);
    PrintWingShort(stderr, wing);
    PrintWingShort(stderr, BrepEdgeOtherWing(wing));
    fprintf(stderr, "\n");
#endif
#ifdef DDEBUG
    fprintf(stderr, "file %s line %d: after BrepMakeEdgeSplitContour():\n", __FILE__, __LINE__);
    fprintf(stderr, "\t  wing->prev = "); PrintWingShort(stderr, wing->prev); fprintf(stderr, "\n");
    fprintf(stderr, "\t  wing       = "); PrintWingShort(stderr, wing); fprintf(stderr, "\n");
    fprintf(stderr, "\t  wing->next = "); PrintWingShort(stderr, wing->next); fprintf(stderr, "\n");
    fprintf(stderr, "\to wing->prev = "); PrintWingShort(stderr, BrepEdgeOtherWing(wing)->prev); fprintf(stderr, "\n");
    fprintf(stderr, "\to wing       = "); PrintWingShort(stderr, BrepEdgeOtherWing(wing)); fprintf(stderr, "\n");
    fprintf(stderr, "\to wing->next = "); PrintWingShort(stderr, BrepEdgeOtherWing(wing)->next); fprintf(stderr, "\n");
#endif
#ifdef DDEBUG
    fprintf(stderr, "file %s line %d: checking wing1->contour:\n", __FILE__, __LINE__);
    CheckContour(wing1->contour);    
    fprintf(stderr, "file %s line %d: checking vertices:\n", __FILE__, __LINE__);
    CheckVertex(wing1->vertex);
    CheckVertex(wing2->vertex);
#endif
  }

  MarkContours(wing);
  return wing2;		/* the possibly updated wing leaving at the exit endpoint */
}

/* Introduces a notch from the starting vertex of wing1 to v2. Assumes that the
 * starting vertex of wing1 comes before v2 on the intersection line. */
static BREP_WING *MakeNotch1(BREP_WING *wing1, BREP_VERTEX *v2)
{
  EDGE edge_data;
  BREP_WING *wing;

  edge_data.id = 0;
#ifdef DEBUG
  fprintf(stderr, "MakeNotch1:\n");
  fprintf(stderr, "\twing1       = "); PrintWingShort(stderr, wing1); fprintf(stderr, "\n");
  fprintf(stderr, "\tv2          = "); PrintVertexData(stderr, v2->client_data); fprintf(stderr, "\n");
#endif
  wing = BrepMakeNotch(wing1, v2, &edge_data);
#ifdef DEBUG
    fprintf(stderr, "file %s line %d: new wings:\n", __FILE__, __LINE__);
    PrintWingShort(stderr, wing);
    PrintWingShort(stderr, BrepEdgeOtherWing(wing));
    fprintf(stderr, "\n");
#endif
#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: checking contour:\n", __FILE__, __LINE__);
  CheckContour(wing1->contour);
  fprintf(stderr, "file %s line %d: checking vertices:\n", __FILE__, __LINE__);
  CheckVertex(wing1->vertex);
  CheckVertex(v2);
#endif

  MarkContours(wing);
  return wing->next;	/* wing leaving at (exit endpoint) v2 */
}

/* Introduces a notch from v1 to the starting vertex of wing2. Assumes that v1
 * comes before the starting vertex of wing2 on the intersection line. */
static BREP_WING *MakeNotch2(BREP_VERTEX *v1, BREP_WING *wing2)
{
  EDGE edge_data;
  BREP_WING *wing;

  edge_data.id = 0;
#ifdef DEBUG
  fprintf(stderr, "MakeNotch2:\n");
  fprintf(stderr, "\tv1          = "); PrintVertexData(stderr, v1->client_data); fprintf(stderr, "\n");
  fprintf(stderr, "\twing2       = "); PrintWingShort(stderr, wing2); fprintf(stderr, "\n");
#endif
  wing = BrepMakeNotch(wing2, v1, &edge_data);
#ifdef DEBUG
    fprintf(stderr, "file %s line %d: new wings:\n", __FILE__, __LINE__);
    PrintWingShort(stderr, wing);
    PrintWingShort(stderr, BrepEdgeOtherWing(wing));
    fprintf(stderr, "\n");
#endif
#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: checking contour:\n", __FILE__, __LINE__);
  CheckContour(wing2->contour);
  fprintf(stderr, "file %s line %d: checking vertices:\n", __FILE__, __LINE__);
  CheckVertex(v1);
  CheckVertex(wing2->vertex);
#endif

  MarkContours(wing);
  return wing2;
}

/* Introduces a slit in the face between the two vertices. Assumes that v1 comes sbefore
 * v2 on the intersection line. */
static BREP_WING *MakeSlit(BREP_FACE *face, BREP_VERTEX *v1, BREP_VERTEX *v2)
{
  BREP_WING *wing;
  EDGE edge_data;
  CONTOUR contour_data;

  edge_data.id = 0;
  contour_data.id = 0;
#ifdef DEBUG
  fprintf(stderr, "MakeSlit:\n");
  fprintf(stderr, "\tv1          = "); PrintVertexData(stderr, v1->client_data); fprintf(stderr, "\n");
  fprintf(stderr, "\tv2          = "); PrintVertexData(stderr, v2->client_data); fprintf(stderr, "\n");
#endif
  wing = BrepMakeSlit(face, v1, v2, &edge_data, &contour_data);
#ifdef DEBUG
    fprintf(stderr, "file %s line %d: new wings:\n", __FILE__, __LINE__);
    PrintWingShort(stderr, wing);
    PrintWingShort(stderr, BrepEdgeOtherWing(wing));
    fprintf(stderr, "\n");
#endif
#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: checking face:\n", __FILE__, __LINE__);
  CheckFace(face);
  fprintf(stderr, "file %s line %d: checking vertices:\n", __FILE__, __LINE__);
  CheckVertex(v1);
  CheckVertex(v2);
#endif

  MarkContours(wing);
  return wing->next;	/* wing leaving at (exit endpoint) v2 going to v1 */
}

/* The routines Cut(), Notch() and Slit() introduce a cut, notch or slit in
 * a face between endpoints p1 and p2. They return a BREP_WING leaving at the exit end
 * of the cut, notch or slit, i.e. with highest parameter along the intersection
 * line. This vertex may be the start vertex of a next cut, notch or slit. */

/* Both endpoints are contributed by the face. This routine will insert a cut in
 * the face. Assumes that p1 preceeds p2, i.o.w. p1 is the entry endpoint and p2
 * is the exit endpoint. */
static BREP_WING *Cut(BREP_FACE *face, ENDPOINT *p1, ENDPOINT *p2)
{
  BREP_WING *wing1, *wing2;

#ifdef DEBUG
  fprintf(stderr, "--- Cut:\n");
  fprintf(stderr, "face = "); PrintFaceData(stderr, face->client_data); fprintf(stderr, "\n");
  fprintf(stderr, "p1 = "); PrintEndpointData(stderr, p1);
  fprintf(stderr, "p2 = "); PrintEndpointData(stderr, p2);
#endif
  wing1 = MakeVertexSplitEdge(p1);
  if (wing1 == p2->wing) {	
    /* the point at which p1->wing was to be split was its end vertex, so 
     * MakeVertexSplitEdge(p1) returns p2->wing. */
#ifdef DDEBUG
    fprintf(stderr, "file %s line %d: before MakeVertexSplitEdge():\n", __FILE__, __LINE__);
#endif
    wing2 = MakeVertexSplitEdge(p2);
#ifdef DDEBUG
    fprintf(stderr, "file %s line %d: checking wing2->contour:\n", __FILE__, __LINE__);
    CheckContour(wing2->contour);    
    fprintf(stderr, "file %s line %d: checking vertices:\n", __FILE__, __LINE__);
    CheckVertex(wing2->vertex);
    CheckVertex(wing2->next->vertex);
    CheckVertex(wing2->prev->vertex);
#endif
    return wing2;
  }
  wing2 = MakeVertexSplitEdge(p2);

  if (p1->ti < p2->ti)
    return MakeCut(wing1, wing2);
  else
    return MakeCut(wing2, wing1);
}

/* Only the first endpoint was contributed by the face, the second not. This
 * routine inserts a notch in the face, splitting p1->wing in two at the
 * intersection point and connecting to the new vertex defined by p2. */
static BREP_WING *Notch(BREP_FACE *face, ENDPOINT *p1, ENDPOINT *p2)
{
  BREP_WING *wing;
  BREP_VERTEX *v;
#ifdef DEBUG
  fprintf(stderr, "--- Notch:\n");
  fprintf(stderr, "face = "); PrintFaceData(stderr, face->client_data); fprintf(stderr, "\n");
  fprintf(stderr, "p1 = "); PrintEndpointData(stderr, p1);
  fprintf(stderr, "p2 = "); PrintEndpointData(stderr, p2);
#endif
  wing = MakeVertexSplitEdge(p1);
  v = MakeVertex(face, p2);

  if ((lastwingA && v == lastwingA->vertex) || 
      (lastwingB && v == lastwingB->vertex)) {
    /* a notch to the end vertex of the previous cutting interval cuts the
     * contour. */
    BREP_WING *wing2 = ((lastwingA && v == lastwingA->vertex) ? lastwingA : lastwingB);
    if (p1->ti < p2->ti)
      return MakeCut(wing, wing2);
    else
      return MakeCut(wing2, wing);
  }

  if (p1->ti < p2->ti)
    return MakeNotch1(wing, v);
  else
    return MakeNotch2(v, wing);
}

/* None of the endpoints was contributed by the face. This routine inserts a slit 
 * in the face. It is assumed that p1 preceeds p2 on the intersection line, i.o.w.
 * p1 is the entry endpoint and p2 is the exit endpoint. */
static BREP_WING *Slit(BREP_FACE *face, ENDPOINT *p1, ENDPOINT *p2)
{
  BREP_VERTEX *v1, *v2;
#ifdef DEBUG
  fprintf(stderr, "--- Slit:\n");
  fprintf(stderr, "face = "); PrintFaceData(stderr, face->client_data); fprintf(stderr, "\n");
  fprintf(stderr, "p1 = "); PrintEndpointData(stderr, p1);
  fprintf(stderr, "p2 = "); PrintEndpointData(stderr, p2);
#endif
  v1 = MakeVertex(face, p1);
  v2 = MakeVertex(face, p2);

  if ((lastwingA && v1 == lastwingA->vertex) || 
      (lastwingB && v1 == lastwingB->vertex)) {
    /* a slit including the endpoint of a previous cutting interval is a notch */
    BREP_WING *wing1 = ((lastwingA && v1 == lastwingA->vertex) ? lastwingA : lastwingB);
    return MakeNotch1(wing1, v2); 
  }

  return MakeSlit(face, v1, v2);
}

/* This routine inserts cuts, notches and slits in the faces as indicated by
 * the endpoint multiplicities. There is at least one and at most two entry or exit 
 * endpoints for a cutting interval: at least from one and at most from both faces. 
 * If there is both an entry and exit endpoint for a given face, the face needs
 * to be cut. If there is only one entry or exit endpoint for a face, a notch 
 * needs to be inserted. If a face didn't contribute neither an entry and neither
 * an exit endpoint, a slit needs to be inserted in the face (In that case, the other
 * face has contributed both entry and exit endpoint and will be cut). Phantom 
 * endpoints should come in pairs. If not, there is something wrong with the endpoint 
 * determination or sorting. A pair of phantom endpoints induces a cut, notch or slit in
 * the face that didn't contribute the endpoints, but the face that contributed them
 * is left intact. This routine is called with the last encountered entry and
 * first encountered exit points on both faces at the start and end of a cutting 
 * interval. */
static void CuttingInterval(ENDPOINT *entryA, ENDPOINT *entryB,
			    ENDPOINT *exitA, ENDPOINT *exitB)
{
  ENDPOINT *ci_startA, *ci_startB, *ci_endA, *ci_endB; /* entry and exit endpoints of
							* the cutting interval */
  ENDPOINT *entry, *exit;
#ifdef DEBUG
  fprintf(stderr, "--- CuttingInterval:\n");
  fprintf(stderr, "entryA = "); PrintEndpointData(stderr, entryA);
  fprintf(stderr, "entryB = "); PrintEndpointData(stderr, entryB);
  fprintf(stderr, "exitA = "); PrintEndpointData(stderr, exitA);
  fprintf(stderr, "exitB = "); PrintEndpointData(stderr, exitB);
#endif
  /* check if an entry and exit point is given for both faces. */
  if (!entryA || !entryB || !exitA || !exitB) {
    Warning("CuttingInterval", "missing endpoints");
    return;
  }

  /* Compare the last encountered entry points on face A and face B:
   * if they are at the same position, keep both as starting
   * points of a cutting interval. If not, save only the one with largest
   * parameter along the intersection line. */
  ci_startA = ci_startB = (ENDPOINT *)NULL;
  if (entryA->ti >= entryB->ti) 
    ci_startA = entryA;  
  if (entryB->ti >= entryA->ti) 
    ci_startB = entryB;
  entry = (ci_startA ? ci_startA : ci_startB);

  /* Compare the endpoints. Keep both if they are at the same position, or only
   * the one at with smallest parameter along the intersection line if different. */
  ci_endA = ci_endB = (ENDPOINT *)NULL;
  if (exitA->ti <= exitB->ti) 
    ci_endA = exitA;
  if (exitB->ti <= exitA->ti) 
    ci_endB = exitB;
  exit = (ci_endA ? ci_endA : ci_endB);

#ifdef DEBUG
  fprintf(stderr, "ci_startA = "); PrintEndpointData(stderr, ci_startA);
  fprintf(stderr, "ci_startB = "); PrintEndpointData(stderr, ci_startB);
  fprintf(stderr, "ci_endA = "); PrintEndpointData(stderr, ci_endA);
  fprintf(stderr, "ci_endB = "); PrintEndpointData(stderr, ci_endB);
#endif

  if (CoincidingEndpoints(entry, exit)) {
    /* don't create zero length features */
#ifdef DEBUG
    fprintf(stderr, "Coinciding endpoints ... \n");
#endif
    return;
  }

  if (entry->ti > exit->ti) {
    Warning("CuttingInterval", "cutting interval determination incorrect");
    return;
  }

  /* check if phantom endpoints come in pairs */
  if ((entryA->type & PHANTOM_ENDPOINT) != (exitA->type & PHANTOM_ENDPOINT) ||
      (entryB->type & PHANTOM_ENDPOINT) != (exitB->type & PHANTOM_ENDPOINT)) {
    Warning("CuttingInterval", "phantom endpoints should come in pairs");
  }

  /* if not a pair of phantom endpoints, it must be a pair of
   * one entry and one exit endpoint. */
  else if (!(entryA->type & ENTRY_ENDPOINT) != !(exitA->type & EXIT_ENDPOINT) ||
	   !(entryB->type & ENTRY_ENDPOINT) != !(exitB->type & EXIT_ENDPOINT)) {
    Warning("CuttingInterval", "entry/exit property of endpoints incorrectly determined");
  }

#ifdef CHECK
  if (check) { 
#ifdef DEBUG
    fprintf(stderr, "file %s line %d: just checking ...\n", __FILE__, __LINE__);
#endif
    return;
  }
#endif /*CHECK*/

  /* partition face A */
  if (!(entryA->type & PHANTOM_ENDPOINT) && !(exitA->type & PHANTOM_ENDPOINT)) {
    if (ci_startA && ci_endA) 
      lastwingA = Cut(faceA, ci_startA, ci_endA);
    else if (ci_startA) 
      lastwingA = Notch(faceA, ci_startA, ci_endB);
    else if (ci_endA) 
      lastwingA = Notch(faceA, ci_endA, ci_startB);
    else 
      lastwingA = Slit(faceA, ci_startB, ci_endB);
  } 

  /* partition face B */
  if (!(entryB->type & PHANTOM_ENDPOINT) && !(exitB->type & PHANTOM_ENDPOINT)) {
    if (ci_startB && ci_endB) 
      lastwingB = Cut(faceB, ci_startB, ci_endB);
    else if (ci_startB) 
      lastwingB = Notch(faceB, ci_startB, ci_endA);
    else if (ci_endB) 
      lastwingB = Notch(faceB, ci_endB, ci_startA);
    else 
      lastwingB = Slit(faceB, ci_startA, ci_endA);
  } 

  lastti = exit->ti;
#ifdef DEBUG
  fprintf(stderr, "file %s line %d: after cutting:\n", __FILE__, __LINE__);
  fprintf(stderr, "\tlastwingA = "); PrintWingShort(stderr, lastwingA); fprintf(stderr, "\n");
  fprintf(stderr, "\tlastwingB = "); PrintWingShort(stderr, lastwingB); fprintf(stderr, "\n");
  fprintf(stderr, "\tlastti = %f\n", lastti);
#endif
}

/* Traverses the intersection line, determines entry and exit endpoints of
 * cutting intervals, and processes cutting intervals. */
static void TraverseIntersectionline(void)
{
  int i, j,
      inA, inB, /* flags indicating whether we are inside/outside faceA and faceB */
      incut; 	/* flag indicating whether we are currently in a cutting interval */
  ENDPOINT *endpoint, *endq,
           *entryA, *entryB, *exitA, *exitB, 	/* last encountered entry and exit
						 * endpoint on face A and B */
           *ci_startA, *ci_startB, *ci_endA, *ci_endB; /* potential entry and
						  * exit endpoint for cutting interval */
#ifdef DEBUG
  fprintf(stderr, "TraverseIntersectionline:\n");
#endif
  /* Initialize intersection line parameter of and wings in faceA and faceB 
   * leaving at the exit endpoint of the previous cutting interval. */
  lastwingA = lastwingB = (BREP_WING *)NULL;
  lastti = -HUGE;

  entryA = exitA = entryB = exitB = (ENDPOINT *)NULL;
  ci_startA = ci_startB = ci_endA = ci_endB = (ENDPOINT *)NULL;
  inA = inB = incut = FALSE;

  /* loop over the sorted endpoints */
  for (i=0, endpoint = endpoints; i<nr_endpoints; i++, endpoint++) {
#ifdef DEBUG
    fprintf(stderr, "--- i=%d ", i);
    PrintEndpointData(stderr, endpoint);
#endif
    /* skip deleted endpoints or purely VERTEX endpoints, which are just there
     * for alignement. */
    if ((endpoint->type & DELETED_ENDPOINT) || (endpoint->type == VERTEX_ENDPOINT))
      continue;

    /* flip the state of the inA or inB flag, update the last encountered entry or
     * exit endpoint for the face accordingly. Forget the last exit endpoint when
     * an entry endpoint is found and vice versa. */
    if (endpoint->wing->contour->face == faceA) {
      inA = 1 - inA; 
      if (inA) {
	entryA = endpoint;
	exitA = (ENDPOINT *)NULL;
      } else {
	entryA = (ENDPOINT *)NULL;
	exitA = endpoint;
      }
    } else if (endpoint->wing->contour->face == faceB) {
      inB = 1 - inB;
      if (inB) {
	entryB = endpoint;
	exitB = (ENDPOINT *)NULL;
      } else {
	entryB = (ENDPOINT *)NULL;
	exitB = endpoint;
      }
    } else {
      /* this should never occur. */
      Warning("TraverseIntersectionline", "endpoint present from unknown face");
      continue;	/* ignore it */
    }

#ifdef DEBUG
    fprintf(stderr, "inA = %d, inB = %d, incut = %d.\n", inA, inB, incut);
#endif
    if (entryA && entryB) { 	/* start of a new cutting interval */
      ci_startA = entryA;
      ci_startB = entryB;
      incut = TRUE;
    } else if (incut && (exitA || exitB)) {	/* end of cutting interval */
      /* look ahead to find the first exit endpoint induced by the other face */
      ci_endA = exitA;
      for (j=i+1; !ci_endA && j<nr_endpoints; j++) {
	endq = &endpoints[j];
	if (endq->wing->contour->face == faceA &&
	    ((endq->type & PHANTOM_ENDPOINT) || (endq->type & EXIT_ENDPOINT)))
	  ci_endA = endq;
      }

      ci_endB = exitB;
      for (j=i+1; !ci_endB && j<nr_endpoints; j++) {
	endq = &endpoints[j];
	if (endq->wing->contour->face == faceB &&
	    ((endq->type & PHANTOM_ENDPOINT) || (endq->type & EXIT_ENDPOINT)))
	  ci_endB = endq;
      }

      CuttingInterval(ci_startA, ci_startB, ci_endA, ci_endB);
      incut = FALSE;
    }
  }

  /* after traversal of the intersection line, we should be outside of both
   * faces again. */
  if (inA || inB) {
    Warning("TraverseIntersectionline", "something wrong with endpoint determination");
  }
}

/* Partitions a pair of possibly transversal intersecting faces along the intersection
 * line. */
static void TransversalIntersection(BREP_FACE *_faceA, BREP_WING *first_wingA, 
				    BREP_FACE *_faceB, BREP_WING *first_wingB)
{
#ifdef DEBUG
  fprintf(stderr, "-----> TransversalIntersection():\n");
#endif
  faceA = _faceA;	/* global variables not to have to pass them as a */
  faceB = _faceB;	/* parameter every time. */

  /* compute the parametric equation t.U + V and tolerance of the face 
   * intersection line */
  isect = ComputePlaneToPlaneIntersection(faceA->client_data, faceB->client_data);

  /* determine cutting interval endpoints = intersections of an edge of one
   * face with the plane of the other */
  nr_endpoints = 0;
  timinA = timinB = HUGE;
  timaxA = timaxB = -HUGE;
  DetermineEndpoints(faceA, first_wingA, faceB);
  DetermineEndpoints(faceB, first_wingB, faceA);

  /* if all endpoints of one face come before the endpoints contributed by the 
   * other face, there is no intersection. */
  if (timinA > timaxB || timinB > timaxA) {
#ifdef DEBUG
    fprintf(stderr, "timinA=%f > timaxB=%f || timinB=%f > timaxA=%f -> no intersection\n",
	    timinA, timaxB, timinB, timaxA);
#endif
    return;
  }

  /* sort endpoint list */
  SortEndpoints();

  /* traverse intersection line, insert cuts, notches and slits in the faces
   * as indicated by the endpoints */
#ifdef CHECK
  check = TRUE;
  TraverseIntersectionline();
  check = FALSE;
#endif /*CHECK*/
  TraverseIntersectionline();
}

/* faceA and faceB are two faces for which the coplanarity factor exceeds the
 * allowed minimum. This routine will return TRUE if faceA really can 
 * be considered to be within the plane of faceB:  that is, when
 * two vertices in faceA lies on opposite sides of the plane of faceB, or
 * when any vertex of faceA lies within a distance equal to the extent of faceB. */
static int TestCoplanarIntersection(BREP_FACE *faceA, BREP_FACE *faceB)
{
  BREP_CONTOUR *contour;
  BREP_WING *wing;
  int side, prevside;
#ifdef DEBUG
  fprintf(stderr, "-----> TestCoplanarIntersection:\n");
#endif

  side = UNKNOWN_SIDE;	/* some value <> 0 */

  /* check the signed distance of all vertices of faceA to the plane of faceB */
  contour = faceA->outer_contour;
  if (!contour)	/* face without contours */
    return FALSE; /* ignore it */
  do {
    if (contour->wings) {
      wing = contour->wings;
      SignedDistanceToPlane(wing->vertex->client_data, faceB->client_data, &side);
      if (side == 0)	/* vertex within the extent of the plane of faceB */
	return TRUE;
      wing = wing->next;

      while (wing != contour->wings) {
	prevside = side;
	SignedDistanceToPlane(wing->vertex->client_data, faceB->client_data, &side);

	if (side != prevside || side == 0) {
	  /* the starting vertex of wing was on the other side of the plane of 
	   * faceB than the starting vertex of wing->prev or the vertex was within the
	   * extent of the plane of faceB. */
	  return TRUE;
	}

	wing = wing->next;
      }
      /* there's no need to check the last wing: if the starting vertex of the
       * last wing is on the other side of the starting vertex of the first,
       * there is a crossing in one of the other wings as well. */
    }

    contour = contour->next;
  } while (contour != faceA->outer_contour);

  return FALSE;
}

/* faceA and faceB are two faces for which the coplanarity factor is lower than the
 * minimum allowed coplanarity factor. This routine classifies faceA w.r.t. faceB:
 * If an edge of faceA has endpoints on different
 * sides of the plane of faceB, a pointer to the crossing edge is returned.
 * If faceA is entirely on the same side of the plane of faceB NULL is returned.
 * Two faces faceA and faceB can only intersect transversally if both faces
 * have an edge that crosses the plane of the other. */
static BREP_WING *TestTransversalIntersection(BREP_FACE *faceA, BREP_FACE *faceB)
{
  BREP_CONTOUR *contour;
  BREP_WING *wing;
  int side, prevside;

#ifdef DEBUG
  fprintf(stderr, "-----> TestTransversalIntersection:\n");
#endif
  side = UNKNOWN_SIDE;	/* some value <> 0 */

  /* check the signed distance of all vertices of faceA to the plane of faceB */
  contour = faceA->outer_contour;
  if (!contour)	/* face without contours */
    return (BREP_WING *)NULL;	/* ignore it */
  do {
    if (contour->wings) {
      wing = contour->wings;
      SignedDistanceToPlane(wing->vertex->client_data, faceB->client_data, &side);
      if (side == 0)
	return wing;
      wing = wing->next;

      while (wing != contour->wings) {
	prevside = side;
	SignedDistanceToPlane(wing->vertex->client_data, faceB->client_data, &side);
	if (side == 0 || side != prevside) {
	  /* the starting vertex of wing was on the other side of the plane of 
	   * faceB than the starting vertex of wing->prev */
	  return wing->prev;
	}
	wing = wing->next;
      }
      /* there's no need to check the last wing: if the starting vertex of the
       * last wing is on the other side of the starting vertex of the first,
       * there is a crossing in one of the other wings as well. */
    }
    contour = contour->next;
  } while (contour != faceA->outer_contour);

  return (BREP_WING *)NULL;	/* no need to classify faceB w.r.t. to faceA */
}

/* in triangulate.c */
extern float FaceShortestEdgeLength(BREP_FACE *face);
extern float minedgelength;

/* Given two BREP_FACES, with appropriate geometric data (see geometry.[ch]), this
 * routine splits the faces into nonintersecting parts. Plane equations and tolerances 
 * have been computed when reading in the objects, in CloseFaceCallback(). */
int PartitionFaces(BREP_FACE *faceA, BREP_FACE *faceB)
{
  BREP_WING *wingA = (BREP_WING *)NULL, *wingB = (BREP_WING *)NULL;
#ifdef DEBUG
fprintf(stderr, "=====> PartitionFaces:\n");
fprintf(stderr, "\tfaceA = "); PrintFaceData(stderr, faceA->client_data); fprintf(stderr, "\n");
fprintf(stderr, "\tfaceB = "); PrintFaceData(stderr, faceB->client_data); fprintf(stderr, "\n");
fprintf(stderr, "Coplanarity factor = %g\n", CoplanarityFactor(faceA->client_data, faceB->client_data));
#endif

  if (FaceShortestEdgeLength(faceA) < minedgelength ||
      FaceShortestEdgeLength(faceB) < minedgelength)
    return NO_ISECT;	/* do not consider small faces */

  if (CoplanarityFactor(faceA->client_data, faceB->client_data) > min_coplanarity_factor) {
    if (TestCoplanarIntersection(faceA, faceB) || 
	TestCoplanarIntersection(faceB, faceA))
      return COPLANAR_ISECT;
  } else {
    if ((wingA = TestTransversalIntersection(faceA, faceB)) && 
	(wingB = TestTransversalIntersection(faceB, faceA)))
      TransversalIntersection(faceA, wingA, faceB, wingB);
    return TRANSVERSAL_ISECT;
  }

  return NO_ISECT;
}

