/* main.c */
/*
  The BREP Library.
  Copyright (C) 1996 Philippe Bekaert

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "pools.h"
#include "brep.h"
#include "geometry.h"
#include "readmgf.h"
#include "writemgf.h"
#include "isect.h"
#include "error.h"
#include "raytrace.h"
#include "vector.h"
#include "globals.h"
#include "brepit.h"

clock_t start_clock;
int verbose;	/* to print statistics etc ... */
void (*csg_select)(BREP_SOLID *, BREP_SOLID *);
BREP_SOLID *solidA, *solidB;
char *progname;
int ignore_normals_on_input; 	/* ignore normals in input file */
int ignore_materials;		/* ignore materials when outputting MGF */

/* the algorithm requires to remember coplanar face pairs until partitioning
 * and inside/outside mark propagation has been done. */
typedef struct COPLANAR_FACELIST {
  BREP_FACE *faceA, *faceB;
  struct COPLANAR_FACELIST *next;
} COPLANAR_FACELIST;

static COPLANAR_FACELIST *coplanar_faces = (COPLANAR_FACELIST *)NULL;

#ifndef NOPOOLS
static POOL *coplanar_facePool = (POOL *)NULL;
#define NEW_COPLANAR_FACEPAIR()  	(COPLANAR_FACELIST *)New(sizeof(COPLANAR_FACELIST), &coplanar_facePool)
#define DISPOSE_COPLANAR_FACEPAIR(ptr) Dispose((unsigned char *)(ptr), &coplanar_facePool)
#else /*NOPOOLS*/
#define NEW_COPLANAR_FACEPAIR()	(COPLANAR_FACELIST *)Alloc(sizeof(COPLANAR_FACELIST))
#define DISPOSE_COPLANAR_FACEPAIR(ptr) Free((char *)ptr, sizeof(COPLANAR_FACEPAIR))
#endif /*NOPOOLS*/

/* record a pair of coplanar faces */
static void RecordCoplanarFacepair(BREP_FACE *faceA, BREP_FACE *faceB)
{
  COPLANAR_FACELIST *pair;

  pair = NEW_COPLANAR_FACEPAIR();
  pair->faceA = faceA;
  pair->faceB = faceB;
  pair->next = coplanar_faces;
  coplanar_faces = pair;
}

/* Given a pair of coplanar faces, IterateMatchingContours() calls a given
 * function for each pair of matching contours. Two contours match after 
 * partitioning as soon as they have one wing connecting the same vertices 
 * and their enclosed region is on the same side. 
 * See M. Segal, IEEE CG&A Jan '88 p 61-62 */
static void (*matching_contour_func)(BREP_CONTOUR *, BREP_CONTOUR *);
static BREP_CONTOUR *matching_contourA;
static BREP_FACE *matching_contour_faceB;

static void do_matching_contours(BREP_WING *wingB)
{
  BREP_CONTOUR *contourB = wingB->contour;

  if (!contourB || contourB->face != matching_contour_faceB)
    return;

  /* wingB has endpoints at the same location as the first wing of 
   * matching_contourA. The contours match if the enclosed region is on the same
   * side, i.o.w. the first vertex of wingB is at the same location as the 
   * first vertex of the first wing of matching_contourA. */
  if (VertexCompareLocation(matching_contourA->wings->vertex->client_data, wingB->vertex->client_data) == 8)
    matching_contour_func(matching_contourA, contourB);
}

static void IterateMatchingContours(BREP_FACE *faceA, BREP_FACE *faceB, 
				    void (*func)(BREP_CONTOUR *contourA, BREP_CONTOUR *contourB))
{
  BREP_CONTOUR *contourA;
  void (*old_matching_contour_func)(BREP_CONTOUR *, BREP_CONTOUR *) = matching_contour_func;
  matching_contour_func = func;
  matching_contour_faceB = faceB;

  for (contourA = faceA->outer_contour; contourA; 
       contourA = (contourA->next == faceA->outer_contour ? NULL : contourA->next)) {
    if (contourA->wings && contourA->wings->next != contourA->wings->prev) {
      matching_contourA = contourA;
      BrepIterateWingsBetweenLocations(contourA->wings->vertex->client_data, 
				       contourA->wings->next->vertex->client_data,
				       faceB->shell->solid->vertices,
				       do_matching_contours);
    }
  }

  matching_contour_func = old_matching_contour_func;
}

/* clears the contour masks OUTMASK, INMASK, WRITEMASK, DONTWRITEMASK, WRITEREVERSEMASK */
static void unset_mark(BREP_CONTOUR *contour)
{
  ((CONTOUR *)(contour->client_data))->id &= ~(INMASK|OUTMASK|WRITEMASK|DONTWRITEMASK|WRITEREVERSEMASK);
}

/* returns the inside/outside mask of the contour */
static unsigned int inoutmark(BREP_CONTOUR *contour)
{
  return ((CONTOUR *)(contour->client_data))->id & (INMASK|OUTMASK);
}

/* marks a contour with the given mask */
static void mark_contour(BREP_CONTOUR *contour, unsigned int mask)
{
#ifdef DDEBUG
  fprintf(stderr, "mark_contour: contour = "); 
  PrintContourData(stderr, contour->client_data);
  fprintf(stderr, ", mask = %s%s%s%s%s\n", 
	  mask & OUTMASK ? "OUTMASK" : "",
	  mask & INMASK ? "INMASK" : "",
	  mask & WRITEMASK ? "WRITEMASK" : "",
	  mask & DONTWRITEMASK ? "DONTWRITEMASK" : "",
	  mask & WRITEREVERSEMASK ? "WRITEREVERSEMASK" : "");
#endif
  ((CONTOUR *)(contour->client_data))->id |= mask;
}

/* contourA and contourB are matching contours in coplanar faces. Keep one
 * if they have the same orientation. */
static void keep_one_same(BREP_CONTOUR *contourA, BREP_CONTOUR *contourB)
{
#ifdef DEBUG
  fprintf(stderr, "keep_one_same: contourA = ");
  PrintContourData(stderr, contourA->client_data);
  fprintf(stderr, ", contourB = ");
  PrintContourData(stderr, contourB->client_data);
  fprintf(stderr, "\n");
#endif
  if (VECTORDOTPRODUCT(((CONTOUR *)(contourA->client_data))->normal,
		       ((CONTOUR *)(contourB->client_data))->normal) > 0.) {
    /* same orientation */
    unset_mark(contourA);
    mark_contour(contourA, WRITEMASK);
    unset_mark(contourB);
    mark_contour(contourB, DONTWRITEMASK);
  } else {
    /* opposite orientation */
    unset_mark(contourA);
    mark_contour(contourA, DONTWRITEMASK);
    unset_mark(contourB);
    mark_contour(contourB, DONTWRITEMASK);
  }
}

/* contourA and contourB are matching contours in coplanar faces. Keep one
 * if they have opposite orientation. */
static void keep_one_opposite(BREP_CONTOUR *contourA, BREP_CONTOUR *contourB)
{
#ifdef DEBUG
  fprintf(stderr, "keep_one_opposite: contourA = ");
  PrintContourData(stderr, contourA->client_data);
  fprintf(stderr, ", contourB = ");
  PrintContourData(stderr, contourB->client_data);
  fprintf(stderr, "\n");
#endif
  if (VECTORDOTPRODUCT(((CONTOUR *)(contourA->client_data))->normal,
		       ((CONTOUR *)(contourB->client_data))->normal) > 0.) {
    /* same orientation */
    unset_mark(contourA);
    mark_contour(contourA, DONTWRITEMASK);
    unset_mark(contourB);
    mark_contour(contourB, DONTWRITEMASK);
  } else {
    /* opposite orientation */
    unset_mark(contourA);
    mark_contour(contourA, WRITEMASK);
    unset_mark(contourB);
    mark_contour(contourB, DONTWRITEMASK);
  }
}

/* partition pairs of faces belonging to shellA and shellB */
static void partition_shells(BREP_SHELL *shellA, BREP_SHELL *shellB)
{
  BREP_FACE *faceA, *faceB;

  for (faceA = shellA->faces; faceA;
       faceA = (faceA->next == shellA->faces ? (BREP_FACE *)NULL : faceA->next)) {
    for (faceB = shellB->faces; faceB;
	 faceB = (faceB->next == shellB->faces ? (BREP_FACE *)NULL : faceB->next)) {
      if (!DisjunctFaceBounds(faceA, faceB))
	if (PartitionFaces(faceA, faceB) == COPLANAR_ISECT)
	  RecordCoplanarFacepair(faceA, faceB);
    }
  }
}

void do_isect(BREP_SOLID *solid)
{
  BREP_SHELL *shellA, *shellB;
  /*  COPLANAR_FACELIST *copl; */
  
  for (shellA = solid->shells; shellA; 
       shellA = (shellA->next == solid->shells ? (BREP_SHELL *)NULL : shellA->next)) {
    for (shellB = shellA->next; shellB != solid->shells; shellB = shellB->next) {
      if (!DisjunctShellBounds(shellA, shellB))
	partition_shells(shellA, shellB);
    }
  }

  /*  for (copl = coplanar_faces; copl; copl = copl->next)
    IterateMatchingContours(copl->faceA, copl->faceB, keep_one_same); */
}

/* inside/outside mark propagation: when a contour is cut during face partitioning,
 * the cut contour parts are marked whether they are inside or outside the transversal
 * solid. After partitioning, thi mark can be propagated to neighbouring contours,
 * reducing the number of contours for which an explicit inside/outside test will
 * be necessary. */
static void contour_propagate_mark(BREP_CONTOUR *contour);
static void wing_propagate_mark(BREP_WING *wing, unsigned int mark)
{
  BREP_CONTOUR *ocontour;

  ocontour = BrepEdgeOtherWing(wing)->contour;
  if (!ocontour || inoutmark(ocontour))
    return;

  mark_contour(ocontour, mark);
  contour_propagate_mark(ocontour); 
}

static void contour_propagate_mark(BREP_CONTOUR *contour)
{
  unsigned int mark = inoutmark(contour);

  if (mark)
    BrepContourIterateWings1A(contour, (void (*)(BREP_WING *, void *))wing_propagate_mark, (void *)mark);
}

/* Propagates inside/outside mark of contours for which it was determined 
 * during PartitionFaces() to neighbouring contours. */
static void csg_propagate_mark(BREP_SOLID *solid)
{
  BrepSolidIterateContours(solid, contour_propagate_mark);
}

/* inside/outside mark propagation only reduces the number of contours for which
 * an explicit inside/outside test will be necessary. After propagation, this
 * test is necessary for all contours that remained unmarked. */
static void contour_mark_unmarked(BREP_CONTOUR *contour, BREP_SOLID *othersolid)
{
  unsigned int mark;

  if (inoutmark(contour))
    return;

  if ((mark = ContourInSolid(contour, othersolid))) {
    mark_contour(contour, mark);  
    contour_propagate_mark(contour); 
  } else {
    /* The contour is most probably coplanar with a contour of the other solid
     * and will be handled later */
    Warning("contour_mark_unmarked", "can't determine whether contour is inside or outside other solid");
  }
}

static void csg_mark_unmarked(BREP_SOLID *solid, BREP_SOLID *othersolid)
{
  BrepSolidIterateContours1A(solid, (void (*)(BREP_CONTOUR *, void *))contour_mark_unmarked, (void *)othersolid);
}

/* mark a contour that is outside the transversal solid for writing */
static void mark_out(BREP_CONTOUR *contour)
{
  if (inoutmark(contour) & OUTMASK)
    mark_contour(contour, WRITEMASK);
  else
    mark_contour(contour, DONTWRITEMASK);
}

/* mark a contour that is inside the transversal solid for writing */
static void mark_in(BREP_CONTOUR *contour)
{
  if (inoutmark(contour) & INMASK)
    mark_contour(contour, WRITEMASK);
  else
    mark_contour(contour, DONTWRITEMASK);
}

/* mark a contour that is inside the transversal solid for reverse writing */
static void mark_in_reverse(BREP_CONTOUR *contour)
{
  if (inoutmark(contour) & INMASK)
    mark_contour(contour, WRITEREVERSEMASK);
  else
    mark_contour(contour, DONTWRITEMASK);
}

static void select_out(BREP_SOLID *solid)
{
  BrepSolidIterateContours(solid, mark_out);
}

static void select_in(BREP_SOLID *solid)
{
  BrepSolidIterateContours(solid, mark_in);
}

static void select_in_reverse(BREP_SOLID *solid)
{
  BrepSolidIterateContours(solid, mark_in_reverse);
}

static void select_union(BREP_SOLID *solidA, BREP_SOLID *solidB)
{
  COPLANAR_FACELIST *copl;

  select_out(solidA);
  select_out(solidB);

  for (copl = coplanar_faces; copl; copl = copl->next)
    IterateMatchingContours(copl->faceA, copl->faceB, keep_one_same);
}

static void select_difference(BREP_SOLID *solidA, BREP_SOLID *solidB)
{
  COPLANAR_FACELIST *copl;

  select_out(solidA);
  select_in_reverse(solidB);

  for (copl = coplanar_faces; copl; copl = copl->next)
    IterateMatchingContours(copl->faceA, copl->faceB, keep_one_opposite);
}

static void select_intersection(BREP_SOLID *solidA, BREP_SOLID *solidB)
{
  COPLANAR_FACELIST *copl;

  select_in(solidA);
  select_in(solidB);

  for (copl = coplanar_faces; copl; copl = copl->next)
    IterateMatchingContours(copl->faceA, copl->faceB, keep_one_same);
}

/* CSG main program */
static void do_csg(BREP_SOLID *solidA, BREP_SOLID *solidB)
{
  BREP_SHELL *shellA, *shellB;

  BrepSolidIterateContours(solidA, unset_mark);
  BrepSolidIterateContours(solidB, unset_mark);
  
  if (!DisjunctSolidBounds(solidA, solidB)) {
    for (shellA = solidA->shells; shellA; 
	 shellA = (shellA->next == solidA->shells ? (BREP_SHELL *)NULL : shellA->next)) {
      for (shellB = solidB->shells; shellB; 
	   shellB = (shellB->next == solidB->shells ? (BREP_SHELL *)NULL : shellB->next)) {
	if (!DisjunctShellBounds(shellA, shellB))
	  partition_shells(shellA, shellB);
      }
    }
  } else {
    BrepSolidIterateContours1A(solidA, (void (*)(BREP_CONTOUR *, void *))mark_contour, (void *)OUTMASK);
    BrepSolidIterateContours1A(solidB, (void (*)(BREP_CONTOUR *, void *))mark_contour, (void *)OUTMASK);
  }

  csg_propagate_mark(solidA); 
  csg_mark_unmarked(solidA, solidB); 
  csg_propagate_mark(solidB); 
  csg_mark_unmarked(solidB, solidA); 

  csg_select(solidA, solidB);
}

/* needed for including MGF files */
char *current_directory;

/* pools library error reporting routines */
static void do_pools_info(char *text)
{
  fprintf(stderr, "%s\n", text);
}

static void do_pools_fatal(char *text)
{
  Fatal(-1, NULL, text);
}

static void InitPools(void)
{
  /* uncomment the following line in order to obtain debug information from the 
   * pools library */
  /*  PoolsSetInfoCallback((POOLS_MSG_CALLBACK_FUNC)do_pools_info); */
  PoolsSetFatalCallback((POOLS_MSG_CALLBACK_FUNC)do_pools_fatal); 
}

/* used as CloseSolid callback to catch the solid read with readMGF() */
SOLID *CatchSolidA(BREP_SOLID *solid)
{
  solidA = solid;
  return SolidCloseCallback(solid);
}

SOLID *CatchSolidB(BREP_SOLID *solid)
{
  solidB = solid;
  return SolidCloseCallback(solid);
}

/* prints some statistics in verbose mode */
static void PrintStats(void)
{
  fprintf(stderr, "Nr. of solids   = %d\n", GetNumberOfSolids());
  fprintf(stderr, "Nr. of shells   = %d\n", GetNumberOfShells());
  fprintf(stderr, "Nr. of faces    = %d\n", GetNumberOfFaces());
  fprintf(stderr, "Nr. of contours = %d\n", GetNumberOfContours());
  fprintf(stderr, "Nr. of edges    = %d\n", GetNumberOfEdges());
  fprintf(stderr, "Nr. of vertices = %d\n", GetNumberOfVertices());
  fprintf(stderr, "Memory usage: %ld KBytes\n", GetMemoryUsage()/1024);
  fprintf(stderr, "Processing time: %.1f seconds\n", (float)(clock() - start_clock)/(float)CLOCKS_PER_SEC);
}

/* prints program usage */
static void PrintUsage(FILE *out)
{
  fprintf(out, "Usage: csg [options] fileA.mgf fileB.mgf\n\n");
  fprintf(out, "Options:\n\n");
  fprintf(out, "\t-q n: nr. quart circle divisions (n = positive integer, default = 5)\n");
  fprintf(out, "\t-t x: vertex tolerance (x = positive real number, default = 1e-6)\n");
  fprintf(out, "\t-p x: min. coplanarity factor (x = positive real number, default = 500)\n");
  fprintf(out, "\t-v: verbose mode\n");
  fprintf(out, "\t-d: compute CSG difference\n");
  fprintf(out, "\t-u: compute CSG union\n");
  fprintf(out, "\t-i: compute CSG intersection\n");
  fprintf(out, "\t-I: ignore normals in input file\n");
  fprintf(out, "\t-W: write faces without making separate faces for outer contours contained in holes\n");
  fprintf(out, "\t-M: don't write materials\n");
  fprintf(out, "\t-h: print usage\n\n");
}

/* sets defaults for various options */
static void Defaults(void)
{
  MGFSetNrQuartCircDivs(5);
  SetMinimumVertexTolerance(1e-6);
  SetMinimumEdgeTolerance(1e-12);
  SetMinimumFaceTolerance(1e-12);
  SetMinCoplanarityFactor(500.);
  verbose = FALSE;
  csg_select = select_union;
  ignore_normals_on_input = FALSE;
  MGFRespectVertexOrder(TRUE);
  write_as_is = FALSE;
  SetVertexCompareFlags(VCMP_LOCATION | VCMP_NORMAL | VCMP_NAME |
			VCMP_NO_NAME_IS_EQUAL_NAME);
  ignore_materials = FALSE;
}

/* parse command line options */
static int GetOptions(int argc, char **argv)
{
  int i, j, n;
  float x;

#define next_arg_parm {i++; j=1;}
#define next_arg_no_parm {if (argv[i][j+1]) {j++;} else {i++; j=1;}}

  for (i=1, j=1; i<argc && argv[i][0] == '-';) {
    switch (argv[i][j]) {
    case 'q':
      if ((sscanf(&argv[i][j+1], "%d", &n) != 1 &&
	  (++i >= argc || sscanf(argv[i], "%d", &n) != 1)) ||
	  n <= 0)
	Fatal(1, NULL, "option -q requires a positive integer argument");
      else
	MGFSetNrQuartCircDivs(n);
      next_arg_parm;
      break;
    case 't':
      if ((sscanf(&argv[i][j+1], "%f", &x) != 1 &&
	  (++i >= argc || sscanf(argv[i], "%f", &x) != 1)) ||
	  (x <= 0.))
	Fatal(1, NULL, "option -t requires a positive real argument");
      else
	SetMinimumVertexTolerance(x);      
      next_arg_parm;
      break;
    case 'p':
      if ((sscanf(&argv[i][j+1], "%f", &x) != 1 &&
	  (++i >= argc || sscanf(argv[i], "%f", &x) != 1)) ||
	  (x <= 0.))
	Fatal(1, NULL, "option -p requires a positive real argument");
      else
	SetMinCoplanarityFactor(x);      
      next_arg_parm;
      break;
    case 'v':
      verbose = TRUE;
      next_arg_no_parm;
      break;
    case 'd':
      csg_select = select_difference;
      next_arg_no_parm;
      break;
    case 'u':
      csg_select = select_union;
      next_arg_no_parm;
      break;
    case 'i':
      csg_select = select_intersection;
      next_arg_no_parm;
      break;
    case 'I':
      ignore_normals_on_input = TRUE;
      /*      MGFRespectVertexOrder(FALSE); */
      next_arg_no_parm;
      break;
    case 'W':
      write_as_is = TRUE;
      next_arg_no_parm;
      break;
    case 'M':
      ignore_materials = TRUE;
      next_arg_no_parm;
      break;
    case 'h':
    case '?':
      PrintUsage(stderr);
      exit(0);
    default:
      Fatal(1, NULL, "Invalid option '%c'.\n", argv[i][j]);
    }
  }

  return i;
}

int csg_main(int argc, char **argv)
{
  int i;

  Defaults();
  i = GetOptions(argc, argv);
  if (i >= argc-1) {
    fprintf(stderr, "Usage: %s [options] file1.mgf file2.mgf\n", progname);
    exit(1);
  } 

  start_clock = clock();

  InitPools();
  InitGeometry();

  /* read the objects in the MGF files */
  current_directory = "";

  solidA = (BREP_SOLID *)NULL;
  BrepSetCloseSolidCallback((BREP_CALLBACK_FUNC)CatchSolidA);
  ReadMGF(argv[i]);

  if (verbose) {
    fprintf(stderr, "------------- after reading %s -------------- \n", argv[i]);
    PrintStats();
  }

  if (!solidA)
    exit(1);

  solidB = (BREP_SOLID *)NULL;
  BrepSetCloseSolidCallback((BREP_CALLBACK_FUNC)CatchSolidB);
  ReadMGF(argv[i+1]);

  if (verbose) {
    fprintf(stderr, "------------- after reading %s -------------- \n", argv[i+1]);
    PrintStats();
  }

  if (!solidB)
    exit(1);

  do_csg(solidA, solidB);

  if (verbose) {
    fprintf(stderr, "------------- after doing CSG operation ------------------ \n");
    PrintStats();
  }

  printf("# solidA: \n");
  WriteMGF(solidA, stdout);

  printf("# solidB: \n");
  WriteMGF(solidB, stdout);

  return 0;
}

/* main program. */
int main(int argc, char **argv)
{
  if ((progname = strrchr(argv[0], '/')) != NULL)
    progname = progname+1;
  else
    progname = strdup(argv[0]);

  return csg_main(argc, argv);
}













