/* raytrace.c: */
/*
  The BREP Library.
  Copyright (C) 1996 Philippe Bekaert

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "raytrace.h"
#include "vector.h"
#include "error.h"

static int icount = 0;
static int cull_flags = CULL_NONE;

static BREP_FACE *face1, *face2;

void DontIntersectFaces(BREP_FACE *f1, BREP_FACE *f2)
{
  face1 = f1;
  face2 = f2;
}

unsigned RaytraceCull(unsigned flags)
{
  unsigned old_flags = cull_flags;
  cull_flags = flags;
  return old_flags;
}

/*
 * (This code has been copied from rayshade 4.0.6. and adapted)
 *
 * Quadrants are defined as:
 *        |
 *   1    |   0
 *        |
 * -------c--------
 *        |
 *   2    |   3
 *        |
 */
#define quadrant(p, c) ((p.u<c.u) ? ((p.v<c.v) ? 2 : 1) : ((p.v<c.v) ? 3 : 0))

/* Returns the winding number of the contour around the point, projected to 2D. */
/* this code was taken from Rayshade 4.0.6 */
int ContourWinding(BREP_CONTOUR *contour, DVEC2D *center, int index)
{
  int winding;
  int quad, lastquad;
  double left, right;
  DVEC2D cur, last;
  BREP_WING *wing;

  /*
   * Is the point inside the polygon?
   *
   * Compute the winding number by finding the quadrant each
   * polygon point lies in with respect to the the point in
   * question, and computing a "delta" (winding number).  If we
   * end up going around in a complete circle around
   * the point (winding number is non-zero at the end), then
   * we're inside.  Otherwise, the point is outside.
   *
   * Note that we can turn this into a 2D problem by projecting
   * all the points along the axis defined by poly->index,
   * the "dominant" part of the polygon's normal vector.
   */
  if (!contour->wings || (contour->wings->prev == contour->wings->next)) {
    /* fewer than three vertices in the contour, return zero */
    return 0;
  }

  winding = 0;
      
  VECTORPROJECT(last, ((VERTEX *)(contour->wings->prev->vertex->client_data))->point, index);
  lastquad = quadrant(last, (*center));

  wing = contour->wings;
  do {
    VECTORPROJECT(cur, ((VERTEX *)(wing->vertex->client_data))->point, index);
    quad = quadrant(cur, (*center));

    if (quad != lastquad) {
      if(((lastquad + 1) & 3) == quad)
	winding++;
      else if(((quad + 1) & 3) == lastquad)
	winding--;
      else {
	/*
	 * Find where edge crosses
	 * center's X axis.
	 */
	right = last.u - cur.u;
	left = (last.v - cur.v) * (center->u - last.u);
	if(left + last.v * right > right * center->v)
	  winding += 2;
	else
	  winding -= 2;
      }
    }

    lastquad = quad;    
    last = cur;
    wing = wing->next;
  } while (wing != contour->wings);
  
  return winding;
}

int VertexInsideContour(VERTEX *vertex_data, BREP_CONTOUR *contour)
{
  FACE *face_data = (FACE *)(contour->face->client_data);
  DVEC2D center;
  int index;

  index = VectorDominantCoord(&(face_data->normal));
  VECTORPROJECT(center, vertex_data->point, index);
  return ContourWinding(contour, &center, index);
}

/* Returns winding number of intersection point in the face. The winding number
 * is zero for points outside the face and != zero for contained points */
int IntersectionInsideFace(BREP_FACE *face, RAY *ray, float dist)
{
  FACE *face_data = (FACE *)(face->client_data);
  DVEC2D center;
  int winding, index;

  switch ((index = VectorDominantCoord(&(face_data->normal)))) {
  case XNORMAL:
    center.u = ray->pos.y + dist * ray->dir.y;
    center.v = ray->pos.z + dist * ray->dir.z;
    break;
  case YNORMAL:
    center.v = ray->pos.z + dist * ray->dir.z;
    center.u = ray->pos.x + dist * ray->dir.x;
    break;
  default:	/* ZNORMAL */
    center.u = ray->pos.x + dist * ray->dir.x;
    center.v = ray->pos.y + dist * ray->dir.y;
  }

  winding = 0;
  ForAllContoursInFace(contour, face) {
    int w = ContourWinding(contour, &center, index);
    winding += w;
  } EndForAll;

  return winding;
}

static int BoundsIntersected(float *bounds, RAY *ray, float mindist, float maxdist)
{
  VECTOR vtmp;

  VECTORSUMSCALED(ray->pos, mindist, ray->dir, vtmp);
  if (OutOfBounds(&vtmp, bounds))
    return BoundsIntersect(ray, bounds, mindist, &maxdist);
  return TRUE;
}

BREP_FACE *FaceIntersect(BREP_FACE *face, RAY *ray, float mindist, float *maxdist)
{
  FACE *face_data = (FACE *)(face->client_data);
  double dist, d;

  if (face == face1 || face == face2)
    return NULL;

  /* test against face bounding box */
  if (!BoundsIntersected(face_data->bounds, ray, mindist, *maxdist))
    return (BREP_FACE *)NULL;

  /* fill ray.pos in the face plane equation */
  d = VECTORDOTPRODUCT(face_data->normal, ray->pos) + face_data->d;
  if (cull_flags != CULL_NONE) {
    if ((d < 0. && (cull_flags & CULL_BACK)) ||
	(d > 0. && (cull_flags & CULL_FRONT)))
      return (BREP_FACE *)NULL;
  }

  /* compute intersection distance along the ray */
  dist = VECTORDOTPRODUCT(face_data->normal, ray->dir);
  if (fabs(dist) < EPSILON)
    return (BREP_FACE *)NULL;

  dist = - d / dist;
  if (dist < mindist || dist > *maxdist)
    return (BREP_FACE *)NULL;

  if (IntersectionInsideFace(face, ray, dist)) {
    *maxdist = dist;
    icount++;
    return face;
  }

  return (BREP_FACE *)NULL;
}

BREP_FACE *ShellIntersect(BREP_SHELL *shell, RAY *ray, float mindist, float *maxdist,
		     int shadowtest)
{
  BREP_FACE *hit;

  if (IS_HIDDEN(shell))
    return (BREP_FACE *)NULL;

  if (!BoundsIntersected(((SHELL *)(shell->client_data))->bounds, ray, mindist, *maxdist))
    return (BREP_FACE *)NULL;

  hit = (BREP_FACE *)NULL;
  ForAllFacesInShell(face, shell) {
    BREP_FACE *f = FaceIntersect(face, ray, mindist, maxdist);
    if (f) {
      if (shadowtest)
	return f;
      else
	hit = f;
    }
  } EndForAll;

  return hit;
}

BREP_FACE *SolidIntersect(BREP_SOLID *solid, RAY *ray, float mindist, float *maxdist,
			  int shadowtest)
{
  BREP_FACE *hit;

  if (!BoundsIntersected(((SOLID *)(solid->client_data))->bounds, ray, mindist, *maxdist))
    return (BREP_FACE *)NULL;

  hit = (BREP_FACE *)NULL;
  ForAllShellsInSolid(shell, solid) {
    BREP_FACE *f = ShellIntersect(shell, ray, mindist, maxdist, shadowtest);
    if (f) {
      if (shadowtest)
	return f;
      else
	hit = f;
    }
  } EndForAll;

  return hit;
}

#ifdef NEVER
static double WingLength(BREP_WING *wing)
{
  VERTEX *v1 = (VERTEX *)(wing->vertex->client_data), 
         *v2 = (VERTEX *)(wing->next->vertex->client_data);

  return VectorDist(&(v1->point), &(v2->point)) 
         - ((EDGE *)(wing->edge->client_data))->extent;
}

static BREP_WING *FindGoodWing(BREP_SOLID *solid)
{
  BREP_SHELL *shell;
  BREP_FACE *face;
  BREP_CONTOUR *contour;
  BREP_WING *wing;

  for (shell = solid->shells; shell; shell = (shell->next == solid->shells ? NULL : shell->next)) {
    for (face = shell->faces; face; face = (face->next == shell->faces ? NULL : face->next)) {
      for (contour = face->outer_contour; contour; contour = (contour->next == face->outer_contour ? NULL : contour->next)) {
	for (wing = contour->wings; wing; wing = wing->next) {
	  if (WingLength(wing) > 0.)
	    return wing;
	}
      }
    }
  }

  return (BREP_WING *)NULL;
}

/* simple application of raytracing: testing whether a vertex is inside the solid */
int VertexInSolid(BREP_VERTEX *vertex, BREP_SOLID *solid)
{
  RAY ray;
  DVECTOR p, u;
  BREP_WING *wing;
  float dist;

  VECTORCOPY(((VERTEX *)(vertex->client_data))->point, ray.pos);

  /* an arbitrary ray direction suffices, but only for a closed solid. If the solid is
   * not closed, the result would depend on the ray direction. To handle non-closed
   * solids, we prefer a point-in-polyhedron inclusion test based on solid angles. */

  /* find an arbitrary point on an arbitrary face of the solid */
  wing = FindGoodWing(solid);
  if (!wing) {
    Error("VertexInSolid", "bad solid 1");
    return 0;
  }

  /* take the midpoint of the edge */
  MIDPOINT(((VERTEX *)(wing->vertex->client_data))->point,
	   ((VERTEX *)(wing->next->vertex->client_data))->point,
	   p);
  
  /* move it a little inward to the face */
  VECTORDIFF(((VERTEX *)(wing->next->vertex->client_data))->point,
	     ((VERTEX *)(wing->vertex->client_data))->point,
	     u);
  VECTORCROSSPRODUCT(((FACE *)(wing->contour->face->client_data))->normal, 
		     u, u);
  VECTORNORMALIZE(u);
  VECTORSUMSCALED(p, 1.5 * ((EDGE *)(wing->edge->client_data))->extent, u, p);

  VECTORDIFF(p, ray.pos, ray.dir);
  dist = VECTORNORM(ray.dir) - EPSILON;
  VECTORSCALEINVERSE(dist, ray.dir, ray.dir);

  DontIntersectFaces(wing->contour->face, NULL); 
  icount=1;	/* one intersection with the face being excluded from testing */

  dist = HUGE;
  SolidIntersect(solid, &ray, ((VERTEX *)(vertex->client_data))->tolerance, &dist, FALSE);

  DontIntersectFaces(NULL, NULL);

  return icount&1 ? OUTMASK : INMASK;
}
#endif
