/* readmgf.c: reads a MGF format file and creates one BREP_SOLID containing 
 * all its objects. */
/*
  The BREP Library.
  Copyright (C) 1996 Philippe Bekaert

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#undef DEBUG 
/* #define DDEBUG */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "readmgf.h"
#include "brep.h"
#include "geometry.h"
#include "error.h"
#include "vector.h"
#include "pools.h"
#include "cie.h"
#include "globals.h"

static BREP_SOLID *currentSolid;
static BREP_SHELL *currentShell;
static BREP_FACE *currentFace;
static BREP_CONTOUR *currentContour;
static BREP_VERTEX_OCTREE *autoVertices;	/* automatically created 
						 * vertices for sph ... */

static char *currentmatname, currentobjectname[100];
static int newobject, incomplex, objectnestlevel;

void MGFSetNrQuartCircDivs(int divs)
{
  if (divs <= 0) {
    Error(NULL, "Number of quarter circle divisions (%d) should be positive", divs);
    return;
  }

  mg_nqcdivs = divs;
}

static void do_error(char *errmsg)
{
  Error(NULL, "%s line %d: %s", mg_file->fname, mg_file->lineno, errmsg); 
}

static void do_warning(char *errmsg)
{
  Warning(NULL, "%s line %d: %s", mg_file->fname, mg_file->lineno, errmsg); 
}

static void NewSolid(void)
{
  SOLID solid;

  solid.id = 0;
  solid.filename = mg_file->fname;
  currentSolid = BrepCreateSolid(&solid);
}

static void CloseSolid(void)
{
  /* compute a normal for vertices that don't have one yet. */
  /* BrepIterateVertices(currentSolid->vertices, ComputeVertexNormal); */
  BrepCloseSolid(currentSolid);
}

static RGB_MATERIAL ConvertMaterial(C_MATERIAL cmat)
{
  RGB_MATERIAL rgbmat;

  if (cmat.ed > 0.)
    MGFGetColor(&cmat.ed_c, 1., &rgbmat.emission);
  else
    RGBSET(rgbmat.emission, 0., 0., 0.);

  MGFGetColor(&cmat.rd_c, cmat.rd, &rgbmat.diffuse);
  MGFGetColor(&cmat.rs_c, cmat.rs, &rgbmat.specular);

  if (cmat.rs_a > 0.) {
    rgbmat.shininess = 0.6/cmat.rs_a;
    rgbmat.shininess *= rgbmat.shininess;
  } else
    rgbmat.shininess = 0.;
  rgbmat.alpha = 1. - cmat.td;

  return rgbmat;
}

static void NewShell(int newobject)
{
  SHELL shell;

  shell.id = mg_file->lineno;
  if (newobject)
    shell.name = currentobjectname;
  else
    shell.name = NULL;
  shell.matname = currentmatname;
  shell.material = *c_cmaterial;
  shell.rgbmat = ConvertMaterial(shell.material);
  currentShell = BrepCreateShell(currentSolid, &shell);
}

static void NewFace(void)
{
  FACE face;

  face.id = mg_file->lineno;
  currentFace = BrepCreateFace(currentShell, &face);
}

static void NewContour(void)
{
  CONTOUR contour;

  contour.id = mg_file->lineno;
  currentContour = BrepCreateContour(currentFace, &contour);
}

static void NewWing(BREP_VERTEX *v1, BREP_VERTEX *v2)
{
  EDGE edge;
  edge.id = mg_file->lineno;
  
  BrepCreateWing(v1, v2, currentContour, &edge);
}

/* returns the name of the current material */
static char *GetMaterialName(void)
{
  char *matname;

  matname = c_cmname;
  if (!matname || *matname == '\0')	/* this would cause strcmp to crash */
    matname = "unnamed";

  return matname;
}

/* checks whether the material has changed since the last time 
 * ResetMaterialClock() was called. */
static int MaterialChanged(void)
{
  return (strcmp(GetMaterialName(), currentmatname) != 0 ||
	  c_cmaterial->clock != 0);
}

/* reset the clock value so we will be aware of changes in future */
static void ResetMaterialClock(void)
{
  currentmatname = GetMaterialName();
  c_cmaterial->clock = 0;
}

/* looks up/creates a vertex */
static BREP_VERTEX *GetVertex(char *name)
{
  FVECT	vert, norm;
  C_VERTEX *vp;
  VERTEX v;
  BREP_VERTEX *thevertex;

  if ((vp = c_getvert(name)) == NULL)
    return (BREP_VERTEX *)NULL;

  xf_xfmpoint(vert, vp->p);
  v.point.x = vert[0];
  v.point.y = vert[1];
  v.point.z = vert[2];

  if (is0vect(vp->n) || ignore_normals_on_input) {
    v.normal.x = 0.;
    v.normal.y = 0.;
    v.normal.z = 0.;
  } else {
    xf_xfmvect(norm, vp->n);
    v.normal.x = norm[0];
    v.normal.y = norm[1];
    v.normal.z = norm[2];
    VECTORNORMALIZE(v.normal);
  }

  v.tolerance = vertex_tolerance;

  if (name && name[0] != '\0')
    v.name = name;
  else
    v.name = "???";

  v.id = mg_file->lineno;

  thevertex = BrepFindVertex(&v, currentSolid->vertices);
  if (!thevertex) {
    /* automatically created vertices have a name starting with a '_' */
    if (incomplex && name && name[0] == '_') {
      v.name = NULL;	/* name is unimportant */

      /* look up the vertex in the current list of automatically
       * created vertices */
      thevertex = BrepFindVertex(&v, autoVertices);
      if (!thevertex) {	
	thevertex = BrepInstallVertex(&v, &(currentSolid->vertices));
        BrepAttachVertex(thevertex, &autoVertices);
      }
    } else {
      if (name && name[0] == '_')
	name[0] = '$';	/* something else than '_' to avoid confusion */
      thevertex = BrepInstallVertex(&v, &(currentSolid->vertices));
    }
  }

  return thevertex;
}

/* Creates a contour respecting the order in which the vertices are specified */
static int do_contour_respect_order(int startidx, int endidx, char **argv)
{
  BREP_VERTEX *first, *prev, *cur;
  int i, errcode;

  errcode = MG_OK;

  if (!currentFace) {
    do_error("no current face for contour");
    return MG_OK;
  }

  NewContour();

  /* first vertex */
  cur = GetVertex(argv[startidx]);
  if (!cur)
    return MG_EUNDEF;
  first = cur;

  for (i=startidx+1; i<=endidx && errcode == MG_OK; i++) {
    prev = cur;
    cur = GetVertex(argv[i]);
    if (!cur)
      return MG_EUNDEF;
    else
      NewWing(prev, cur);
  }

  /* close the loop */
  NewWing(cur, first);

  return MG_OK;
}

/* Creates a contour ignoring the vertex order: the specified order is
 * however preferred if it does not lead to more conflicts han the reverse
 * order. We have a conflict if an edge is already used in the 
 * sense as we want to create one now. --- needs to be debugged. */
static int do_contour_no_order(int startidx, int endidx, char **argv)
{
  int num = endidx-startidx+1;
  BREP_VERTEX *vbuf[100], **v=vbuf, *cur, *prev;
  BREP_EDGE *edge;
  int i, nr_conflicts, nr_conflicts_reverse;

  if (num > 100)
    v = (BREP_VERTEX **)Alloc(num * sizeof(BREP_VERTEX *));

  /* get the vertices */
  for (i=0; i<num; i++) {
    if (!(v[i] = GetVertex(argv[startidx+i])))
      return MG_EUNDEF;
  }
  
  /* Count number of conflicts when creating the contour in the specified 
   * and reverse order. */
  nr_conflicts = nr_conflicts_reverse = 0;
  cur = v[num-1];
  for (i=0; i<num; i++) {
    prev = cur;
    cur = v[i];
    edge = BrepFindEdge(prev, cur);
    if (edge) {	/* there is an edge between the two vertices */
      if (edge->wing[0].vertex == prev) {
	if (edge->wing[0].contour) nr_conflicts++;
	if (edge->wing[1].contour) nr_conflicts_reverse++;
      } else {
	if (edge->wing[1].contour) nr_conflicts++;
	if (edge->wing[0].contour) nr_conflicts_reverse++;
      }
    }
  }

  NewContour();

  if (nr_conflicts <= nr_conflicts_reverse) {
    /* create the contour in the specified order */
    cur = v[num-1];
    for (i=0; i<num; i++) {
      prev = cur;
      cur = v[i];
      NewWing(prev, cur);
    }
  } else {
    /* create the contour in reverse order */
    cur = v[0];
    for (i=num-1; i>=0; i--) {
      prev = cur;
      cur = v[i];
      NewWing(prev, cur);
    }
  }

  if (num > 100)
    Free((char *)v, num * sizeof(BREP_VERTEX *));

  return MG_OK;
}

static int (*do_contour)(int startidx, int endidx, char **argv) = do_contour_respect_order;

/* Specify whether the vertex order in an MGF file to be read is to be
 * respected or not. When specifying that the vertex order is not relevant,
 * a test is performed to see what order yields to the least number of conflicts.
 * There is a conflict if a wing to be created already exists. This is e.g.
 * the case when vertex order is not consistent in the input file. Specifying
 * FALSE (don't respect vertex order) may lead to a consistent vertex ordering
 * in each shell. */
void MGFRespectVertexOrder(int truefalse)
{
  if (truefalse)
    do_contour = do_contour_respect_order;
  else
    do_contour = do_contour_no_order;
}

/* processes 'f' entities */
static int do_face(int argc, char **argv)
{
  if (newobject || MaterialChanged()) {
    ResetMaterialClock();	/* also gets the current materialname */
    NewShell(newobject && !incomplex);
    newobject = FALSE;
  }

  NewFace();
  return do_contour(1, argc-1, argv);
}

static int do_face_with_holes(int argc, char **argv)
{
  int startidx, endidx, errcode;

  if (newobject || MaterialChanged()) {
    ResetMaterialClock();	/* also gets the current materialname */
    NewShell(newobject && !incomplex);
    newobject = FALSE;
  }

  errcode = MG_OK;
  NewFace();

  startidx=1;
  while (startidx < argc && errcode == MG_OK) {
    /* find last vertex of contour: look for the next '-', note that
     * there's no dash at the end of the argument list */
    for (endidx=startidx+1; endidx<argc && *argv[endidx] != '-'; endidx++) {}
    endidx--;	/* we went one too far */

    errcode = do_contour(startidx, endidx, argv);
    startidx = endidx+2;	/* skip also the '-' */
  }

  return errcode;
}

/* the mgf parser already contains some good routines for discretizing spheres,
 * ... into polygons. The point is that we want to do some extra work except 
 * discretising when a sphere, cone, ... is read. In the official release of 
 * the parser library, the discretisation  routines are internal (declared 
 * static in parser.c and no reference to them in parser.h). The parser was 
 * changed so we can call them in order not to have to duplicate the code. */
static int do_discretize(int argc, char **argv)
{
  int en = mg_entity(argv[0]);
  
  switch (en) {
  case MG_E_SPH: 	return e_sph(argc, argv);
  case MG_E_TORUS: 	return e_torus(argc, argv);
  case MG_E_CYL: 	return e_cyl(argc, argv);
  case MG_E_RING: 	return e_ring(argc, argv);
  case MG_E_CONE: 	return e_cone(argc, argv);
  case MG_E_PRISM: 	return e_prism(argc, argv);
  default:
    Error("readmgf.c: do_discretize()", "Unknown geometry entity number %d", en);
    return MG_OK;	/* ignore the unknown entity */
  }
  
  return MG_EILL;	/* definitely illegal when this point is reached */
}	

/* processes 'sph', 'torus', ... entities */
static int do_surface(int argc, char **argv)
{
  int errcode;
  
  if (incomplex) /* e_sph calls e_cone ... */
    return do_discretize(argc, argv);  
  else {
    incomplex = TRUE;
    newobject = TRUE;
    autoVertices = BrepCreateVertexOctree();

    errcode = do_discretize(argc, argv);

    BrepDestroyVertexOctree(autoVertices);
    incomplex = FALSE;
    return errcode;
  }
}

/* processes 'o' entities */
static int do_object(int argc, char **argv)
{
  int i;

  newobject = TRUE;
  if (argc > 1) { 	/* beginning of a new object */
    sprintf(currentobjectname, "%s", argv[1]);
    for (i=0; i<objectnestlevel; i++)
      fprintf(stderr, "\t");
    fprintf(stderr, "%s ...\n", currentobjectname);
    objectnestlevel++;
  } else {
    objectnestlevel--;
  }
  
  return obj_handler(argc, argv);
}

static int unknown_count;	/* unknown entity count */

static int do_unknown(int argc, char **argv)
{
  do_warning("unknown entity");
  unknown_count++;
  
  return MG_OK;
}

/* MGF parser intialisations */
static void InitMGF(void)
{
  mg_ehand[MG_E_FACE] = do_face;
  mg_ehand[MG_E_FACEH] = do_face_with_holes;
  
  mg_ehand[MG_E_VERTEX] = c_hvertex;
  mg_ehand[MG_E_POINT] = c_hvertex;
  mg_ehand[MG_E_NORMAL] = c_hvertex;
  
  mg_ehand[MG_E_COLOR] = c_hcolor;
  mg_ehand[MG_E_CXY] = c_hcolor;
  mg_ehand[MG_E_CMIX] = c_hcolor;
/* we don't use spectra.... let the mgf parser library convert to tristimulus
 * values itself 
  mg_ehand[MG_E_CSPEC] = c_hcolor;
  mg_ehand[MG_E_CCT] = c_hcolor;
 */

  mg_ehand[MG_E_MATERIAL] = c_hmaterial;
  mg_ehand[MG_E_ED] = c_hmaterial;
  mg_ehand[MG_E_IR] = c_hmaterial;
  mg_ehand[MG_E_RD] = c_hmaterial;
  mg_ehand[MG_E_RS] = c_hmaterial;
  mg_ehand[MG_E_SIDES] = c_hmaterial;
  mg_ehand[MG_E_TD] = c_hmaterial;
  mg_ehand[MG_E_TS] = c_hmaterial;
  
  mg_ehand[MG_E_OBJECT] = do_object;	
  
  mg_ehand[MG_E_XF] = xf_handler;
  
  mg_ehand[MG_E_SPH] = do_surface;
  mg_ehand[MG_E_TORUS] = do_surface;
  mg_ehand[MG_E_RING] = do_surface;
  mg_ehand[MG_E_CYL] = do_surface;
  mg_ehand[MG_E_CONE] = do_surface;
  mg_ehand[MG_E_PRISM] = do_surface;

  unknown_count = 0;
  mg_uhand = do_unknown;    

  mg_init();
}	

/* reads the objects in a MGF file and collects them into one BREP_SOLID.
 * The calling procedure should set a BREP CloseSolid callback function
 * to catch the created solid. In future, we might change the code
 * so that multiple BREP_SOLIDs are created. */
void ReadMGF(char *filename)
{
  MG_FCTXT fctxt;
  int err;
  
  InitMGF();

  /* some impossible material name */
  currentmatname = " ";
  objectnestlevel = 0;
  newobject = TRUE;
  incomplex = FALSE;
  
  err = mg_open(&fctxt, filename);
  if (err) 
    Error("readMGF", mg_err[err]);
  else {
    NewSolid();

    while (mg_read() > 0 && !err) {
      err = mg_parse();
      if (err) 
	do_error(mg_err[err]);
    }
    mg_close();

    if (!err)
      CloseSolid();
  }
  mg_clear();
}



