/* vector.c: vectors */
/*
  The BREP Library.
  Copyright (C) 1996 Philippe Bekaert

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <math.h>

#include "vector.h"
#include "pools.h"

#ifndef NOPOOLS
static POOL *vectorPool = (POOL *)NULL;
#define NEWVECTOR()  	(VECTOR *)New(sizeof(VECTOR), &vectorPool)
#define DISPOSEVECTOR(ptr) Dispose((unsigned char *)(ptr), &vectorPool)
#else /*NOPOOLS*/
#define NEWVECTOR()	(VECTOR *)Alloc(sizeof(VECTOR))
#define DISPOSEVECTOR(ptr) Free((char *)ptr, sizeof(VECTOR))
#endif /*NOPOOLS*/

static int TotalVectors = 0;

/* creates a vector with given components */
VECTOR *VectorCreate(float x, float y, float z)
{	
	VECTOR *vector;

	vector = NEWVECTOR();

	vector->x = x;
	vector->y = y;
	vector->z = z;

	TotalVectors++;

	return vector;
}

/* destroys a vector */
void VectorDestroy(VECTOR *vector)
{
	DISPOSEVECTOR(vector);

	TotalVectors--;
}

/* copies a vector and returns a pointer to the nex vector */
VECTOR *VectorCopy(VECTOR *vector)
{
	VECTOR *newvector;

	newvector = NEWVECTOR();
	*newvector = *vector;
	return newvector;
}

/* normalizes a vector, returns the length of the supplied vector */
double VectorNormalize(VECTOR *vector)
{
	double norm;
	DVECTOR dvec;

	dvec.x = vector->x;
	dvec.y = vector->y;
	dvec.z = vector->z;

	norm = VECTORNORM(dvec);
	if (norm < EPSILON)
		return norm;

	VECTORSCALEINVERSE(norm, dvec, *vector);
	return norm;
}

/*
 * Find the "dominant" part of the vector (eg patch-normal).  This
 * is used to turn the point-in-polygon test into a 2D problem.
 */
int VectorDominantCoord(VECTOR *v)
{
	DVECTOR anorm;
	double indexval;

	anorm.x = fabs(v->x);
	anorm.y = fabs(v->y);
	anorm.z = fabs(v->z);
	indexval = MAX(anorm.y, anorm.z);
	indexval = MAX(anorm.x, indexval);

	return (indexval == anorm.x ? XNORMAL :
		(indexval == anorm.y ? YNORMAL : ZNORMAL));
}

double VectorDist(DVECTOR *v1, DVECTOR *v2)
{
  DVECTOR d;

  VECTORDIFF(*v2,*v1,d);
  return VECTORNORM(d);
}


