/* vertinsol.c: vertex in solid test */
/*
  The BREP Library.
  Copyright (C) 1996 Philippe Bekaert

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#undef DEBUG

#include "raytrace.h"
#include "vector.h"
#include "error.h"

#ifdef DEBUG
static int debug = FALSE;
#endif

#define ON_THE_BORDER 4*M_PI

/* Paulo Cezar Pinto Carvalho and Paulo Roma Cavalcanti, "Point in Polyhedron 
 * Testing Using Spherical Polygons", Graphics Gems V, p 42 */
static double GeoSolidAngle(VERTEX *v, BREP_CONTOUR *contour)
{
  BREP_WING *wing;
  double area, ang, s, l1, l2;
  DPOINT p1, p2, r1, a, b, n1, n2;
  VECTOR norm;
  FACE *plane = (FACE *)(contour->face->client_data);
  int nr_verts, winding;

#ifdef DDEBUG
  if (debug) {
  fprintf(stderr, "---> GeoSolidAngleDebug:\n");
  }
#endif

  /* Contour with less than 3 vertices */
  if (!contour->wings || contour->wings->prev == contour->wings->next) {
#ifdef DDEBUG
  if (debug) {
    fprintf(stderr, "%s %d: < 3 verts.\n", __FILE__, __LINE__);
  }
#endif
    return 0.0;
  }

  /* Test whether v->point lies within the plane of the contour. If so, test
   * whether the point is inside of outside the contour. If outside, the solid angle
   * under which the contour is seen is zero. If inside, the solid angle is
   * 2pi or -2pi depending on the orientation of the contour around the vertex. */
  if (fabs(VECTORDOTPRODUCT(plane->normal, v->point) + plane->d) < plane->extent + v->tolerance) {
    winding = VertexInsideContour(v, contour);
#ifdef DDEBUG
  if (debug) {
    fprintf(stderr, "%s %d: vertex on contour plane\n", __FILE__, __LINE__);
    fprintf(stderr, "winding number = %d\n", winding);
  }
#endif
    return (winding == 0 ? 0. : /* (winding > 0 ? 2*M_PI : -2*M_PI) */ ON_THE_BORDER);
  }

  norm = ((CONTOUR *)(contour->client_data))->normal;
  area = 0.0; nr_verts = 0;

  p2 = ((VERTEX *)(contour->wings->prev->vertex->client_data))->point;
  p1 = ((VERTEX *)(contour->wings->vertex->client_data))->point;
  VECTORDIFF(p2, p1, a);
#ifdef DDEBUG
  if (debug) {
  fprintf(stderr, "%s %d;\n\tp1 = (%f,%f,%f)\n\tp2 = (%f,%f,%f)\n\ta = (%f,%f,%f)\n", __FILE__, __LINE__,
	  p1.x, p1.y, p1.z, p2.x, p2.y, p2.z, a.x, a.y, a.z);
  }
#endif

  for (wing = contour->wings; wing; 
       wing = (wing->next == contour->wings ? NULL : wing->next), nr_verts++) {
#ifdef DDEBUG
  if (debug) {
    fprintf(stderr, "\tp1 = (%f,%f,%f)\n", p1.x, p1.y, p1.z);
  }
#endif
    VECTORDIFF(p1, v->point, r1);
    p2 = ((VERTEX *)(wing->next->vertex->client_data))->point;
    VECTORDIFF(p2, p1, b);
    VECTORCROSSPRODUCT(a, r1, n1);
    VECTORCROSSPRODUCT(r1, b, n2);
#ifdef DDEBUG
  if (debug) {
  fprintf(stderr, "%s %d;\n", __FILE__, __LINE__);
  fprintf(stderr, "\tp1 = (%f,%f,%f)\n", p1.x, p1.y, p1.z);
  fprintf(stderr, "\tp2 = (%f,%f,%f)\n", p2.x, p2.y, p2.z);
  fprintf(stderr, "\ta  = (%f,%f,%f)\n", a.x, a.y, a.z);
  fprintf(stderr, "\tb  = (%f,%f,%f)\n", b.x, b.y, b.z);
  fprintf(stderr, "\t*p = (%f,%f,%f)\n", v->point.x, v->point.y, v->point.z);
  fprintf(stderr, "\tr1 = (%f,%f,%f)\n", r1.x, r1.y, r1.z);
  fprintf(stderr, "\tn1 = (%f,%f,%f)\n", n1.x, n1.y, n1.z);
  fprintf(stderr, "\tn2 = (%f,%f,%f)\n", n2.x, n2.y, n2.z);
  }
#endif
    
    l1 = VECTORNORM(n1);
    l2 = VECTORNORM(n2);
    s = VECTORDOTPRODUCT(n1, n2) / (l1 * l2);
    ang = acos(MAX(-1.0, MIN(1.0, s)));
#ifdef DDEBUG
  if (debug) {
  fprintf(stderr, "%s %d; l1 = %f, l2 = %f, s = %f, ang = %f\n", 
	  __FILE__, __LINE__, l1, l2, s, ang);
  }
#endif
    VECTORDETERMINANT(b, a, norm, s);
    area += s > -EPSILON ? M_PI - ang : M_PI + ang;
#ifdef DDEBUG
  if (debug) {
  fprintf(stderr, "%s %d; s = %f, s > 0.0 ? M_PI - ang : M_PI + ang = %f, area = %f\n", 
	  __FILE__, __LINE__, s, s > 0.0 ? M_PI - ang : M_PI + ang, area);
  }
#endif

    VECTORSCALE(-1.0 ,b, a);
    p1 = p2; 
  }

  area -= M_PI * (double)(nr_verts - 2);
#ifdef DDEBUG
  if (debug) {
  fprintf(stderr, "%s %d: nr_verts = %d, area = %f, VECTORDOTPRODUCT(norm, r1) = %f\n", 
	  __FILE__, __LINE__, nr_verts, area, VECTORDOTPRODUCT(norm, r1));
  }
#endif

  area = (VECTORDOTPRODUCT(norm, r1) > 0.0) ? -area : area;
#ifdef DDEBUG
  if (debug) {
  fprintf(stderr, "-> area = %f\n", area);
  }
#endif
  return area;
}

int VertexInSolid(VERTEX *v, BREP_SOLID *solid)
{
  double Area = 0., area;
  BREP_SHELL *shell;
  BREP_FACE *face;
  BREP_CONTOUR *contour;
  SOLID *sol = (SOLID *)(solid->client_data);

#ifdef DEBUG
  if (debug) {
  fprintf(stderr, "===> VertexInSolid:\n");
  fprintf(stderr, "\tvertex = "); PrintVertexData(stderr, v); fprintf(stderr, "\n");
  fprintf(stderr, "\tsolid = "); PrintSolidData(stderr, solid->client_data); fprintf(stderr, "\n");
  }
#endif

  if (OutOfBounds(&(v->point), sol->bounds)) {
#ifdef DEBUG
  if (debug) {
    fprintf(stderr, "vertex OUTSIDE bounding box of solid\n");
  }
#endif
    return OUTMASK;
  }

  Area = 0.;
  for (shell = solid->shells; shell; shell = (shell->next == solid->shells ? NULL : shell->next)) {
    for (face = shell->faces; face; face = (face->next == shell->faces ? NULL : face->next)) {
      for (contour = face->outer_contour; contour; contour = (contour->next == face->outer_contour ? NULL : contour->next)) {	
	area = GeoSolidAngle(v, contour);
	if (area < -2*M_PI || area > 2*M_PI) {
	  if (area != ON_THE_BORDER)
	    Error("VertexInSolid", "something wrong with GeoSolidAngle()");
	  return 0;
	}
	Area += area;
#ifdef DDEBUG
  if (debug) {
	fprintf(stderr, "Area = %12.4f\n", Area);
  }
#endif
      }
    }
  }

#ifdef DEBUG
  if (debug) {
  fprintf(stderr, "--> vertex %s solid\n", Area > 2*M_PI || Area < -2*M_PI ? "INSIDE" : "OUTSIDE");
  }
#endif
  return (Area > 2*M_PI || Area < -2*M_PI) ? INMASK : OUTMASK;
}

int ContourInSolid(BREP_CONTOUR *contour, BREP_SOLID *solid)
{
  BREP_WING *wing;
  DVECTOR u, p;
  int mask = 0;
  VERTEX v;
#ifdef DEBUG
  int count = 0;

  /*  if ((((CONTOUR *)(contour->client_data))->id & ~IDMASK) == 30)
    debug = TRUE; */

  if (debug) {
  fprintf(stderr, "ContourInSolid:\n");
  fprintf(stderr, "\tsolid = "); PrintSolidData(stderr, solid->client_data);
  fprintf(stderr, "\tcontour = "); PrintContourData(stderr, contour->client_data); fprintf(stderr, "\n");
  fprintf(stderr, "\tcontour vertices:\n");
  for (wing = contour->wings; wing; wing = (wing->next == contour->wings ? NULL : wing->next)) {
    fprintf(stderr, "\t\tv%d = ", count++); PrintVertexData(stderr, wing->vertex->client_data); fprintf(stderr, "\n");
  }
  }
#endif

  if (!contour->wings)
    return OUTMASK;

  wing = contour->wings;
  do {
    /* take the midpoint of the wing */
    MIDPOINT(((VERTEX *)(wing->vertex->client_data))->point,
	     ((VERTEX *)(wing->next->vertex->client_data))->point,
	     p);

    /* shift it a little bit into the face */
    VECTORSUBTRACT(((VERTEX *)(wing->next->vertex->client_data))->point,
		   ((VERTEX *)(wing->vertex->client_data))->point,
		   u);
    VECTORCROSSPRODUCT(((CONTOUR *)(contour->client_data))->normal, u, u);
    VECTORNORMALIZE(u);

    v = *((VERTEX *)(wing->vertex->client_data));
    VECTORADDSCALED(p, 10.*EPSILON, u, v.point);
    v.tolerance = ((EDGE *)(wing->edge->client_data))->tolerance;

    mask = VertexInSolid(&v, solid);
#ifdef DEBUG
    if (debug) {
    fprintf(stderr, "mask = %s\n", mask == OUTMASK ? "OUTSIDE" : (mask == INMASK ? "INSIDE" : "UNKNOWN"));
    }
#endif
    wing = wing->next;
  } while (wing != contour->wings && !mask);

#ifdef DEBUG
  debug = FALSE;
#endif

  return mask;
}
