/* writemgf.c */
/*
  The BREP Library.
  Copyright (C) 1996 Philippe Bekaert

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#undef DEBUG

#include "writemgf.h"
#include "geometry.h"
#include "pools.h"
#include "error.h"
#include "isect.h"
#include "vector.h"
#include "brepit.h"
#include "raytrace.h"
#include "globals.h"

static FILE *out;
int write_as_is, write_reverse;	/* defaults set in CreateSaveoptionsmenu() in ui_file.c */

extern int ignore_materials;

#ifdef KEEPVERTEXNAMES
#define VERTEX_NAME_LEN 8
#ifndef NOPOOLS
static POOL *namePool = (POOL *)NULL;
#define NEWNAME()  	New(VERTEX_NAME_LEN, &namePool)
#define DISPOSENAME(ptr) Dispose((unsigned char *)(ptr), &namePool)
#else /*NOPOOLS*/
#define NEWNAME()	Alloc(VERTEX_NAME_LEN)
#define DISPOSENAME(ptr) Free(ptr, VERTEX_NAME_LEN)
#endif /*NOPOOLS*/
#endif /*KEEPVERTEXNAMES*/

static void OutputVertexData(VERTEX *vertex_data)
{
#ifdef KEEPVERTEXNAMES
  if (!vertex_data->name) {	/* make a unique name for automatically */
    char *name = NEWNAME();	/* created vertices. The name is disposed of */
    auto_vertex_count++;	/* with FreeAutoVertexName() below */
    sprintf(name, "_a%d", auto_vertex_count);
    vertex_data->name = name;
  }

  if (vertex_data->name[0] == '$')
    fprintf(out, "v _%s =\n", &vertex_data->name[1]);
  else
    fprintf(out, "v %s =\n", vertex_data->name);
#else
  fprintf(out, "v v%d = \n", vertex_data->id & ~IDMASK);
#endif

  fprintf(out, "\tp %f %f %f\n", 
	  vertex_data->point.x, vertex_data->point.y, vertex_data->point.z);

  if (VECTORNORM(vertex_data->normal) > EPSILON)
  fprintf(out, "\tn %f %f %f\n", 
	  vertex_data->normal.x, vertex_data->normal.y, vertex_data->normal.z);
}

static void OutputVertex(BREP_VERTEX *vertex)
{
  VERTEX *vertex_data = (VERTEX *)(vertex->client_data);

  if (vertex->wing_ring && !(vertex_data->id & OUTMASK))
    OutputVertexData(vertex_data);
  vertex_data->id |= OUTMASK;
}

static void OutputVertexName(BREP_VERTEX *vertex)
{
  VERTEX *vertex_data = (VERTEX *)(vertex->client_data);
#ifdef KEEPVERTEXNAME
  if (vertex_data->name[0] == '$')
    fprintf(out, "_%s ", &vertex_data->name[1]);
  else
    fprintf(out,"%s ", vertex_data->name);
#else
  fprintf(out, "v%d ", vertex_data->id & ~IDMASK);
#endif
}

static void OutputContourData(CONTOUR *contour_data)
{
#ifdef DEBUG
  fprintf(out, "# ");
  PrintContourData(out, contour_data);
  fprintf(out, "\n");
  if (contour_data->id & OUTMASK)
    fprintf(out, "# contour outside other solid:\n");
  if (contour_data->id & INMASK)
    fprintf(out, "# contour inside other solid:\n");
#endif
}

static unsigned ContourWriteMask(BREP_CONTOUR *contour)
{
  return (unsigned) ((CONTOUR *)(contour->client_data))->id & (WRITEMASK|WRITEREVERSEMASK|DONTWRITEMASK);
}

void DoOutputContour(char *pre, BREP_CONTOUR *contour, char *post)
{
  unsigned writemask = ContourWriteMask(contour);

  if (writemask == DONTWRITEMASK)
    return;

  OutputContourData(contour->client_data);

  fprintf(out, "%s ", pre);
  if ((writemask & WRITEREVERSEMASK) ^ write_reverse)
    BrepContourReverseIterateVertices(contour, OutputVertexName);
  else
    BrepContourIterateVertices(contour, OutputVertexName);
  fprintf(out, "%s", post);
}

void OutputContour(BREP_CONTOUR *contour)
{
  if (IsHole(contour))
    DoOutputContour("h", contour, "\n");
  else
    DoOutputContour("f", contour, "\n");
}

void OutputContour2(BREP_CONTOUR *contour)
{
  if (IsHole(contour))
    DoOutputContour("-", contour, " ");
  else
    DoOutputContour("\nfh", contour, " ");
}

typedef enum {COEXISTING, CONTAINING, CONTAINED} CONTAINEMENT;

/* test may go wrong if the contours share a vertex without sharing an edge */
static CONTAINEMENT CompareContours(BREP_CONTOUR *c, BREP_CONTOUR *ref)
{
  BREP_WING *w, *w1, *w2;

  /* same contour: COEXISTING */
  if (c == ref) {
    Error("CompareContours", "same contour.");
    return COEXISTING;	/* should never happen */
  }

  /* no wings ... coexisting */
  if (!c->wings || !ref->wings) {
    Error("CompareContours", "no wings.");
    return COEXISTING;
  }

  /* find a wing w1 of c on an edge not shared with ref */
  w1 = (BREP_WING *)NULL;
  for (w = c->wings; w; w = (w->next == c->wings ? (BREP_WING *)NULL : w->next)) {
    if (BrepEdgeOtherWing(w)->contour != ref) {
      w1 = w;
      break;
    }
  }

  /* find a wing w2 of ref on an edge not shared with c */
  w2 = (BREP_WING *)NULL;
  for (w = ref->wings; w; w = (w->next == ref->wings ? (BREP_WING *)NULL : w->next)) {
    if (BrepEdgeOtherWing(w)->contour != c) {
      w2 = w;
      break;
    }
  }

  if (!w1 && !w2) {
    /* coinciding contours: one is a hole, the other is an outer contour. A hole is
     * always contained in some other outer contour. Let it contain the outer contour. */
    if (IsHole(c))
      return CONTAINING;
    else
      return CONTAINED;
  }

  if (w1) {
    /* if this wing is inside the ref contour, the whole contour c is inside ref */
    VERTEX v;
    /* compute midpoint of edge */
    MIDPOINT(((VERTEX *)(w1->edge->wing[0].vertex->client_data))->point,
	     ((VERTEX *)(w1->edge->wing[1].vertex->client_data))->point,
	     v.point);
    if (VertexInsideContour(&v, ref))
      return CONTAINED;
  }

  if (w2) {
    /* if this wing is inside the c contour, the whole contour ref is inside c*/
    VERTEX v;
    /* compute midpoint of edge */
    MIDPOINT(((VERTEX *)(w2->edge->wing[0].vertex->client_data))->point,
	     ((VERTEX *)(w2->edge->wing[1].vertex->client_data))->point,
	     v.point);
    if (VertexInsideContour(&v, c))
      return CONTAINING;
  }

  /* they are disjunct */
  return COEXISTING;
}

/* Containment/Coexistance graph */
typedef struct COCONODE {
  BREP_CONTOUR *contour;		/* reference contour */
  struct COCONODE *contained_contour;	/* a contour contained in the first */
  struct COCONODE *next_coexisting,	/* disjuct contours not contained by the first */
	          *prev_coexisting;	/* (doubly linked circular list) */
} COCONODE;

/* remove the contour from the ring of coexisting contours */
static COCONODE *RemoveCoexistingContour(COCONODE *p, COCONODE *ring)
{			
  if (p->next_coexisting != p) {				
    p->next_coexisting->prev_coexisting = p->prev_coexisting; 	
    p->prev_coexisting->next_coexisting = p->next_coexisting;   
    if (ring == p) ring = p->next_coexisting;			
  } else {							
    /* it's the only element in the ring */			
    ring = (COCONODE *)NULL;					
  }								
  return ring;
}

/* add the contour to a ring of coexisting contours */
static COCONODE *AddCoexistingContour(COCONODE *p, COCONODE *ring)
{
  if (!ring)							
    ring = p;							
  else {							
    p->next_coexisting = ring->next_coexisting;			
    p->prev_coexisting = ring;					
  }								
  p->next_coexisting->prev_coexisting = p;			
  p->prev_coexisting->next_coexisting = p;			
  return ring;
}

/* inserts a new containement/coexistance graph node into the graph */
static COCONODE *InsertContour(COCONODE *new, COCONODE *root, CONTAINEMENT containement)
{
  COCONODE *p, *q;
#ifdef DEBUG
{fprintf(out, "# InsertContour: new contour = ");
if (IsHole(new->contour))
  DoOutputContour("# h", new->contour, ", root contour = ");
else 
  DoOutputContour("# f", new->contour, ", root contour = ");
if (IsHole(root->contour))
  DoOutputContour("# h", root->contour, ", containement = ");
else 
  DoOutputContour("# f", root->contour, ", containement = ");
fprintf(out, "%s\n",
	containement == CONTAINED ? "CONTAINED" : 
	(containement == CONTAINING ? "CONTAINING" :
	 (containement == COEXISTING ? "COEXISTING" : "UNKNOWN")));
}
#endif /*DEBUG*/

  switch (containement) {
  case CONTAINED:
    /* If there is a contained contour of root that contains new, new needs
     * to be inserted in the contained_contour list of that contour. If 
     * there is no such contour, all contained contours of root that are
     * also contained by new, need to be moved to new->contained_contour
     * and new added to the contained contours list of root. */
    p = root->contained_contour;
    while (p) {
      switch (CompareContours(new->contour, p->contour)) {
      case CONTAINED:
	/* p can be contained in at most on contour contained in root. */
	InsertContour(new, p, CONTAINED);
	return root;
	break;
      case CONTAINING:
	/* move p to contained_contour list of new, in other words: add it
	 * to the coexistings list of new->contained_contour. */
	q = p->next_coexisting;
	root->contained_contour = RemoveCoexistingContour(p, root->contained_contour);
	new->contained_contour = AddCoexistingContour(p, new->contained_contour);
	p = q;
	break;
      case COEXISTING:
	p = p->next_coexisting;
	break;
      default:
	Fatal(1, "InsertContour", "Invalid return code from CompareContours");
      }

      if (p == root->contained_contour) p = (COCONODE *)NULL;
    }

    /* add 'new' to the contained_contour list of root. If 'new' were contained 
     * on one of the contours contained by root, we wouldn't arrive to this 
     * point. */
    root->contained_contour = AddCoexistingContour(new, root->contained_contour);
    break;

  case COEXISTING:
    /* If 'new' is contained in a disjunct contour of root, 'new' needs
     * to be inserted in the contained_contour list of that contour.
     * If there is no such contour, all contours coexisting with root that
     * are contained by 'new', need to be inserted in the contained_contour 
     * list of 'new' and 'new' added to the list of contours coexisting
     * with root. */
    p = root->next_coexisting;
    while (p && p != root) {
      switch (CompareContours(new->contour, p->contour)) {
      case CONTAINED:
	/* p can be contained in at most one contour coexisting with root. */
	InsertContour(new, p, CONTAINED);
	return root;
	break;
      case CONTAINING:
	/* move p to coexisting_contours list of new */
	q = p->next_coexisting;
	root->next_coexisting = RemoveCoexistingContour(p, root->next_coexisting);
	new->next_coexisting = AddCoexistingContour(p, new->next_coexisting);
	p = q;
	break;
      case COEXISTING:
	p = p->next_coexisting;
	break;
      default:
	Fatal(1, "InsertContour", "Invalid return code from CompareContours");
      }
    }

    /* add 'new' to the coexisting_contours list of root. If 'new' were contained 
     * on one of the contours contained by root, we wouldn't arrive to this 
     * point.*/
    root->next_coexisting = AddCoexistingContour(new, root->next_coexisting);
    break;

  case CONTAINING:
    /* new contains root. The contours contained by root are automatically
     * also contained by new. The contours coexisting with root, may be
     * coexisting with new, in which case they have to be moved to
     * the list of contours coexisting with new, or may also be contained in
     * new, in which case they can stay where they are. It is not possible
     * that a contour coexisting with root contains new. new becomes the new
     * root node of the graph. */
    p = root->next_coexisting;
    while (p && p != root) {
      switch (CompareContours(new->contour, p->contour)) {
      case COEXISTING:
	/* move 'p' from coexisting contours list of 'root' to coexisting contours 
	 * list of 'new' */
	q = p->next_coexisting;
	root->next_coexisting = RemoveCoexistingContour(p, root->next_coexisting);
	new->next_coexisting = AddCoexistingContour(p, new->next_coexisting);
	p = q;
	break;
      case CONTAINING:
	p = p->next_coexisting;
	break;
      case CONTAINED:
	/* new contains root, some contour disjunct with root contains new
	 * ===> the contour "disjunct" with root, must contain root: impossible */
	Error("InsertContour", "Bad containement graph");
	break;
      default:
	Fatal(1, "InsertContour", "Invalid return code from CompareContours");
      }
    }

    /* new becomes the new root of the graph. The old root is a contained contour
     * of new. */
    new->contained_contour = root;
    root = new;
    break;

  default:
    Fatal(1, "InsertContour", "Invalid return code from CompareContours");
  }

  return root;
}

static void OutputHoles(COCONODE *c)
{
  COCONODE *p;

  for (p=c; p; p=(p->next_coexisting == c ? (COCONODE *)NULL : p->next_coexisting)) {
    if (IsHole(p->contour))
      DoOutputContour("-", p->contour, "");
  }
}

static void OutputContoursInContainementGraph(COCONODE *root, int recursdepth)
{ 
  COCONODE *p;

  if (!root)
    return;

  for (p=root; p; p = (p->next_coexisting == root ? (COCONODE *)NULL : p->next_coexisting)) {
    if (p->contained_contour)
      /* outputs the lower level faces and their holes */
      OutputContoursInContainementGraph(p->contained_contour, recursdepth+1);

    if (!IsHole(p->contour)) {
      if (p->contained_contour) {
	/* face with holes (faces cannot directly contain other faces) */
	DoOutputContour("fh", p->contour, "");
	OutputHoles(p->contained_contour);
	fprintf(out, "\n");
      } else
	/* simple face */
	DoOutputContour("f", p->contour, "\n");
    } else {
      /* the holes will be written after the containing outer contour */
    }
  }
}

static void MakeContainementGraphAndOutput(BREP_FACE *face, int nrcontours)
{
  COCONODE coconodes[nrcontours], *root, *new;
  BREP_CONTOUR *contour;

#ifdef DEBUG
{BREP_CONTOUR *c;
fprintf(out, "# face contours:\n");
for (c=face->outer_contour; c; c = (c->next == face->outer_contour ? (BREP_CONTOUR *)NULL : c->next)) {
  if (!IsHole(c))
    DoOutputContour("# f", c, "\n");
  else
    DoOutputContour("# h", c, "\n");
}}
#endif

  root = (COCONODE *)NULL; new = &coconodes[0];
  for (contour = face->outer_contour; contour;
       contour = (contour->next == face->outer_contour ? (BREP_CONTOUR *)NULL : contour->next)) {
    if (ContourWriteMask(contour) != DONTWRITEMASK) {
      new->contour = contour;
      new->contained_contour = (COCONODE *)NULL;
      new->prev_coexisting = new->next_coexisting = new;

      if (!root)
	root = new;
      else
	root = InsertContour(new, root, CompareContours(new->contour, root->contour));

      new++;
    }
  }

  OutputContoursInContainementGraph(root, 0);
}

static void OutputFaceWithHoles(BREP_FACE *face)
{
  BREP_CONTOUR *contour;
  int nrcontours;

  /* count the number of contours to be written */
  nrcontours = 0;
  for (contour = face->outer_contour; contour;
       contour = (contour->next == face->outer_contour ? (BREP_CONTOUR *)NULL : contour->next))
    if (ContourWriteMask(contour) != DONTWRITEMASK)
      nrcontours++;
  
  /* no contours to be written */
  if (nrcontours == 0)
    return;
  /*  
  MakeContainementGraphAndOutput(face, nrcontours);
  */
  BrepFaceIterateContours(face, OutputContour2);
  printf("\n");
}

static int FaceWithHoles(BREP_FACE *face)
{
  BREP_CONTOUR *c;

  for (c=face->outer_contour; c; c = (c->next == face->outer_contour ? (BREP_CONTOUR *)NULL : c->next)) {
    if (ContourWriteMask(c) != DONTWRITEMASK && IsHole(c))
      return TRUE;
  }

  return FALSE;
}

void OutputFace(BREP_FACE *face)
{
  BREP_CONTOUR *c;

  ForAllContoursInFace(contour, face) {
    ForAllWingsInContour(wing, contour) {
      OutputVertex(wing->vertex);
    } EndForAll;
  } EndForAll;
  /*  BrepFaceIterateContours1A(face, (void (*)(BREP_CONTOUR *, void *))BrepContourIterateVertices, (void *)OutputVertex); */

  if (write_as_is) {
    if (face->outer_contour->next == face->outer_contour) {
      /* face with one contour */
      DoOutputContour("f ", face->outer_contour, "\n");
    } else {
      int first=TRUE;
      ForAllContoursInFace(contour, face) {
	DoOutputContour(first ? "fh " : "- ", contour, " ");
	first = FALSE;
      } EndForAll;
      fprintf(out, "\n");
    }
  } else {
    if (FaceWithHoles(face))
      OutputFaceWithHoles(face);
    else {
      for (c=face->outer_contour; c; c = (c->next == face->outer_contour ? (BREP_CONTOUR *)NULL : c->next))
	DoOutputContour("f", c, "\n");
    }
  }
}

static void OutputColor(C_COLOR *color)
{
  if (color->flags & C_CSXY)	/* xy chromaticities are set */
    fprintf(out, "\tc\n\t\tcxy %f %f\n", color->cx, color->cy);
  else {	/* not needed since the parser converts spectra ...
		 * to CIE xy when reading an MGF file */
    fprintf(out, "# I don't know how to output a spectrum\n");
    Error("OutputColor", "I don't know how to output a spectrum");
  }
}

/* only non-default values are written, see MGF docs */
void OutputMaterial(char *name, C_MATERIAL *material)
{
  fprintf(out, "m %s =\n", name);
  if (material->sided)
    fprintf(out, "\tsides %d\n", material->sided ? 1 : 2);
  if (material->rd > 0.) {
    OutputColor(&(material->rd_c));
    fprintf(out, "\trd %f\n", material->rd);
  }
  if (material->td > 0.) {
    OutputColor(&(material->td_c));
    fprintf(out, "\ttd %f\n", material->td);
  }
  if (material->ed > 0.) {
    OutputColor(&(material->ed_c));
    fprintf(out, "\ted %f\n", material->ed);
  }
  if (material->rs > 0.) {
    OutputColor(&(material->rs_c));
    fprintf(out, "\trs %f %f\n", material->rs, material->rs_a);
  }
  if (material->ts > 0.) {
    OutputColor(&(material->ts_c));
    fprintf(out, "\tts %f %f\n", material->ts, material->ts_a);
  }
  if (material->nr != 1. || material->ni != 0.)
    fprintf(out, "ir %f %f\n", material->nr, material->ni);
}

static void OutputShellData(SHELL *shell_data)
{
  if (shell_data->name && *shell_data->name) {
    fprintf(out, "o %s\n", shell_data->name);    
    fprintf(stderr, "%s ...\n", shell_data->name);    
  } else {
    fprintf(out, "o o%d\n", shell_data->id & ~IDMASK);
    fprintf(stderr, "o%d ...\n", shell_data->id & ~IDMASK);
  }

  if (!ignore_materials)
    OutputMaterial(shell_data->matname, &(shell_data->material));
}

void PreShell(BREP_SHELL *shell)
{
  OutputShellData((SHELL *)(shell->client_data));
}

void PostShell(BREP_SHELL *shell)
{
  fprintf(out, "o\n");
}

void OutputShell(BREP_SHELL *shell)
{
  PreShell(shell);
  BrepShellIterateFaces(shell, OutputFace);
  PostShell(shell);
}

static void ResetVertexOutbit(BREP_VERTEX *vertex)
{
  VERTEX *vertex_data = (VERTEX *)(vertex->client_data);
  vertex_data->id &= ~OUTMASK;
}

static void FreeAutoVertexName(BREP_VERTEX *vertex)
{
#ifdef KEEPVERTEXNAMES
  VERTEX *vertex_data = (VERTEX *)(vertex->client_data);

  if (vertex_data->name &&
      vertex_data->name[0] == '_' &&
      vertex_data->name[1] == 'a') {
    DISPOSENAME(vertex_data->name);
    vertex_data->name = NULL;
  }
#endif /*KEEPVERTEXNAMES*/
}

void OutputSolid(BREP_SOLID *solid)
{
  BrepSolidIterateShells(solid, OutputShell);
}

void InitMGFOutput(BREP_SOLID *solid, FILE *_out)
{
  BrepIterateVertices(solid->vertices, ResetVertexOutbit);
  out = _out;
}

void CloseMGFOutput(BREP_SOLID *solid)
{
  BrepIterateVertices(solid->vertices, FreeAutoVertexName);
}

void WriteFace(BREP_FACE *face, FILE *out)
{
  InitMGFOutput(face->shell->solid, out);
  OutputFace(face);
  CloseMGFOutput(face->shell->solid);
}

void WriteShell(BREP_SHELL *shell, FILE *out)
{
  InitMGFOutput(shell->solid, out);
  OutputShell(shell);
  CloseMGFOutput(shell->solid);
}

void WriteMGF(BREP_SOLID *solid, FILE *out)
{
  InitMGFOutput(solid, out);
  OutputSolid(solid);
  CloseMGFOutput(solid);
}
