/* List.c: generic doubly linked linear lists */

#include "DList.h"
#include "private.h"
#include "pools.h"

#ifndef NOPOOLS
static POOL *dlistPool = (POOL *)NULL;
#define NEWDLIST()  	(DLIST *)New(sizeof(DLIST), &dlistPool)
#define DISPOSEDLIST(ptr) Dispose((unsigned char *)(ptr), &dlistPool)
#else /*NOPOOLS*/
#define NEWDLIST()	(DLIST *)Alloc(sizeof(DLIST))
#define DISPOSEDLIST(ptr) Free((char *)ptr, sizeof(DLIST))
#endif /*NOPOOLS*/


#ifndef NODLISTMACROS
void *__pdlistel__;	/* wordt gebruikt in de DlistNext() macro */
#endif /*NODLISTMACROS*/

#ifdef NODLISTMACROS
/* creeert een lege lijst */
DLIST *DListCreate(void)
{
	return (DLIST *)NULL;
}
#endif /*NODLISTMACROS*/

/* voegt het element vooraan aan de lijst toe, geeft de nieuwe wijzer naar 
 * de lijst terug */
DLIST *DListAdd(DLIST *dlist, void *pelement)
{
	DLIST *newdlist;

/* weiger NULL-pointers in de lijst */
	if (pelement == (void *)NULL)
		return dlist;

	newdlist = NEWDLIST();
	newdlist->pelement = pelement;
	newdlist->next = dlist;
	newdlist->prev = (DLIST *)NULL;

	if (dlist)
	  dlist->prev = newdlist;

	return newdlist;
}

/* telt het aantal elementen in de lijst */
int DListCount(DLIST *dlist)
{
	int count = 0;

	while (dlist) {
		count++;
		dlist = dlist->next;
	}

	return count;
}

/* geeft het index-ste element uit de lijst terug of NULL als er
 * geen index elementen in de lijst zijn. index telt vanaf 0. */
void *DListGet(DLIST *dlist, int index)
{
	while (dlist && index > 0) {
		dlist = dlist->next;
		index--;
	}

	if (!dlist)
		return (void *)NULL;
	else
		return dlist->pelement;
}

#ifdef NODLISTMACROS
/* het eerste argument is het adres van een DLIST *. 
 * Aanvankelijk moet deze DLIST * naar het begin van een lijst wijzen.
 * Bij herhaaldelijke oproepen zal de lijst doorlopen worden.
 * De functie geeft bij herhaaldelijk oproepen een na een alle
 * elementen van de lijst terug. Indien de lijst doorlopen is,
 * wordt een NULL wijzer teruggegeven */
void *DListNext(DLIST **dlist)
{
	void *pelement = (void *)NULL;

	if (*dlist) {
		pelement = (*dlist)->pelement;
		*dlist = (*dlist)->next;
	} else
		pelement = (void *)NULL;

	return pelement;
}
#endif /*NODLISTMACROS*/

/* voegt twee lijsten samen: de elementen van lijst dlist2 komen voor de elementen van 
 * lijst dlist1. Een wijzer naar de verlengde lijst dlist1 wordt teruggegeven. */
DLIST *DListMerge(DLIST *dlist1, DLIST *dlist2)
{
	void *pelement = (void *)NULL;

	while ((pelement = DListNext(&dlist2)))
		dlist1 = DListAdd(dlist1, pelement);

	return dlist1;
}

/* dupliceert een lijst: de elementen zelf worden niet gedupliceerd */
DLIST *DListDuplicate(DLIST *dlist)
{
	DLIST *newdlist = (DLIST *)NULL;
	void *pelement = (void *)NULL;

	while ((pelement = DListNext(&dlist)))
		newdlist = DListAdd(newdlist, pelement);

	return newdlist;
}

/* verwijdert de cel p uit de lijst dlist */
DLIST *DListRemoveCell(DLIST *dlist, DLIST *p)
{
	if (p == dlist) {
		dlist = dlist->next;
	        dlist->prev = (DLIST *)NULL;
		DISPOSEDLIST(p);
	} else {
		if (p->prev) p->prev->next = p->next;
		if (p->next) p->next->prev = p->prev;
		DISPOSEDLIST(p);
	}

/* geef een wijzer naar het begin van de lijst terug */
	return dlist;
}

/* verwijdert een element uit de lijst. Geeft een pointer naar de gewijzigde 
 * lijst terug */
DLIST *DListRemove(DLIST *dlist, void *pelement)
{
	DLIST *p;

/* als de lijst ledig is */
	if (!dlist) {
		GdtError("DlistRemove", "attempt to remove an element from an empty dlist");
		return dlist;	/* voor het geval we ooit besluiten tolerant te zijn */
	}

/* zoek het element in de lijst */
	for (p = dlist; p && p->pelement != pelement; p = p->next) {}

/* als p de NULL pointer is komt het te verwijderen element niet in de lijst
 * voor: fout */
	if (!p) {
		GdtError("DlistRemove", "attempt to remove a nonexisting element from a dlist");
		return dlist;
	}

/* in het andere geval is p een wijzer naar de lijst-cel die eruit gehaald moet
 * worden */
	return DListRemoveCell(dlist, p);
}

#ifdef NODLISTMACROS
/* iterator: voer de procedure uit vooral alle elementen van de lijst.
 * De procedure heeft als argument een wijzer naar een element */
void DListIterate(DLIST *dlist, void (*proc)(void *))
{
	void *pelement;

	while (dlist) {
		pelement = dlist->pelement;
		dlist = dlist->next;
		proc(pelement);
	}		
}
#endif /*NODLISTMACROS*/

#ifdef NODLISTMACROS
/* iterator: voer de procedure uit vooral alle elementen van de lijst.
 * De procedure heeft twee argumenten: eerst een wijzer naar
 * het element, dan een wijzer naar andere data de gelijk is voor alle
 * elementen */
void DListIterate1A(DLIST *dlist, void (*proc)(void *, void *), void *extra)
{
	void *pelement;

	while (dlist) {
		pelement = dlist->pelement;
		dlist = dlist->next;
		proc(pelement, extra);
	}		
}
#endif /*NODLISTMACROS*/

#ifdef NODLISTMACROS
/* iterator: voer de procedure uit vooral alle elementen van de lijst.
 * De procedure heeft twee argumenten: eerst een wijzer naar
 * data die gelijk is voor alle elementen, dan een wijzer naar het element */
void DListIterate1B(DLIST *dlist, void (*proc)(void *, void *), void *extra)
{
	void *pelement;

	while (dlist) {
		pelement = dlist->pelement;
		dlist = dlist->next;
		proc(extra, pelement);
	}		
}
#endif /*NODLISTMACROS*/

/* een volledige lijst vernietigen: vernietigt niet de elementen! */
void DListDestroy(DLIST *dlist)
{
	DLIST *p;

	while (dlist) {
		p = dlist->next;
		DISPOSEDLIST(dlist);
		dlist = p;
	}
}




