/* pools.c: memory management routines */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#define POOLS_PRIVATE
#include "pools.h"

int pools_debug = 0;
#ifdef DEBUG
#define DPRINTF 	if (pools_debug) printf
#endif

static long bytesAllocated=0, bytesOverhead=0;

/* Returns the total number of bytes allocated with the library,
 * excluding the overhead of the storage allocator. */
long GetMemoryUsage(void)
{
  return bytesAllocated - bytesOverhead;
}

/* Returns total bytes of overhead caused by the memory allocator. There
 * are three ways by which you can reduce the overhead: tune POOL_ALIGN, 
 * determining the default alignment of the cells, use NewAligned() instead
 * of New() wherever possible (watch out: architecture dependent!),
 * or don't define POOL_HEADERS, by which you save at least one pointer
 * per cell, at the cost of a much more expensive Dispose(). */
long GetMemoryOverhead(void)
{
  return bytesOverhead;
}

/* Allocate nrbytes bytes of memory and return a pointer to the allocated 
 * memory. */
char *Alloc(int nrbytes)
{
  char *buf;
  
  if ((buf = (char *)malloc(nrbytes)) == NULL) 
    PoolsFatal("Alloc", "failed to allocate %d bytes", nrbytes);
  bytesAllocated += nrbytes;
#ifdef DEBUG
  DPRINTF("Alloc %d bytes OK: %ld bytes allocated in total now.", nrbytes, bytesAllocated);
#endif
  
  return buf;
}

/* Return the memory pointed to by buf to the system. The second argument
 * is only used to keep track of how much memory is being used by the
 * library. It should be the number of bytes being pointed to by buf,
 * same argument as when it was allocate dusing Alloc() above. */
void Free(char *buf, int nrbytes)
{
  free(buf);
  bytesAllocated -= nrbytes;
#ifdef DEBUG
  DPRINTF("Free %d bytes ... OK: %ld bytes allocated in total now.", nrbytes, bytesAllocated);
#endif
}

/* Changes the size of the piece of memory being pointed to by buf:
 * new size is size+extra. extra can be negative. */
char *Realloc(char *buf, int size, int extra)
{
  if ((buf = (char *)realloc(buf, size+extra)) == NULL) 
    PoolsFatal("Realloc", "failed to reallocate %d bytes", 
	       size+extra);
  bytesAllocated += extra;
#ifdef DEBUG
  DPRINTF("Realloc %d to %d bytes ... OK: %ld bytes allocated in total now.", size, size+extra, bytesAllocated);
#endif
  return buf;
}

/* ----------------------------------------------------------------- */
/* Returns a magic number. Two magic numbers are used: one to mark the
 * start of a POOL and one to mark the start of a POOL_TRAILER. */
static unsigned GetMagic(int which)
{
  static char magic[2][4] = {
    {'P', 'o', 'E', 'l'},
    {'p', 'O', 'e', 'L'}
  };
  union {char c[4]; unsigned i;} magic_union;

  which = which-1;
  magic_union.i = 0;
  magic_union.c[0] = magic[which][0];
  magic_union.c[1] = magic[which][1];
  magic_union.c[2] = magic[which][2];
  magic_union.c[3] = magic[which][3];
  return magic_union.i;
}

/* macros to mark cells as being allocated or free in the allocmap of
 * a pool */
#define MARK_ALLOCED(allocmap,i)   (allocmap)[(i)>>3] |= (0x80 >> ((i)&7))
#define MARK_FREED(allocmap,i)   (allocmap)[(i)>>3] &= ~(0x80 >> ((i)&7))
#define MARKED_ALLOCED(allocmap,i)   (allocmap)[(i)>>3] & (0x80 >> ((i)&7))

#define INCREMENT_MODULO(count,base)  {(count)++; if ((count) >= (base)) (count) = 0;}

#define PageIsFull(page)	((page)->allocedcells >= (page)->maxcells-(page)->reservedcells)
#define CellAddress(page,index)	((page)->base + (page)->offset + (index)*((page)->cellsize))
#define CellIndex(page,ptr)	(((unsigned char *)(ptr) - ((page)->base + (page)->offset))/(page)->cellsize)
#define CellInPage(page,ptr)	(((unsigned char *)(ptr) >= CellAddress(page,0)) && ((unsigned char *)(ptr) < CellAddress(page,(page)->maxcells)))

/* Tests whether 'page' is a pointer to a valid page in the chain with given 'head' */
static int ValidPage(POOL *page, POOL *head)
{
  static int first_time = 1;
  if (first_time) {
    PoolsWarning("ValidPage", "Sanity checks are on");
    first_time = 0;
  }

  if (!page)
    return 1;

  if (!
    ((page->magic == head->magic) && /* really pointer to the start of a POOL? */
     (page->cellsize == head->cellsize) && /* really a valid pool ? */
     (page->maxcells == head->maxcells) &&
     (page->overheadpercell == head->overheadpercell))) {
    fprintf(stderr, "Invalid page: wrong 'magic number' or incompatible pool.\n");
    return 0;
  }
  /*  
  while (head && head != page)
    head = head->next;
  if (!head) {
    fprintf(stderr, "Invalid page: not found in the pool.\n");
    return 0;
  }
  */
  return 1;  
}

#ifdef DDEBUG
static void TrailerTest(FILE *out, POOL *pool, POOL_TRAILER *trailer, unsigned char *ptr)
{
  unsigned magic2 = GetMagic(2);
  unsigned magic1 = GetMagic(1);
  if (trailer->magic == magic2) {
    if (trailer->pool == pool)
      fprintf(out, "(trailer copy");
    else
      fprintf(out, "(almost a trailer copy");
    if (((unsigned char *)(trailer->pool) > ptr-sizeof(POOL)) &&    /* valid pointer? */
	((unsigned char *)(trailer->pool) < ptr) &&
	(trailer->pool->magic == magic1)) { /* really pointer to the start of a POOL? */
      if (trailer->pool != pool)
	fprintf(out, ": not OK!!!");
    }
    fprintf(out, ") ");
  }
}

static void PrintPool(FILE *out, POOL *pool)
{
  int i;
  unsigned magic2 = GetMagic(2);

  fprintf(out, "Pool %p, next = %p, prev = %p, free = %p, magic1 = %u (should be %u), magic2 = %u\n",
	  pool, pool->next, pool->prev, pool->free, pool->magic, GetMagic(1), magic2);
  fprintf(out, "base = %p, offset = %d, base+offset = %p, cellsize = %d, overheadpercell = %d\n",
	  pool->base, pool->offset, pool->base+pool->offset, 
	  pool->cellsize, pool->overheadpercell);
  fprintf(out, "fixedoverhead = %d, maxcells = %d, reservedcells = %d, allocedcells = %d, nextfreecell = %d\n",
	  pool->fixedoverhead, pool->maxcells, pool->reservedcells, pool->allocedcells, pool->nextfreecell);
  fprintf(out, "Allocated cells: ");
  for (i=0; i<pool->maxcells; i++) {
    if (MARKED_ALLOCED(pool->allocmap,i)) {
      fprintf(out, " %d", i);
      TrailerTest(out, pool, (POOL_TRAILER *)CellAddress(pool,i), CellAddress(pool,i));
    }
  }
  fprintf(out, "\nTrailer: ");
  TrailerTest(out, pool, (POOL_TRAILER *)CellAddress(pool,pool->maxcells), CellAddress(pool,pool->maxcells));
  fprintf(out, "\n");
}
#endif /*DDEBUG*/

/* Searches the POOL chain for a page containing free cells. Returns
 * (POOL *)NULL if the pool is empty, or all pages are full. */
static POOL *FindFreePage(POOL *stor)
{
  int count = 0;

  if (stor) {
    /* first see if 'stor' itself still contains free cells */
    if (!PageIsFull(stor))
      return stor;

    /* no free cells, have a look at the page pointed to by stor->free,
     * which is the last page from which a cell was disposed.
     * With a little luck, this page still contains unallocated 
     * cells. */
    if (stor->free) {
#ifdef SANITY_CHECKS
      if (!ValidPage(stor->free, stor))
	PoolsFatal("FindFreePage", "Invalid free block pointer");	/* sanity check */
#endif
      if (!PageIsFull(stor->free))
	return stor->free;
    }
  }

  /* revert to searching the whole chain for a page containing free cells. */
  while (stor != (POOL *)NULL && count < MAX_SEARCH_PAGES) {
    if (!PageIsFull(stor)) 
      return stor;
    stor = stor->next;
    count++;
  }

  return (POOL *)NULL;
}

/* For a faster Dispose(), some cells in the page are reserved and filled
 * with a copy of the trailer. This trailer contains a magic number in order
 * to recognize it, and a pointer to the page to which it belongs. It is
 * used to determine the page to which an arbitrary given cell (pointer)
 * belongs. See FindWhichPage() and Dispose() below. */
static int ReserveCells(POOL *pool, POOL_TRAILER *trailer)
{
  int nrreserved;

#ifdef POOL_HEADERS
  /* no need for "trailers", since each cell is preceeded by a pointer
   * to the page to which it belongs. */
  nrreserved = 0;
#else /*POOL_HEADERS*/
  nrreserved = pool->maxcells / CELLS_PER_TRAILER;
  if (nrreserved > 0) {
    int i, gap = pool->maxcells/(nrreserved+1);
    for (i=1; i<=nrreserved; i++) {
      unsigned char *p = CellAddress(pool, i*gap);
      MARK_ALLOCED(pool->allocmap, i*gap);
      *(POOL_TRAILER *)p = *trailer;
    }
  }
#endif /*POOL_HEADERS*/
  return (pool->reservedcells = nrreserved);
}

/* Creates a new page containing cells of at leasy the given size. The
 * cells are aligned to an 'aling'-bytes boundary. */
static POOL *NewPage(unsigned int cellsize, unsigned int align)
{
  POOL *newpage;
  POOL_TRAILER *trailer;
  int reservedcells, fixedoverhead;

  newpage = (POOL *)Alloc(sizeof(POOL));
  memset(newpage, 0, sizeof(POOL)); /* clear the storage as it may
				     * be a previously disposed POOL page
				     * still containing old trailers that
				     * would confuse FindWhichPage() */

  newpage->magic = GetMagic(1);
  newpage->free = newpage->next = newpage->prev = (POOL *)NULL;

  /* cellsize must be at least MINCELLSIZE, add space for the header (0 bytes if
   * POOL_HEADERS is not defined, round up to a multiple of 'align' and compute how
   * much bytes we waste per cell. */
  newpage->cellsize = cellsize;
  if (newpage->cellsize < MINCELLSIZE) 
    newpage->cellsize = MINCELLSIZE;
  newpage->cellsize = ROUND_UP(newpage->cellsize+HEADERSIZE, align);
  newpage->overheadpercell = newpage->cellsize - cellsize;

  /* newpage->base + newpage->offset is an aligned address with enough space before
   * for a cell header (if one is needed). It will be the address of the first cell. */
  newpage->offset = ROUND_UP((unsigned)(newpage->base)+HEADERSIZE, align) - (unsigned)(newpage->base);

  newpage->maxcells = PAGESIZE/newpage->cellsize;
  newpage->allocedcells = 0;
  newpage->nextfreecell = 0;

  memset(newpage->allocmap, 0, ALLOCMAPSIZE);

  /* 'trailer' points to after the space provided for the cells. Put a magic number
   * there to mark the end of the POOL, and a pointer to the beginning of it. */
  trailer = (POOL_TRAILER *)CellAddress(newpage, newpage->maxcells);
  trailer->magic = GetMagic(2);
  trailer->pool = newpage;

  /* reserve a number of cells and put a copy of the trailer there */
  reservedcells = ReserveCells(newpage, trailer);

  fixedoverhead = sizeof(POOL) - (newpage->maxcells-newpage->reservedcells)*newpage->cellsize;

  if (sizeof(newpage->reservedcells)==1 && reservedcells>255)
    PoolsFatal("NewPage", "cannot store reservedcells > 255");
  if (sizeof(newpage->fixedoverhead)==1 && fixedoverhead>255)
    PoolsFatal("NewPage", "cannot store fixedoverhead > 255");

  newpage->reservedcells = reservedcells;
  newpage->fixedoverhead = fixedoverhead;
  bytesOverhead += newpage->fixedoverhead;

  return newpage;
}

/* the new page is attached in front of the previoud\sly allocated ones.
 * A pointer to the new chain is returned. */
static POOL *AttachPage(POOL *page, POOL *stor)
{
#ifdef DEBUG
DPRINTF("AttachPage (begin): page = %p, stor = %p\n", page, stor);
#endif

  if (stor) {
    stor->prev = page;
    page->free = stor->free;	/* remember last page from which a cell was */
    stor->free = (POOL *)NULL;	/* disposed at the head of the chain. */
  }
  page->next = stor;
  page->prev = (POOL *)NULL;

#ifdef DEBUG
DPRINTF("AttachPage (end): page = %p, stor = %p\n", page, stor);
#endif

  return page;
}

/* Releases a page from the chain */
static POOL *ReleasePage(POOL *page, POOL *stor)
{
  if (page == stor) 	/* first page in chain */
    stor = page->next;
  else 
    page->prev->next = page->next;

  if (page->next) page->next->prev = page->prev;

  if (stor && (!stor->free || stor->free == page))	/* make sure stor->free doesn't point to */
    stor->free = page->next;		/* the page to be deleted. */

  bytesOverhead -= page->fixedoverhead;

  Free((char *)page, sizeof(POOL));
  
  return stor;
}

/* allocates a cell in the given page */
static void *AllocCell(POOL *page)
{
  unsigned char *cellptr;

  MARK_ALLOCED(page->allocmap, page->nextfreecell);
  page->allocedcells ++;
  cellptr = CellAddress(page, page->nextfreecell);
#ifdef POOL_HEADERS
  ((CELLHDR *)(cellptr-HEADERSIZE))->page = page;
#endif
#ifdef SANITY_CHECKS
if (((unsigned)cellptr & (ALIGN-1)) != 0) 
  PoolsFatal("AllocCell", "cellptr = %p not aligned!!!\n", cellptr);
#endif
  bytesOverhead += page->overheadpercell;

  /* find next free cell */
  if (!PageIsFull(page)) {
    int startidx = page->nextfreecell;
    INCREMENT_MODULO(page->nextfreecell, page->maxcells);
    while (MARKED_ALLOCED(page->allocmap, page->nextfreecell) &&
	   page->nextfreecell != startidx) /* prevents infinite loops */ 
      INCREMENT_MODULO(page->nextfreecell, page->maxcells);

    if (page->nextfreecell == startidx)
      PoolsFatal("AllocCell", "Page not full, but can't seem to find a free cell???");
  }

  return (void *)cellptr;
}

/* Allocate a new cell with given cellsize from a POOL. The cells
 * are aligned to an 8 byte boundary unless POOL_ALIGN is defined
 * to be something else (but always a power of two!). */
void  *New(unsigned int cellsize, POOL **stor)
{
  POOL *page;
#ifdef SANITY_CHECKS
  void *ptr;
  if (*stor && !ValidPage((*stor)->free, *stor))
    PoolsFatal("New (start)", "Invalid free block pointer.");
#endif

  if ((page = FindFreePage(*stor)) == (POOL *)NULL) {
    page = NewPage(cellsize, ALIGN);
    *stor = AttachPage(page, *stor);
  }

#ifdef SANITY_CHECKS
  ptr = AllocCell(page);
  if (*stor && !ValidPage((*stor)->free, *stor))
    PoolsFatal("New (end)", "Invalid free block pointer.");
  return ptr;
#else
  return AllocCell(page);
#endif
}

/* same as above, but the cells are aligned to an 'align'-byte boundary.
 * 'align' must be a power of 2! */
void  *NewAligned(unsigned int cellsize, unsigned int align, POOL **stor)
{
  POOL *page;

  if ((page = FindFreePage(*stor)) == (POOL *)NULL) {
    page = NewPage(cellsize, align);
    *stor = AttachPage(page, *stor);
  }

  return AllocCell(page);
}

#ifndef POOL_HEADERS
/* FindWhichPage: returns the page to which the cell pointed to by 'ptr'
 * belongs. Not used when POOL_HEADERS is defined. In that case, each
 * cell is preceeded by a pointer to the cell to which it belongs. */
static POOL *FindWhichPage(unsigned char *ptr, POOL *stor)
{
  int i;
  unsigned char *p;
  unsigned magic2 = GetMagic(2), magic1 = GetMagic(1);
  POOL_TRAILER *trailer;

  /* stor->free is the last page in which a cell was disposed. With a little
   * luck, the cell to be disposed now is a cell in the same page. */
  if (stor->free) { 
#ifdef SANITY_CHECKS
    if (!ValidPage(stor->free, stor))
      PoolsFatal("FindWhichPage", "Invalid free block pointer."); /* sanity check */
#endif
    if (CellInPage(stor->free, ptr))
      return stor->free;
  }

  /* Another good candidate page that might contain the cell pointer is the
   * current head of the chain */
  if (stor != stor->free) {
    if (CellInPage(stor, ptr))
      return stor;
  }

  /* If ptr is a valid cell pointer, a trailer containing the magic2
   * number and a pointer to the cell to which it belongs follows 
   * within stor->maxcells*stor->cellsize bytes. */
  for (p=ptr+stor->cellsize, i=1; i<=stor->maxcells; p+=stor->cellsize, i++) {
    trailer = (POOL_TRAILER *)p;
    if ((trailer->magic == magic2) &&	      /* magic2 marks (a copy of) the trailer */
	((unsigned char *)(trailer->pool) > ptr-sizeof(POOL)) &&    /* valid pointer? */
	((unsigned char *)(trailer->pool) < ptr) &&
	(trailer->pool->magic == magic1)) {    /* really start of a pool ? */
      return trailer->pool;
    }
  }

  PoolsFatal("FindWhichPage", "Invalid cell pointer");
  return (POOL *)NULL;
}
#endif /*POOL_HEADERS*/

/* Dispose of the cell pointed to by cellptr from a POOL. */
void Dispose(void *cellptr, POOL **stor)
{
  POOL *page;

  if (cellptr == (void *)NULL) {
    PoolsError("Dispose", "NULL pointer passed");
    return;
  }

#ifdef SANITY_CHECKS
    if (*stor && !ValidPage((*stor)->free, *stor))
      PoolsFatal("Dispose (start)", "Invalid free block pointer.");
#endif

#ifdef POOL_HEADERS
  page = ((CELLHDR *)((unsigned char *)cellptr-HEADERSIZE))->page;
#else
  page = FindWhichPage(cellptr, *stor);
#endif
#ifdef SANITY_CHECKS
  if (!MARKED_ALLOCED(page->allocmap, CellIndex(page, cellptr)))
    PoolsFatal("Dispose", "cell was already disposed of");
  if (*stor && !ValidPage(page, *stor))
    PoolsFatal("Dispose", "FindWhichPage() returns invalid page.");
#endif

  page->allocedcells --;
  page->nextfreecell = CellIndex(page, cellptr);
#ifdef SANITY_CHECKS
  if (page->nextfreecell >= page->maxcells || 
      CellAddress(page,page->nextfreecell) != cellptr) {
    PoolsFatal("Dispose", "Invalid cell index.");	/* sanity check */
  }
#endif
  MARK_FREED(page->allocmap, page->nextfreecell);
  bytesOverhead -= page->overheadpercell;

  if (page->allocedcells <= 0) 
    *stor = ReleasePage(page, *stor);
  else
    (*stor)->free = page;

#ifdef SANITY_CHECKS
    if (*stor && !ValidPage((*stor)->free, *stor))
      PoolsFatal("Dispose (end)", "Invalid free block pointer.");
#endif
}

/* Disposes all cells in the pool at once. (Much faster than disposing 
 * of the cells one by one.) */
void DisposeAll(POOL **pool)
{
  while (*pool) {
    bytesOverhead -= (*pool)->allocedcells * (*pool)->overheadpercell;
    *pool = ReleasePage(*pool, *pool);
  }
}
