/* pools.h: memory management routines */

#ifndef _POOLS_H_
#define _POOLS_H_

#ifndef NULL
#define NULL (char *)0
#endif /*NULL*/

/* ------------------------------------------------------------------ */
/* routines for communication of informational (debug) and fatal error
 * messages to the user of the library. The routines to be specified
 * have one char * parameter: the message being passed. */
typedef void (*POOLS_MSG_CALLBACK_FUNC)(char *);

/* max. length of a message communicated by the pools library. */
#define POOLS_MAX_MESSAGE_LENGTH 200

/* default informational message callback function: unset. */
extern POOLS_MSG_CALLBACK_FUNC PoolsSetInfoCallback(POOLS_MSG_CALLBACK_FUNC func);

/* default fatal error callback function prints the message and exits
 * the program with -1 return code. */
extern POOLS_MSG_CALLBACK_FUNC PoolsSetFatalCallback(POOLS_MSG_CALLBACK_FUNC func);


/* ------------------------------------------------------------------ */
/* Memory management routines offering easy debug possibilities: debug output
 * is generated when the library is compiled with the -DDEBUG flag and
 * the user of the library has specified a callback function for informational
 * messages. */

/* allocate nrbytes bytes of memory and return a pointer to the allocated 
 * memory. */
extern char *Alloc(int nrbytes);

/* return the memory pointed to by buf to the system. The second argument
 * is only used to keep track of how much memory is being used by the
 * library. It should be the number of bytes being pointed to by buf,
 * same argument as when it was allocate dusing Alloc() above. */
extern void Free(char *buf, int nrbytes);

/* changes the size of the piece of memory being pointed to by buf:
 * new size is size+extra. extra can be negative. */
extern char *Realloc(char *buf, int size, int extra);

/* returns the total number of bytes allocated with the library,
 * excluding the overhead of the storage allocator. */
extern long GetMemoryUsage(void);


/* ------------------------------------------------------------------ */
/* A memory management system on top of the standard C runtime library
 * storage allocator. Goal: reduce the overhead (both time and space) when
 * dealing with a large quantity of small cells of equal size. Such cells
 * are grouped in pools. Each pool consists of zero, one or more blocks
 * of memory (pages). In each page there is place for a certain number of
 * cells. Only when a new page is needed (because e.g. all others are filled up)
 * the C runtime library storage allocator is called to create a new, empty
 * page. A new cell is allocated or disposed of using the routines New()
 * and Dispose() below. A user of these routines should never self
 * manipulate the POOL structures. 
 *
 *
 * Example of usage:
 *
 * Somewhere on top of a source file:
 *
 * 	typedef some_type THE_THING;
 *
 * 	#ifndef NOPOOLS
 *	#include "pools.h"
 *	static POOL *thingPool = (POOL *)NULL; 
 *	#define NEWTHING()	(THE_THING *)New(sizeof(THE_THING), &thingPool)
 *	#define DISPOSETHING(ptr) Dispose((unsigned char *)ptr, &thingPool)
 *	#else 
 *	#include <stdlib.h>
 *	#define NEWTHING	(THE_THING *)alloc(sizeof(THE_THING))
 *	#define DISPOSETHING(ptr) free((char *)ptr)
 *	#endif 
 *
 * Somewhere in a routine:
 * 
 *	THE_THING *thing;
 *	...
 *	thing = NEWTHING();
 *	...
 *	DISPOSETHING(thing);
 *
 *
 * Bugs: New() and Dispose() may not work because of alignment problems if 
 * the cellsize is not a multiple of the wordsize of the computer architecture 
 * this library is used on. */

#ifndef POOLS_PRIVATE	/* POOLS_PRIVATE is defined when compiling the 
			 * pools library when the private declarations 
			 * are needed. */

typedef struct POOL {	/* should correspond with the declaration in private.h */
  unsigned 	 magic;		/* magic number marking the start of
				 * a POOL struct. */
  struct POOL *next, *prev;	/* pointers to the next and previous pages
				 * in the pool */
  struct POOL 	 *free;		/* pointer to the last POOL from which a 
				 * cell was disposed. */
  unsigned short allocedcells, 	/* nr of allocated cells in page */
                 maxcells,	/* max. number of cells in page */
                 cellsize, 	/* effective (fixed) size of the 
				 * cells in the page, including header */
                 nextfreecell;	/* index of next free cell in page */
  unsigned short fixedoverhead,	/* fixed overhead for the POOL in bytes. */
                 reservedcells;	/* reserved cells (for trailer copies) */
  unsigned char  offset,	/* base+offset is addres of first cell
				 * in page (aligned to ALIGN bytes). */
                 overheadpercell; /* bytes overhead per cell */
  /*  unsigned char allocmap[0]; * bitmap indicating status of
				 * each cell (occupied or free) */
} POOL;

#else /* POOLS_PRIVATE */

#include "private.h"

#endif /* POOLS_PRIVATE */

/* Allocate a new cell with given cellsize from a POOL. The cells
 * are aligned to an 8 byte boundary unless POOL_ALIGN is defined
 * to be something else (but always a power of two!). */
extern void *New(unsigned int cellsize, POOL **stor);

/* same as above, but the cells are aligned to an 'align'-byte boundary.
 * 'align' must be a power of 2! */
extern void *NewAligned(unsigned int cellsize, unsigned int align, POOL **stor);

/* Dispose the cell pointed to by cellptr from a POOL. */
extern void Dispose(void *cellptr, POOL **stor);

/* Returns total bytes of overhead caused by the memory allocator. There
 * are three ways by which you can reduce the overhead: tune POOL_ALIGN, 
 * determining the default alignment of the cells, use NewAligned() instead
 * of New() wherever possible (watch out: architecture dependent!),
 * or don't define POOL_HEADERS, by which you save at least one pointer
 * per cell, at the cost of a much more expensive Dispose(). */
extern long GetMemoryOverhead(void);

/* Disposes all cells in the pool at once. (Much faster than disposing 
 * of the cells one by one.) */
extern void DisposeAll(POOL **pool);

#endif /*_POOLS_H_*/
